package com.nordicpeak.flowengine.evaluators.agequerystateevaluator;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.populators.UniversalSwedishSocialSecurityPopulator;
import se.unlogic.standardutils.validation.Birthdate6DigitValidator;
import se.unlogic.standardutils.validation.Birthdate8DigitValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.evaluators.calculatequerystateevaluator.CalculatedQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.CitizenIdentifierQuery;
import com.nordicpeak.flowengine.interfaces.CitizenIdentifierQueryInstance;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.queries.textfieldquery.TextField;
import com.nordicpeak.flowengine.queries.textfieldquery.TextFieldQuery;
import com.nordicpeak.flowengine.queries.textfieldquery.TextFieldQueryInstance;

public class AgeQueryStateEvaluationProviderModule extends CalculatedQueryStateEvaluationProviderModule<AgeQueryStateEvaluator> {
	
	
	@XSLVariable(prefix = "java.")
	protected String invalidSSNFormatMessage = "Invalid social security number or birthdate";
	
	private static final UniversalSwedishSocialSecurityPopulator SOCIAL_SECURITY_NUMBER_POPULATOR = new UniversalSwedishSocialSecurityPopulator();
	private static final Birthdate6DigitValidator BIRTHDATE_6_DIGIT_VALIDATOR = new Birthdate6DigitValidator();
	private static final Birthdate8DigitValidator BIRTHDATE_8_DIGIT_VALIDATOR = new Birthdate8DigitValidator();
	
	protected AgeQueryStateEvaluatorCRUD evaluatorCRUD;
	
	protected AnnotatedDAO<AgeQueryStateEvaluator> evaluatorDAO;
	
	protected QueryParameterFactory<AgeQueryStateEvaluator, Integer> evaluatorIDParamFactory;
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {
		
		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, AgeQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));
		
		if (upgradeResult.isUpgrade()) {
			
			log.info(upgradeResult.toString());
		}
		
		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(AgeQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);
		
		evaluatorCRUD = new AgeQueryStateEvaluatorCRUD(AgeQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}
	
	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return this.evaluatorCRUD.update(req, res, user, uriParser);
	}
	
	@Override
	protected BigDecimal getCalculatedValue(QueryInstance queryInstance, AgeQueryStateEvaluator evaluator, User poster) throws ValidationException {
		
		String citizenIdentifier = null;
		boolean test = false;
		
		if (queryInstance instanceof CitizenIdentifierQueryInstance) {
			
			CitizenIdentifierQueryInstance citizenIdentifierQueryInstance = (CitizenIdentifierQueryInstance) queryInstance;
			
			citizenIdentifier = citizenIdentifierQueryInstance.getCitizenIdentifier();
			test = citizenIdentifierQueryInstance.isTestCitizenIdentifier();
			
		} else if (queryInstance instanceof TextFieldQueryInstance) {
			
			citizenIdentifier = ((TextFieldQueryInstance) queryInstance).getFieldValue(evaluator.getTextFieldID());
		}
		
		if (citizenIdentifier != null) {
			
			if (!SOCIAL_SECURITY_NUMBER_POPULATOR.validateFormat(citizenIdentifier) && !BIRTHDATE_6_DIGIT_VALIDATOR.validateFormat(citizenIdentifier) && !BIRTHDATE_8_DIGIT_VALIDATOR.validateFormat(citizenIdentifier) && !test) {
				
				throw new ValidationException(new ValidationError("" + evaluator.getTextFieldID(), invalidSSNFormatMessage, ValidationErrorType.InvalidFormat));
			}
			
			citizenIdentifier = citizenIdentifier.replaceAll("[-+]", "");
			
			Calendar today = Calendar.getInstance();
			
			if (citizenIdentifier.length() == 10 || citizenIdentifier.length() == 6) {
				
				int currentCentury = today.get(Calendar.YEAR) / 100;
				int currentDecennium = today.get(Calendar.YEAR) % 100;
				int ssnDecennium = Integer.valueOf(citizenIdentifier.substring(0, 2));
				
				if (ssnDecennium > currentDecennium) {
					currentCentury -= 1;
				}
				
				citizenIdentifier = Integer.toString(currentCentury) + citizenIdentifier;
				
			}
			
			Calendar calendar = Calendar.getInstance();
			calendar.set(Calendar.YEAR, Integer.valueOf(citizenIdentifier.substring(0, 4)));
			calendar.set(Calendar.MONTH, Integer.valueOf(citizenIdentifier.substring(4, 6)) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.valueOf(citizenIdentifier.substring(6, 8)));
			
			int age = today.get(Calendar.YEAR) - calendar.get(Calendar.YEAR);
			
			if(evaluator.getCalculateMode().equals(AgeCalculateMode.BY_DATE) && !isBirthdayPassed(today, calendar)) {
				
				age--;
			}
			
			return BigDecimal.valueOf(age);
			
		}
		
		return null;
	}
	
	private boolean isBirthdayPassed(Calendar today, Calendar calendar) {

		if(today.get(Calendar.MONTH) < calendar.get(Calendar.MONTH)) {
			
			return false;
			
		} else if(today.get(Calendar.MONTH) == calendar.get(Calendar.MONDAY) && today.get(Calendar.DAY_OF_MONTH) < calendar.get(Calendar.DAY_OF_MONTH)) {
			
			return false;
		}
		
		return true;
	}

	@Override
	protected Class<AgeQueryStateEvaluator> getEvaluatorClass() {
		
		return AgeQueryStateEvaluator.class;
	}
	
	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {
		
		return TextFieldQuery.class.isAssignableFrom(queryClass) || CitizenIdentifierQuery.class.isAssignableFrom(queryClass);
	}
	
	@Override
	protected boolean supportsQueryInstance(QueryInstance queryInstance) {
		
		return queryInstance instanceof TextFieldQueryInstance || queryInstance instanceof CitizenIdentifierQueryInstance;
	}
	
	@Override
	protected AnnotatedDAO<AgeQueryStateEvaluator> getEvaluatorDAO() {
		
		return evaluatorDAO;
	}
	
	@Override
	protected QueryParameterFactory<AgeQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {
		
		return evaluatorIDParamFactory;
	}
	
	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {
		
		AgeQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<AgeQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));
		
		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());
		
		if (sourceQuery instanceof TextFieldQuery) {
			
			List<TextField> sourceTextFields = ((TextFieldQuery) sourceQuery).getFields();
			List<TextField> copyTextFields = ((TextFieldQuery) copyQuery).getFields();
			
			if (sourceTextFields != null && copyTextFields != null) {
				
				int textFieldIndex = 0;
				
				while (textFieldIndex < sourceTextFields.size()) {
					
					if (sourceTextFields.get(textFieldIndex).getTextFieldID().equals(evaluator.getTextFieldID())) {
						
						evaluator.setTextFieldID(copyTextFields.get(textFieldIndex).getTextFieldID());
						
						break;
					}
					
					textFieldIndex++;
				}
				
			}
		}
		
		this.evaluatorDAO.add(evaluator, transactionHandler, null);
		
	}
	
	@Override
	public AgeQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {
		
		AgeQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);
		
		if (evaluator.getTextFieldID() != null && query instanceof TextFieldQuery && ((TextFieldQuery) query).getTextFieldConversionMap() != null) {
			
			Map<Integer, Integer> conversionMap = ((TextFieldQuery) query).getTextFieldConversionMap();
			
			Integer newTextFieldID = conversionMap.get(evaluator.getTextFieldID());
			
			evaluator.setTextFieldID(newTextFieldID);
		}
		
		this.evaluatorDAO.add(evaluator, transactionHandler, null);
		
		return evaluator;
	}
	
}
