package com.nordicpeak.flowengine.alternativelist;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.collections.CollectionUtils;

import com.nordicpeak.flowengine.alternativelist.beans.AlternativeListDescriptor;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListInstance;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListProvider;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListQuery;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListQueryProvider;
import com.nordicpeak.flowengine.interfaces.QueryHandler;

public class AlternativeListHandlerModule extends AnnotatedForegroundModule {
	
	@InstanceManagerDependency
	private QueryHandler queryHandler;
	
	private ConcurrentHashMap<String, AlternativeListProvider> providerMap = new ConcurrentHashMap<String, AlternativeListProvider>();
	
	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {
		
		super.init(moduleDescriptor, sectionInterface, dataSource);
		
		if (!systemInterface.getInstanceHandler().addInstance(AlternativeListHandlerModule.class, this)) {
			
			throw new RuntimeException("Unable to register module " + moduleDescriptor + " in global instance handler using key " + AlternativeListHandlerModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}
	
	@Override
	public void unload() throws Exception {
		
		systemInterface.getInstanceHandler().removeInstance(AlternativeListHandlerModule.class, this);
		
		providerMap = null;
		
		super.unload();
	}
	
	public void addProvider(AlternativeListProvider provider) {
		
		if (providerMap.containsKey(provider.getProviderID())) {
			
			log.error("Provider " + provider.getProviderName() + " (ID: " + provider.getProviderID() + ") is already registered");
			return;
		}
		
		providerMap.put(provider.getProviderID(), provider);
		
		log.info("Provider " + provider.getProviderName() + " registered");
	}
		
	public boolean removeProvider(AlternativeListProvider provider) {
		
		boolean removed = providerMap.remove(provider.getProviderID()) != null;
		
		if (removed) {
			
			log.info("Provider " + provider.getProviderName() + " unregistered");
		}
		
		return removed;
	}
	
	public AlternativeListDescriptor getAlternativeList(String listName) {
		
		for(AlternativeListProvider listProvider : providerMap.values()) {
			
			AlternativeListDescriptor list = listProvider.getAlternativeList(listName);
			
			if (list != null) {
				
				return list;
			}
		}
		
		return null;
	}
	
	public Collection<AlternativeListDescriptor> getAlternativeLists() {
		
		ArrayList<AlternativeListDescriptor> lists = new ArrayList<AlternativeListDescriptor>(providerMap.size() * 10);
		
		for(AlternativeListProvider listProvider : providerMap.values()) {
			
			Collection<AlternativeListDescriptor> providerLists = listProvider.getAlternativeLists();
			
			if (!CollectionUtils.isEmpty(providerLists)) {
				
				lists.addAll(providerLists);
			}
		}
		
		return lists;
	}
	
	public AlternativeListProvider getProvider(String providerID) {
		
		return providerMap.get(providerID);
	}
	
	public AlternativeListInstance getListInstance(AlternativeListDescriptor descriptor, AttributeHandler attributeHandler, User poster){
		
		AlternativeListProvider provider = providerMap.get(descriptor.getProviderID());
		
		if(provider != null){
			
			return provider.getAlternativeListInstance(descriptor.getName(), attributeHandler, poster);
		}
		
		return null;
	}
	
	public List<AlternativeListQuery> getQueriesUsingAlternativeList(AlternativeListDescriptor descriptor) {
		
		if (queryHandler != null) {
			
			List<AlternativeListQueryProvider> queryProviders = queryHandler.getAssignableQueryProviders(AlternativeListQueryProvider.class);
			
			if (!CollectionUtils.isEmpty(queryProviders)) {
				
				List<AlternativeListQuery> queries = new ArrayList<AlternativeListQuery>(queryProviders.size() * 100);
				
				for (AlternativeListQueryProvider queryProvider : queryProviders) {
					
					Collection<? extends AlternativeListQuery> queryProviderQueries = queryProvider.getQueriesUsingAlternativeList(descriptor);
					
					if (!CollectionUtils.isEmpty(queryProviderQueries)) {
						
						queries.addAll(queryProviderQueries);
					}
				}
				
				return queries;
			}
		}
		
		return null;
	}
	
	public boolean hasQueriesUsingAlternativeList(AlternativeListDescriptor descriptor) throws SQLException {

		if (queryHandler != null) {
			
			List<AlternativeListQueryProvider> queryProviders = queryHandler.getAssignableQueryProviders(AlternativeListQueryProvider.class);
			
			if (!CollectionUtils.isEmpty(queryProviders)) {
				
				for (AlternativeListQueryProvider queryProvider : queryProviders) {
					
					if(queryProvider.hasQueriesUsingAlternativeList(descriptor)) {
						
						return true;
					}
				}
			}
		}
		
		return false;
	}		
	
	public void checkForListNameCollisions(AlternativeListProvider testedProvider) {
		
		Collection<AlternativeListDescriptor> testedProviderLists = testedProvider.getAlternativeLists();
		
		if (!CollectionUtils.isEmpty(testedProviderLists)) {
			
			Map<String, AlternativeListDescriptor> listMap = new HashMap<String, AlternativeListDescriptor>();
			
			for (AlternativeListProvider listProvider : providerMap.values()) {
				
				if (listProvider != testedProvider) {
					
					Collection<AlternativeListDescriptor> providerLists = listProvider.getAlternativeLists();
					
					if (!CollectionUtils.isEmpty(providerLists)) {
						
						for (AlternativeListDescriptor descriptor : providerLists) {
							listMap.put(descriptor.getName(), descriptor);
						}
					}
				}
			}
			
			for (AlternativeListDescriptor descriptor : testedProviderLists) {
				
				AlternativeListDescriptor existingDescriptor = listMap.get(descriptor.getName());
				
				if (existingDescriptor != null) {
					
					AlternativeListProvider existingProvider = providerMap.get(existingDescriptor.getProviderID());
					
					log.error("Provider " + testedProvider.getProviderName() + " contains list " + descriptor.getName() + " but a list with that name already exists from provider " + existingProvider.getProviderName() + ".");
				}
			}
		}
	}

}
