package com.nordicpeak.flowengine.alternativelist.provider;

import java.io.IOException;
import java.sql.SQLException;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.crud.BeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.CRUDDAO;
import se.unlogic.standardutils.serialization.SerializationUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.BeanRequestPopulator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.alternativelist.beans.AlternativeListDescriptor;
import com.nordicpeak.flowengine.alternativelist.exceptions.UnableToDeleteListInUseException;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListInstance;
import com.nordicpeak.flowengine.alternativelist.validationerrors.ListNameAlreadyInUseValidationError;

public abstract class BaseAlternativeListCRUD<BaseList extends BaseAlternativeList<BaseList, BaseListEvent>, ListInstance extends AlternativeListInstance, ListCRUD extends BaseAlternativeListCRUD<BaseList, ListInstance, ListCRUD, Provider, BaseListEvent>, Provider extends BaseAlternativeListProvider<BaseList, ListInstance, ListCRUD, Provider, BaseListEvent>, BaseListEvent extends BaseAlternativeListEvent<BaseList, BaseListEvent>> extends ModularCRUD<BaseList, Integer, User, Provider> {
	
	public BaseAlternativeListCRUD(BeanIDParser<Integer> idParser, CRUDDAO<BaseList, Integer> crudDAO, BeanRequestPopulator<BaseList> populator, String typeElementName, String typeLogName, String listMethodAlias, Provider callback) {
		
		super(idParser, crudDAO, populator, typeElementName, typeLogName, listMethodAlias, callback);
		
		setRequirePostForDelete(true);
	}
	
	public BaseAlternativeListCRUD(BeanIDParser<Integer> idParser, CRUDDAO<BaseList, Integer> crudDAO, BeanRequestPopulator<BaseList> populator, String typeElementName, String typeElementPluralName, String typeLogName, String typeLogPluralName, String listMethodAlias, Provider callback) {
		
		super(idParser, crudDAO, populator, typeElementName, typeElementPluralName, typeLogName, typeLogPluralName, listMethodAlias, callback);
	}
	
	@Override
	protected void appendShowFormData(BaseList list, Document doc, Element showTypeElement, User user, HttpServletRequest req, HttpServletResponse res, URIParser uriParser) throws SQLException, IOException, Exception {
		
		super.appendShowFormData(list, doc, showTypeElement, user, req, res, uriParser);
		
		XMLUtils.append(doc, showTypeElement, "Events", callback.getRecentListEvents(list));
	}

	@Override
	protected void checkDeleteAccess(BaseList list, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {
		
		super.checkDeleteAccess(list, user, req, uriParser);
		
		if (!CollectionUtils.isEmpty(callback.getAlternativeListHandlerModule().getQueriesUsingAlternativeList(callback.getNewDescriptor(list)))) {
			
			throw new UnableToDeleteListInUseException();
		}
	}
	
	@Override
	protected BaseList populateFromUpdateRequest(BaseList oldList, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {
		
		req.setAttribute("oldList", SerializationUtils.cloneSerializable(oldList));
		
		return super.populateFromUpdateRequest(oldList, req, user, uriParser);
	}
	
	// Needs to be called from sub classes
	protected void validate(BaseList list, HttpServletRequest req, User user, URIParser uriParser, List<ValidationError> validationErrors) throws ValidationException, Exception {
		
		@SuppressWarnings("unchecked")
		BaseList oldList = (BaseList) req.getAttribute("oldList");
		
		if (oldList != null && oldList == list) {
			throw new RuntimeException("Old list should be a separate object from new list");
		}
		
		AlternativeListDescriptor existingList = callback.getAlternativeListHandlerModule().getAlternativeList(list.getListName());
		
		if (existingList != null) {
			
			if (!existingList.getProviderID().equals(callback.getProviderID())) {
				
				validationErrors.add(new ListNameAlreadyInUseValidationError(existingList, callback.getAlternativeListHandlerModule().getProvider(existingList.getProviderID())));
				
			} else if (oldList == null || !oldList.getListName().equals(list.getListName())) {
				
				validationErrors.add(new ListNameAlreadyInUseValidationError(existingList, callback));
			}
		}
	}
	
	protected String getShowURL(BaseList list, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) {
		
		return req.getContextPath() + callback.getFullAlias() + "/show/" + list.getListID();
	}
	
	@Override
	protected ForegroundModuleResponse filteredBeanAdded(BaseList list, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		callback.cacheAlternatives(list);
		callback.addListEvent(callback.getEventListAddedMessage(), list, user);
		
		res.sendRedirect(getShowURL(list, req, res, user, uriParser));
		return null;
	}
	
	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(BaseList list, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		callback.cacheAlternatives(list);
		callback.addListEvent(callback.getEventListUpdatedMessage(), list, user);
		
		res.sendRedirect(getShowURL(list, req, res, user, uriParser));
		return null;
	}
	
	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(BaseList list, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		callback.unCacheList(list);
		
		return super.filteredBeanDeleted(list, req, res, user, uriParser);
	}
	
}
