package com.nordicpeak.flowengine.booking;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.apache.log4j.Level;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.EnumDropDownSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.ModuleConfigurationException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionDescriptor;
import se.unlogic.hierarchy.core.interfaces.events.EventHandler;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.enums.Order;
import se.unlogic.standardutils.fileattachments.FileAttachmentHandler;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.PositiveStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.booking.beans.Booking;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.beans.BookingEvent;
import com.nordicpeak.flowengine.booking.beans.BookingSlot;
import com.nordicpeak.flowengine.booking.beans.ExternalMessage;
import com.nordicpeak.flowengine.booking.beans.ExternalMessageAttachment;
import com.nordicpeak.flowengine.booking.beans.InternalMessage;
import com.nordicpeak.flowengine.booking.beans.InternalMessageAttachment;
import com.nordicpeak.flowengine.booking.cruds.ExternalMessageCRUD;
import com.nordicpeak.flowengine.booking.enums.BookingEventType;
import com.nordicpeak.flowengine.booking.enums.ExternalMessageType;
import com.nordicpeak.flowengine.booking.enums.InternalMessageType;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.MessageCRUDCallback;

public class BaseBookingModule extends AnnotatedRESTModule implements MessageCRUDCallback {

	public static final ValidationError QUEUE_DISABLED_VALIDATION_ERROR = new ValidationError("BookingDisabled");

	@XSLVariable(prefix = "java.")
	protected String eventLeaveMessage = "eventLeaveMessage";

	@XSLVariable(prefix = "java.")
	protected String eventArchiveMessage = "eventArchiveByManagerMessage";

	@XSLVariable(prefix = "java.")
	protected String eventAdminSubmittedMessage = "eventAdminSubmittedMessage";

	@XSLVariable(prefix = "java.")
	protected String eventAdminRestoredMessage = "eventAdminRestoredMessage";

	@XSLVariable(prefix = "java.")
	protected String eventSeeExternalMessage = "eventSeeExternalMessage";

	@XSLVariable(prefix = "java.")
	protected String eventSeeInternalMessage = "eventSeeInternalMessage";

	@XSLVariable(prefix = "java.")
	protected String eventChangedBookingSlot1 = "eventChangedBookingSlot";

	@XSLVariable(prefix = "java.")
	protected String eventChangedBookingSlot2 = "eventChangedBookingSlot";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Temp dir", description = "Directory for temporary files. Should be on the same filesystem as the file store for best performance. If not set system default temp directory will be used")
	protected String tempDir;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max request size", description = "Maxmium allowed request size in megabytes", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer maxRequestSize = 1000;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "RAM threshold", description = "Maximum size of files in KB to be buffered in RAM during file uploads. Files exceeding the threshold are written to disk instead.", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer ramThreshold = 500;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Max file size", description = "Maxmium allowed file size in megabytes", required = true, formatValidator = PositiveStringIntegerValidator.class)
	protected Integer maxFileSize = 50;

	@ModuleSetting
	@EnumDropDownSettingDescriptor(name = "Message sort order", description = "The order of internal and external messages when displayed in this module", required = true)
	protected Order messageSortOrder = Order.ASC;

	protected AnnotatedDAO<Booking> bookingDAO;
	protected AnnotatedDAO<BookingSlot> bookingSlotDAO;
	protected AnnotatedDAO<BookingEvent> bookingEventDAO;
	protected AnnotatedDAO<ExternalMessage> externalMessageDAO;
	protected AnnotatedDAO<InternalMessage> internalMessageDAO;
	protected AnnotatedDAO<ExternalMessageAttachment> externalMessageAttachmentDAO;

	protected AnnotatedDAOWrapper<BookingEvent, Integer> bookingEventDAOWrapper;

	protected ExternalMessageCRUD externalMessageCRUD;
	protected HierarchyAnnotatedDAOFactory daoFactory;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, CalendarAdminModule.class.getName(), new XMLDBScriptProvider(CalendarAdminModule.class.getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), false, true, false);

		bookingDAO = daoFactory.getDAO(Booking.class);
		bookingSlotDAO = daoFactory.getDAO(BookingSlot.class);
		bookingEventDAO = daoFactory.getDAO(BookingEvent.class);
		bookingEventDAOWrapper = bookingEventDAO.getWrapper(Integer.class);

		bookingEventDAOWrapper.addRelations(BookingEvent.ATTRIBUTES_RELATION);

		externalMessageDAO = daoFactory.getDAO(ExternalMessage.class);
		internalMessageDAO = daoFactory.getDAO(InternalMessage.class);
		externalMessageAttachmentDAO = daoFactory.getDAO(ExternalMessageAttachment.class);

		externalMessageCRUD = new ExternalMessageCRUD(externalMessageDAO, externalMessageAttachmentDAO, this, true);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN);
	}

	public boolean hasCalendarAccess(BookingCalendar calendar, User user) {

		return calendar.isEnabled() && AccessUtils.checkAccess(user, calendar);
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");

		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));

		doc.appendChild(documentElement);
		return doc;
	}

	public SectionDescriptor getSectionDescriptor() {

		return sectionInterface.getSectionDescriptor();
	}

	public UserHandler getUserHandler() {

		return systemInterface.getUserHandler();
	}

	public boolean hasActiveBookingInCalendar(BookingCalendar calendar, String citizenIdentifier) throws SQLException {

		return BookingUtils.hasActiveBookingInCalendar(dataSource, calendar, citizenIdentifier);
	}

	public boolean isActiveMemberOfBookingSlot(BookingSlot bookingSlot, String citizenIdentifier) throws SQLException {

		return BookingUtils.isActiveMemberOfBookingSlot(dataSource, bookingSlot, citizenIdentifier);
	}

	public EventHandler getEventHandler() {

		return systemInterface.getEventHandler();
	}

	@Override
	public int getRamThreshold() {

		return ramThreshold;
	}

	@Override
	public long getMaxRequestSize() {

		return maxRequestSize;
	}

	@Override
	public long getMaxFileSize() {

		return maxFileSize;
	}

	@Override
	public String getTempDir() {

		return tempDir;
	}

	@Override
	public boolean isOperatingStatusDisabled(ImmutableFlowInstance flowInstance, boolean manager) {

		throw new UnsupportedOperationException();
	}

	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> errors) throws SQLException, ModuleConfigurationException {

		throw new UnsupportedOperationException();
	}

	@WebPublic(alias = "externalattachment")
	public ForegroundModuleResponse getExternalMessageAttachment(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws ModuleConfigurationException, SQLException, AccessDeniedException, URINotFoundException {

		return externalMessageCRUD.getRequestedMessageAttachment(req, res, user, uriParser);
	}

	public BookingEvent addBookingEvent(Booking booking, BookingEventType eventType, String details, User user, Timestamp timestamp, Map<String, String> attributes) throws SQLException {

		if (timestamp == null) {

			timestamp = TimeUtils.getCurrentTimestamp();
		}

		if (details != null && details.length() > 1024) {

			details = StringUtils.substring(details, 1024);
		}

		BookingEvent bookingEvent = new BookingEvent(timestamp, user, eventType, details, attributes);
		bookingEvent.setBookingSlot(booking);

		bookingEventDAOWrapper.add(bookingEvent);

		return bookingEvent;
	}

	public ExternalMessage addBookingExternalMessage(Booking booking, String message, boolean postedByManager, User user, ExternalMessageType type, List<ExternalMessageAttachment> attachments) throws SQLException {

		if (message != null && message.length() > 16777215) {

			message = StringUtils.substring(message, 16777215);
		}

		ExternalMessage externalMessage = new ExternalMessage(message, user, TimeUtils.getCurrentTimestamp(), booking, attachments, postedByManager, type);

		externalMessageDAO.add(externalMessage);

		return externalMessage;
	}

	public InternalMessage addBookingInternalMessage(Booking booking, String message, User user, InternalMessageType type, List<InternalMessageAttachment> attachments) throws SQLException {

		if (message != null && message.length() > 16777215) {

			message = StringUtils.substring(message, 16777215);
		}

		InternalMessage internalMessage = new InternalMessage(message, user, TimeUtils.getCurrentTimestamp(), booking, attachments, type);

		internalMessageDAO.add(internalMessage);

		return internalMessage;
	}

	public String getEventAdminRestoredMessage() {

		return eventAdminRestoredMessage;
	}

	public String getEventSeeExternalMessage() {

		return eventSeeExternalMessage;
	}

	public String getEventAdminSubmittedMessage() {

		return eventAdminSubmittedMessage;
	}

	public String getEventChangedBookingSlotFrom() {

		return eventChangedBookingSlot1;
	}

	public String getEventChangedBookingSlotTo() {

		return eventChangedBookingSlot2;
	}

	public Map<Integer, User> getAllowedManagers(BookingCalendar calendar) {

		LinkedHashMap<Integer, User> managingUsers = new LinkedHashMap<>();

		List<Integer> managerIDs = calendar.getManagerUserIDs();
		List<Integer> managerGroupIDs = calendar.getManagerGroupIDs();

		if (!CollectionUtils.isEmpty(managerIDs)) {

			List<User> users = getUserHandler().getUsers(managerIDs, false, true);

			CollectionUtils.forEach(users, user -> managingUsers.put(user.getUserID(), user));
		}

		if (!CollectionUtils.isEmpty(managerGroupIDs)) {

			List<User> users = getUserHandler().getUsersByGroups(managerGroupIDs, true);

			CollectionUtils.forEach(users, user -> managingUsers.put(user.getUserID(), user));
		}

		return managingUsers;
	}

	public boolean isCalendarManager(BookingCalendar calendar, User user) {

		if (calendar == null || user == null) {

			return false;
		}

		return getAllowedManagers(calendar).containsKey(user.getUserID());
	}

	public Order getMessageSortOrder() {

		return messageSortOrder;
	}

	@Override
	public FileAttachmentHandler getFileAttachmentHandler() {

		// TODO Not implemented in this module yet
		return null;
	}

}
