package com.nordicpeak.flowengine.booking.cruds;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.crud.FragmentLinkScriptFilter;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.enums.Order;
import se.unlogic.standardutils.io.BinarySizeFormater;
import se.unlogic.standardutils.io.BinarySizes;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.populators.StringSwedishPhoneNumberPopulator;
import se.unlogic.standardutils.populators.SwedishSocialSecurityPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.SimpleRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.booking.BookingSlotAdminModule;
import com.nordicpeak.flowengine.booking.BookingUtils;
import com.nordicpeak.flowengine.booking.beans.Booking;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.beans.BookingEvent;
import com.nordicpeak.flowengine.booking.beans.BookingSlot;
import com.nordicpeak.flowengine.booking.beans.ExternalMessage;
import com.nordicpeak.flowengine.booking.beans.InternalMessage;
import com.nordicpeak.flowengine.booking.beans.PrimaryContact;
import com.nordicpeak.flowengine.booking.enums.BookingEventType;
import com.nordicpeak.flowengine.booking.enums.UserType;
import com.nordicpeak.flowengine.booking.events.BookingManuallyAddedEvent;
import com.nordicpeak.flowengine.booking.events.ExternalMessageAddedEvent;
import com.nordicpeak.flowengine.booking.exceptions.PersonAlreadyInBookingSlotValidationException;
import com.nordicpeak.flowengine.booking.exceptions.PersonAlreadyInCalendarValidationException;
import com.nordicpeak.flowengine.booking.interfaces.BookingExtensionProvider;
import com.nordicpeak.flowengine.enums.SenderType;

public class BookingCRUD extends ModularCRUD<Booking, Integer, User, BookingSlotAdminModule> {

	private static final String CITIZEN_IDENTIFIER_REQUEST_PARAMETER_NAME = "citizenIdentifier";
	private static final String PHONE_REQUEST_PARAMETER_NAME = "phone";
	private static final String SIZE_REQUEST_PARAMETER_NAME = "size";

	private static final String BOOKING_SLOT_REQUEST_ATTRIBUTE_NAME = "bookingSlot";

	private static final SwedishSocialSecurityPopulator CITIZEN_ID_POPULATOR = new SwedishSocialSecurityPopulator(true, true, false, true);

	protected final ExternalMessageCRUD externalMessageCRUD;
	protected final InternalMessageCRUD internalMessageCRUD;
	protected final AdvancedAnnotatedDAOWrapper<Booking, Integer> bookingSlotDAOWrapper;

	public BookingCRUD(AdvancedAnnotatedDAOWrapper<Booking, Integer> crudDAO, InternalMessageCRUD internalMessageCRUD, ExternalMessageCRUD externalMessageCRUD, BookingSlotAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), crudDAO, new SimpleRequestPopulator<>(Booking.class), "Booking", "booking", "", callback);

		this.bookingSlotDAOWrapper = crudDAO;
		this.internalMessageCRUD = internalMessageCRUD;
		this.externalMessageCRUD = externalMessageCRUD;

		this.setRequirePostForDelete(true);
	}

	@Override
	public Booking getBean(Integer beanID, String getMode, HttpServletRequest req) throws SQLException, AccessDeniedException {

		Booking booking = super.getBean(beanID, getMode, req);

		if (booking != null) {

			if (booking.getFlowInstanceID() != null) {

				booking.setFlowInstance(callback.getFlowInstance(booking.getFlowInstanceID()));
			}

			booking.getPoster(callback.getUserHandler());
		}

		return booking;
	}

	@Override
	protected List<Booking> getAllBeans(User user, HttpServletRequest req, URIParser uriParser) throws SQLException {

		throw new UnsupportedOperationException();
	}

	@Override
	protected void appendBean(Booking booking, Element targetElement, Document doc, User user) {

		if (callback.getMessageSortOrder() == Order.DESC) {

			if (booking.getInternalMessages() != null) {

				Collections.reverse(booking.getInternalMessages());
			}

			if (booking.getExternalMessages() != null) {

				Collections.reverse(booking.getExternalMessages());
			}

			XMLUtils.appendNewElement(doc, targetElement, "DefaultMessageOrder", "DESC");
		}

		targetElement.appendChild(booking.toXML(doc, callback.getUserHandler(), callback.hasSMSSender(), callback.isOverrideContactBySMS()));
	}

	@Override
	protected void appendShowFormData(Booking booking, Document doc, Element showTypeElement, User user, HttpServletRequest req, HttpServletResponse res, URIParser uriParser) throws Exception {

		super.appendShowFormData(booking, doc, showTypeElement, user, req, res, uriParser);

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (req.getParameter("externalmessage") != null && BookingUtils.isNewExternalMessagesAllowed(booking)) {

				if (BookingUtils.getUserType(booking, callback.getUserHandler(), callback.hasSMSSender(), false) == UserType.DIGITAL || booking.getPoster() != null) {

					//TODO append message or request parameters
					ExternalMessage externalMessage = externalMessageCRUD.add(req, user, doc, showTypeElement, booking, true);

					if (externalMessage != null) {

						BookingEvent bookingEvent = callback.addBookingEvent(booking, BookingEventType.MANAGER_MESSAGE_SENT, null, user, null, null);

						callback.getEventHandler().sendEvent(Booking.class, new ExternalMessageAddedEvent(booking, bookingEvent, externalMessage, SenderType.MANAGER), EventTarget.ALL);
						callback.getEventHandler().sendEvent(ExternalMessage.class, new CRUDEvent<>(CRUDAction.ADD, externalMessage), EventTarget.ALL);

						res.sendRedirect(req.getContextPath() + uriParser.getFormattedURI() + "#messages");
						return;
					}

				} else {

					log.warn("User " + user + " tried to add external message for Booking slot " + booking + " which has no digital user");
				}

			} else if (req.getParameter("internalmessage") != null) {

				//TODO append message or request parameters
				InternalMessage internalMessage = internalMessageCRUD.add(req, user, doc, showTypeElement, booking);

				if (internalMessage != null) {

					res.sendRedirect(req.getContextPath() + uriParser.getFormattedURI() + "#notes");

					callback.getEventHandler().sendEvent(InternalMessage.class, new CRUDEvent<>(CRUDAction.ADD, internalMessage), EventTarget.ALL);

					return;
				}
			}
		}

		if (hasDeleteAccess(booking, user)) {

			XMLUtils.appendNewElement(doc, showTypeElement, "DeleteAccess");
		}

		XMLUtils.appendNewElement(doc, showTypeElement, "FormatedMaxFileSize", BinarySizeFormater.getFormatedSize(callback.getMaxFileSize() * BinarySizes.MegaByte));

		XMLUtils.appendNewElement(doc, showTypeElement, "DefaultArchiveText", callback.getDefaultArchiveText());

		XMLUtils.appendNewElement(doc, showTypeElement, "IsNewExternalMessagesAllowed", BookingUtils.isNewExternalMessagesAllowed(booking));

		appendExtensionViewFragments(booking, doc, showTypeElement, user, req, res, uriParser);
	}

	protected void appendExtensionViewFragments(Booking booking, Document doc, Element element, User user, HttpServletRequest req, HttpServletResponse res, URIParser uriParser) throws SQLException, IOException, TransformerException {

		if (!callback.getBookingExtensionProviders().isEmpty()) {

			Element pluginFragments = XMLUtils.appendNewElement(doc, element, "ExtensionFragments");

			for (BookingExtensionProvider extensionProvider : callback.getBookingExtensionProviders()) {

				ViewFragment extensionSettings = extensionProvider.getAdminBookingViewFragment(booking, req, res, user, uriParser);

				if (extensionSettings != null) {

					pluginFragments.appendChild(extensionSettings.toXML(doc));
					FragmentLinkScriptFilter.addViewFragment(extensionSettings, req);
				}
			}
		}
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);

		if (!callback.hasPersonProvider()) {

			XMLUtils.appendNewElement(doc, addTypeElement, "NoPersonProvider");
		}

		appendFormData(doc, addTypeElement, req);
	}

	private void appendFormData(Document doc, Element element, HttpServletRequest req) {

		BookingSlot bookingSlot = getRequestBookingSlot(req);

		element.appendChild(bookingSlot.toXML(doc));
	}

	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {

		if (validationErrors != null) {

			return callback.listCalendars(req, user, uriParser, validationErrors);
		}

		throw new UnsupportedOperationException();
	}

	@Override
	protected Booking populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		BookingSlot bookingSlot = getRequestBookingSlot(req);
		Timestamp now = new Timestamp(System.currentTimeMillis());

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		Integer size = ValidationUtils.validateParameter(SIZE_REQUEST_PARAMETER_NAME, req, false, PositiveStringIntegerPopulator.getPopulator(), validationErrors);

		if (size == null) {

			size = 1;
		}

		validateBookingSlot(bookingSlot, size, true);

		Booking booking = null;

		try {

			booking = super.populateFromAddRequest(req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (booking != null) {

			PrimaryContact contact = populatePrimaryContact(req, bookingSlot, validationErrors);

			if (contact == null) {

				validationErrors.add(new ValidationError("ErrorImportingCitizen"));
			}

			//TODO required?

			booking.addEvent(new BookingEvent(now, user, BookingEventType.SUBMITTED, callback.getEventAdminSubmittedMessage(), null));

			booking.setBookingSlot(bookingSlot);
			booking.setAdded(new Timestamp(System.currentTimeMillis()));
			booking.setPrimaryContact(contact);
			booking.setSize(size);
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return booking;
	}

	public static void validateBookingSlot(BookingSlot bookingSlot, Integer size, boolean manager) throws ValidationException {

		validateBookingSlot(bookingSlot, size, manager, bookingSlot.getCalendar());
	}

	public static void validateBookingSlot(BookingSlot bookingSlot, Integer size, boolean manager, BookingCalendar calendar) throws ValidationException {

		if (bookingSlot == null) {

			throw new ValidationException(new ValidationError("NoLongerValidForSubmit"));
		}

		if (!manager && !bookingSlot.isBookable()) {

			throw new ValidationException(new ValidationError("NotBookable"));
		}

		if (bookingSlot.hasPassed()) {

			throw new ValidationException(new ValidationError("HasPassed"));
		}

		if (bookingSlot.getFreeSlots() <= 0) {

			throw new ValidationException(new ValidationError("NoFreeSlots"));
		}

		if (size > 1) {

			boolean allowedToSelectSize = calendar.isBookingSizeSelectorEnabled() && (manager || calendar.isAllowUserBookingSizeSelection());
			boolean selectedTooManySlots = !manager && calendar.getUserBookingMaxSize() != null && size > calendar.getUserBookingMaxSize();

			if (!allowedToSelectSize || selectedTooManySlots) {

				throw new ValidationException(new ValidationError("NotAllowedToSelectSize"));
			}
		}

		if (bookingSlot.getFreeSlots() < size) {

			throw new ValidationException(new ValidationError("NotEnoughFreeSlots"));
		}
	}

	private PrimaryContact populatePrimaryContact(HttpServletRequest req, BookingSlot bookingSlot, ArrayList<ValidationError> validationErrors) throws SQLException {

		String citizenIdentifier = populateCitizenIdentifier(req, validationErrors);

		if (StringUtils.isEmpty(citizenIdentifier)) {

			return null;
		}

		if (citizenIdentifier.length() == 10) {

			citizenIdentifier = SwedishSocialSecurityPopulator.addCentury(citizenIdentifier);
		}

		BookingCalendar calendar = bookingSlot.getCalendar();

		try {

			validateCitizenIdentifier(citizenIdentifier, calendar, bookingSlot);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		String phone = ValidationUtils.validateParameter(PHONE_REQUEST_PARAMETER_NAME, req, false, StringSwedishPhoneNumberPopulator.getPopulator(), validationErrors);

		return callback.getPrimaryContact(citizenIdentifier, phone);
	}

	private String populateCitizenIdentifier(HttpServletRequest req, ArrayList<ValidationError> validationErrors) {

		String citizenIdentifierParameter = req.getParameter(CITIZEN_IDENTIFIER_REQUEST_PARAMETER_NAME);

		if (StringUtils.isEmpty(citizenIdentifierParameter)) {

			validationErrors.add(new ValidationError(CITIZEN_IDENTIFIER_REQUEST_PARAMETER_NAME, ValidationErrorType.RequiredField));
			return null;
		}

		citizenIdentifierParameter = citizenIdentifierParameter.trim().replace("-", "");

		if (!CITIZEN_ID_POPULATOR.validateFormat(citizenIdentifierParameter)) {

			validationErrors.add(new ValidationError(CITIZEN_IDENTIFIER_REQUEST_PARAMETER_NAME, ValidationErrorType.InvalidFormat));
		}

		return CITIZEN_ID_POPULATOR.getValue(citizenIdentifierParameter);
	}

	private void validateCitizenIdentifier(String citizenIdentifier, BookingCalendar calendar, BookingSlot bookingSlot) throws ValidationException, SQLException {

		validateCitizenIdentifier(citizenIdentifier, calendar, bookingSlot, false, bookingSlotDAOWrapper.getAnnotatedDAO().getDataSource());
	}

	public static void validateCitizenIdentifier(String citizenIdentifier, BookingCalendar calendar, BookingSlot bookingSlot, boolean rebooking, DataSource dataSource) throws PersonAlreadyInCalendarValidationException, PersonAlreadyInBookingSlotValidationException, SQLException {

		if (!rebooking && !calendar.isAllowMultipleBookings() && BookingUtils.hasActiveBookingInCalendar(dataSource, calendar, citizenIdentifier)) {

			throw new PersonAlreadyInCalendarValidationException();
		}

		if (!calendar.isAllowDuplicateBookings() && BookingUtils.isActiveMemberOfBookingSlot(dataSource, bookingSlot, citizenIdentifier)) {

			throw new PersonAlreadyInBookingSlotValidationException();
		}
	}

	@Override
	public void checkAddAccess(User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		BookingSlot bookingSlot = getRequestBookingSlot(req);

		if (!callback.hasCalendarAccess(bookingSlot.getCalendar(), user)) {

			throw new AccessDeniedException("User " + user + " is not manager for calendar " + bookingSlot.getCalendar());
		}
	}

	@Override
	public void checkShowAccess(Booking booking, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!callback.hasBookingAccess(booking, user)) {

			throw new AccessDeniedException("User " + user + " is not manager for calendar " + booking.getBookingSlot().getCalendar());
		}
	}

	@Override
	public void checkUpdateAccess(Booking booking, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!callback.hasBookingAccess(booking, user)) {

			throw new AccessDeniedException("User " + user + " is not manager for calendar " + booking.getBookingSlot().getCalendar());
		}
	}

	public boolean hasDeleteAccess(Booking booking, User user) {

		return callback.hasBookingAccess(booking, user);
	}

	@Override
	protected void checkDeleteAccess(Booking booking, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!hasDeleteAccess(booking, user)) {

			throw new AccessDeniedException("User " + user + " is not manager for calendar " + booking.getBookingSlot().getCalendar());
		}

		if (!booking.getBookingSlot().hasPassed() && !booking.isArchived()) {

			throw new AccessDeniedException("User " + user + " is not allowed to delete unarchived and active booking " + booking);
		}
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(Booking booking, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		BookingSlot bookingSlot = getRequestBookingSlot(req);

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showslot/" + bookingSlot.getBookingSlotID());

		callback.getEventHandler().sendEvent(Booking.class, new BookingManuallyAddedEvent(booking, user), EventTarget.ALL);

		return null;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(Booking booking, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showslot/" + booking.getBookingSlot().getBookingSlotID());
		return null;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(Booking booking, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		callback.getEventHandler().sendEvent(Booking.class, new CRUDEvent<>(CRUDAction.DELETE, booking), EventTarget.ALL);

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showslot/" + booking.getBookingSlot().getBookingSlotID());
		return null;
	}

	private static BookingSlot getRequestBookingSlot(HttpServletRequest req) {

		return (BookingSlot) req.getAttribute(BOOKING_SLOT_REQUEST_ATTRIBUTE_NAME);
	}

	public static void setRequestBookingSlot(HttpServletRequest req, BookingSlot bookingSlot) {

		req.setAttribute(BOOKING_SLOT_REQUEST_ATTRIBUTE_NAME, bookingSlot);
	}

}
