package com.nordicpeak.flowengine.booking.queries;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.handlers.UserHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.dao.annotations.ManyToOne;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.BaseInvoiceLine;
import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.booking.BookingUtils;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.beans.BookingSlot;
import com.nordicpeak.flowengine.interfaces.ColumnExportableQueryInstance;
import com.nordicpeak.flowengine.interfaces.ImmutableAlternative;
import com.nordicpeak.flowengine.interfaces.InvoiceLine;
import com.nordicpeak.flowengine.interfaces.PaymentQueryInstance;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.interfaces.QueryRequestProcessor;
import com.nordicpeak.flowengine.interfaces.StringValueQueryInstance;
import com.nordicpeak.flowengine.interfaces.SubmitCheck;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryInstance;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryUtils;
import com.nordicpeak.flowengine.queries.fixedalternativesquery.FixedAlternativesQueryInstance;

@Table(name = "booking_query_instances")
@XMLElement
public class BookingQueryInstance extends BaseQueryInstance implements FixedAlternativesQueryInstance, PaymentQueryInstance, SubmitCheck, ColumnExportableQueryInstance, StringValueQueryInstance {

	private static final long serialVersionUID = -7761759005604863873L;

	public static final Field QUERY_RELATION = ReflectionUtils.getField(BookingQueryInstance.class, "query");
	public static final Field BOOKING_SLOT_RELATION = ReflectionUtils.getField(BookingQueryInstance.class, "bookingSlot");

	@DAOManaged
	@Key
	@XMLElement
	private Integer queryInstanceID;

	@DAOManaged
	@XMLElement
	private String comment;

	@DAOManaged
	@XMLElement
	private boolean accepted;

	@DAOManaged
	@XMLElement
	private Integer size;

	@DAOManaged(columnName = "queryID")
	@ManyToOne
	@XMLElement
	private BookingQuery query;

	@DAOManaged(columnName = "bookingSlotID")
	@ManyToOne
	@XMLElement
	private BookingSlot bookingSlot;

	private List<ValidationError> submitValidationErrors;

	@XMLElement
	private boolean rebooking;

	public Integer getQueryInstanceID() {

		return queryInstanceID;
	}

	public void setQueryInstanceID(Integer queryInstanceID) {

		this.queryInstanceID = queryInstanceID;
	}

	public String getComment() {

		return comment;
	}

	public void setComment(String freeTextAlternativeValue) {

		this.comment = freeTextAlternativeValue;
	}

	@Override
	public BookingQuery getQuery() {

		return query;
	}

	public void setQuery(BookingQuery query) {

		this.query = query;
	}

	@Override
	public void reset(MutableAttributeHandler attributeHandler) {

		resetAttribute(attributeHandler);
		super.reset(attributeHandler);

		this.comment = null;
		this.accepted = false;
		this.bookingSlot = null;
	}

	public void setAttribute(MutableAttributeHandler attributeHandler, UserHandler userHandler, BookingCalendar calendar) {

		if (query.getAttributeName() != null) {

			attributeHandler.setAttribute(query.getAttributeName() + ".calendar", bookingSlot.getCalendar().getName());
			attributeHandler.setAttribute(query.getAttributeName() + ".date", DateUtils.DATE_FORMATTER.format(bookingSlot.getStart()));
			attributeHandler.setAttribute(query.getAttributeName() + ".time", bookingSlot.getTimeFormatted(bookingSlot.getCalendar()));
			attributeHandler.setAttribute(query.getAttributeName() + ".location", bookingSlot.getLocation());
			List<User> managers = BookingUtils.getManagers(bookingSlot, calendar, userHandler, false);
			attributeHandler.setAttribute(query.getAttributeName() + ".managers", BookingUtils.getManagerNames(managers));
			attributeHandler.setAttribute(query.getAttributeName() + ".managerFirstNames", BookingUtils.getManagerFirstNames(managers));
			attributeHandler.setAttribute(query.getAttributeName() + ".size", size);
			attributeHandler.setAttribute(query.getAttributeName() + ".bookingSlotID", bookingSlot.getBookingSlotID());
		}
	}

	public void resetAttribute(MutableAttributeHandler attributeHandler) {

		if (query.getAttributeName() != null) {

			attributeHandler.removeAttribute(query.getAttributeName() + ".calendar");
			attributeHandler.removeAttribute(query.getAttributeName() + ".date");
			attributeHandler.removeAttribute(query.getAttributeName() + ".time");
			attributeHandler.removeAttribute(query.getAttributeName() + ".location");
			attributeHandler.removeAttribute(query.getAttributeName() + ".managers");
			attributeHandler.removeAttribute(query.getAttributeName() + ".managerFirstNames");
			attributeHandler.removeAttribute(query.getAttributeName() + ".size");
			attributeHandler.removeAttribute(query.getAttributeName() + ".bookingSlotID");
		}
	}

	public boolean isAccepted() {

		return accepted;
	}

	public void setAccepted(boolean accepted) {

		this.accepted = accepted;
	}

	public Integer getSize() {

		return size;
	}

	public void setSize(Integer size) {

		this.size = size;
	}

	@Override
	public String toString() {

		return this.getClass().getSimpleName() + " (queryInstanceID=" + queryInstanceID + ")";
	}

	@Override
	public Element toExportXML(Document doc, QueryHandler queryHandler) throws Exception {

		Element element = getBaseExportXML(doc);

		XMLUtils.appendNewElement(doc, element, "Comment", comment);
		XMLUtils.appendNewElement(doc, element, "Accepted", accepted);

		if (bookingSlot != null && bookingSlot.getCalendar() != null) {

			XMLUtils.appendNewElement(doc, element, "Calendar", bookingSlot.getCalendar().getName());
			XMLUtils.appendNewElement(doc, element, "Date", DateUtils.DATE_FORMATTER.format(bookingSlot.getStart()));
			XMLUtils.appendNewElement(doc, element, "Time", bookingSlot.getTimeFormatted(bookingSlot.getCalendar()));
			XMLUtils.appendNewElement(doc, element, "Location", bookingSlot.getLocation());
		}

		return element;
	}

	@Override
	public List<? extends ImmutableAlternative> getAlternatives() {

		if (accepted && bookingSlot != null) {

			return CollectionUtils.getList(query.getBookingSlotSelectedAlternative());
		}

		return null;
	}

	@Override
	public String getFreeTextAlternativeValue() {

		return null;
	}

	public BookingSlot getBookingSlot() {

		return bookingSlot;
	}

	public void setBookingSlot(BookingSlot bookingSlot) {

		this.bookingSlot = bookingSlot;
	}

	@Override
	public List<? extends InvoiceLine> getInvoiceLines() {

		BookingCalendar calendar = query.getCalendar();

		if (queryInstanceDescriptor.isPopulated() && accepted && calendar.getPrice() != null && calendar.getPrice() > 0) {

			String invoiceLineText = query.getInvoiceLineText();

			if (invoiceLineText == null) {

				invoiceLineText = query.getInvoiceLineTextFallback();
			}

			return Collections.singletonList(new BaseInvoiceLine(size, calendar.getPrice(), invoiceLineText, ""));
		}

		return null;
	}

	@Override
	public QueryRequestProcessor getQueryRequestProcessor(HttpServletRequest req, User user, User poster, URIParser uriParser, QueryHandler queryHandler, InstanceRequestMetadata requestMetadata) throws Exception {

		return BaseQueryUtils.getGenericQueryInstanceProvider(this.getClass(), queryHandler, queryInstanceDescriptor.getQueryDescriptor().getQueryTypeID()).getQueryRequestProcessor(this, req, user, poster, uriParser, requestMetadata);
	}

	@Override
	public boolean isValidForSubmit(MutableAttributeHandler attributeHandler, User poster, QueryHandler queryHandler, FlowInstanceManager flowInstanceManager) throws SQLException {

		return queryHandler.getQueryProvider(queryInstanceDescriptor.getQueryDescriptor().getQueryTypeID(), BookingQueryProviderModule.class).isValidBooking(this, poster, flowInstanceManager);
	}

	public List<ValidationError> getSubmitValidationErrors() {

		return submitValidationErrors;
	}

	public void setSubmitValidationErrors(List<ValidationError> submitErrors) {

		this.submitValidationErrors = submitErrors;
	}

	public void setSubmitValidationError(ValidationError submitError) {

		this.submitValidationErrors = Collections.singletonList(submitError);
	}

	public void resetSubmitValidationError() {

		this.submitValidationErrors = null;
	}

	public boolean isRebooking() {

		return rebooking;
	}

	public void setRebooking(boolean rebooking) {

		this.rebooking = rebooking;
	}

	@Override
	public String getStringValue() {

		if (bookingSlot == null) {

			return null;
		}

		StringBuilder builder = new StringBuilder();

		if (query.showLocationInList() && !StringUtils.isEmpty(bookingSlot.getLocation())) {

			builder.append(bookingSlot.getLocation());
			builder.append(" ");
		}

		builder.append(DateUtils.DATE_FORMATTER.format(bookingSlot.getStart()));
		builder.append(bookingSlot.getTimeFormatted(query.getCalendar()));

		if (!query.showLocationInList() && !StringUtils.isEmpty(bookingSlot.getLocation())) {

			builder.append(", ");
			builder.append(bookingSlot.getLocation());
		}

		return builder.toString();
	}

	@Override
	public List<String> getColumnLabels(QueryHandler queryHandler) {

		BookingQueryProviderModule queryProvider = queryHandler.getQueryProvider(getQueryInstanceDescriptor().getQueryDescriptor().getQueryTypeID(), BookingQueryProviderModule.class);

		List<String> labels = new ArrayList<>();

		String queryName = getQueryInstanceDescriptor().getQueryDescriptor().getName();

		labels.add(queryName + ": " + queryProvider.getExportDate());
		labels.add(queryName + ": " + queryProvider.getExportTime());
		labels.add(queryName + ": " + queryProvider.getExportLocation());

		if (query.showCommentField()) {

			labels.add(queryName + ": " + queryProvider.getExportComment());
		}

		return labels;
	}

	@Override
	public List<String> getColumnValues() {

		List<String> values = new ArrayList<>();

		values.add(bookingSlot != null ? DateUtils.DATE_FORMATTER.format(bookingSlot.getStart()) : null);
		values.add(bookingSlot != null ? bookingSlot.getTimeFormatted(query.getCalendar()) : null);
		values.add(bookingSlot != null ? bookingSlot.getLocation() : null);

		if (query.showCommentField()) {

			values.add(bookingSlot != null ? comment : null);
		}

		return values;
	}

}
