package com.nordicpeak.flowengine.integration.boverket;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.apache.log4j.Level;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.enums.EnumUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.integration.boverket.enums.PersonType;
import com.nordicpeak.flowengine.interfaces.FlowAdminExtensionViewProvider;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.XMLProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class BoverketIntegrationProvider extends AnnotatedForegroundModule implements FlowAdminExtensionViewProvider, ViewFragmentModule<ForegroundModuleDescriptor> {

	private static final Charset UTF8 = Charset.forName("UTF-8");

	public static final String BOVERKET_EXTERNAL_ID_ATTRIBUTE = "LovarendeUUID";
	public static final String BOVERKET_STATUS_ATTRIBUTE = "LovarendeStatus";

	protected static final RelationQuery FLOW_INSTANCE_EVENT_ATTRIBUTE_RELATION_QUERY = new RelationQuery(FlowInstanceEvent.ATTRIBUTES_RELATION);

	@ModuleSetting
	@XSLVariable(prefix = "java.")
	private String eventDetailText = "Not set";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Provider priority", description = "The priority of this provider", required = true)
	private Integer priority = 100;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable fragment XML debug", description = "Enables debugging of fragment XML")
	private boolean debugFragmententXML;

	@ModuleSetting(allowsNull = true)
	@TextAreaSettingDescriptor(name = "Supported actionID's", description = "The action ID's which will trigger export XML to be generated and stored when a submit event is detected", required = true)
	private List<String> supportedActionIDs;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "Connection timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer connectionTimeout = 5;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Request timeout", description = "Request timeout in seconds", formatValidator = NonNegativeStringIntegerValidator.class, required = true)
	private Integer requestTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "URL", description = "URL to Boverket API", required = true)
	protected String apiURL = "https://apitest.boverket.se/bygglov-api/api/v1";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Log generated XML", description = "Toggles logging of the generated XML")
	protected boolean logXML;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Disable posting", description = "Disable posting to Boverket API")
	protected boolean disablePosting;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Add integration external id attribute ", description = "Controls wheter integration external id should be added or not")
	protected boolean addExternalIDAttribute;

	private AnnotatedDAO<BoverketIntegrationSettings> settingsDAO;
	private QueryParameterFactory<BoverketIntegrationSettings, Integer> flowFamilyIDParamFactory;

	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;

	@InstanceManagerDependency
	protected XMLProvider xmlProvider;

	private FlowAdminModule flowAdminModule;

	private ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;

	protected boolean configured;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(BoverketIntegrationProvider.class, this)) {

			throw new RuntimeException("Unable to register module in global instance handler using key " + BoverketIntegrationProvider.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}

	@Override
	protected void moduleConfigured() throws Exception {

		this.configured = ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.WARN);

		this.viewFragmentTransformer = new ModuleViewFragmentTransformer<ForegroundModuleDescriptor>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());

		this.viewFragmentTransformer.setDebugXML(debugFragmententXML);

	}

	@Override
	public void unload() throws Exception {

		if (flowAdminModule != null) {

			setFlowAdminModule(null);
		}

		systemInterface.getInstanceHandler().removeInstance(BoverketIntegrationProvider.class, this);

		super.unload();
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, BoverketIntegrationProvider.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		settingsDAO = new AnnotatedDAO<BoverketIntegrationSettings>(dataSource, BoverketIntegrationSettings.class, new SimpleAnnotatedDAOFactory(dataSource));

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
	}
	
	@Override
	public FlowAdminExtensionShowView getShowView(Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerConfigurationException, TransformerException, SQLException {

		if (!AccessUtils.checkAccess(user, this.moduleDescriptor)) {

			return null;
		}

		Document doc = createDocument(req, uriParser, user);

		Element showViewElement = doc.createElement("ShowIntegrationSettings");
		doc.getDocumentElement().appendChild(showViewElement);

		showViewElement.appendChild(flow.toXML(doc));
		
		BoverketIntegrationSettings integrationSettings = getIntegrationSettings(flow);

		boolean enabled = false;
		
		if(integrationSettings != null) {
		
			XMLUtils.append(doc, showViewElement, integrationSettings);
			enabled = true;
		}

		return new FlowAdminExtensionShowView(viewFragmentTransformer.createViewFragment(doc), enabled);
	}

	public BoverketIntegrationSettings getIntegrationSettings(Flow flow) throws SQLException {

		HighLevelQuery<BoverketIntegrationSettings> query = new HighLevelQuery<BoverketIntegrationSettings>();

		query.addParameter(flowFamilyIDParamFactory.getParameter(flow.getFlowFamily().getFlowFamilyID()));

		return settingsDAO.get(query);
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");

		XMLUtils.appendNewElement(doc, document, "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, document, "StaticContentURL", staticContentModule.getModuleContentURL(moduleDescriptor));

		doc.appendChild(document);

		return doc;
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws SQLException {

		if (!configured) {

			log.warn("Module " + moduleDescriptor + " is not properly configured, ignoring event " + event);

			return;
		}

		if (event.getEvent().getEventType() != EventType.SUBMITTED || event.getActionID() == null || !supportedActionIDs.contains(event.getActionID())) {

			return;
		}

		processEvent(event.getFlowInstanceManager().getFlowInstance(), event.getEvent(), false);

	}

	public void processEvent(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, boolean generateXML) throws SQLException {

		BoverketIntegrationSettings integrationSettings = getIntegrationSettings((Flow) flowInstance.getFlow());

		if (integrationSettings == null) {

			return;
		}

		MutableAttributeHandler attributeHandler = (MutableAttributeHandler) flowInstance.getAttributeHandler();

		Document doc = XMLUtils.createDomDocument();

		Element errandElement = doc.createElement("Lovarende");
		doc.appendChild(errandElement);

		String arendeTypID = attributeHandler.getString("ArendetypID");
		String arendeTyp = attributeHandler.getString("Arendetyp");

		Element errandTypesElement = XMLUtils.appendNewElement(doc, errandElement, "Arendetyp");
		Element errandTypeElement = XMLUtils.appendNewElement(doc, errandTypesElement, "Arendetyp");

		XMLUtils.appendNewElement(doc, errandTypeElement, "Id", arendeTypID);
		XMLUtils.appendNewElement(doc, errandTypeElement, "Varde", arendeTyp);

		Integer behorigID = NumberUtils.toInt(attributeHandler.getString("BehorigID"));

		String behorigUUID = getBehorigUUID(behorigID);

		if (behorigUUID == null) {

			log.warn("Cannot find UUID with behorigID " + behorigID + " in cache. Behorig UUID is required skipping import for " + event + " belogning to flow instance " + flowInstance);

			return;
		}

		XMLUtils.appendNewElement(doc, errandElement, "Behorig_UUID", behorigUUID);

		String fastighetsbeteckning = attributeHandler.getString("Fastighet.propertyUnitDesignation");
		XMLUtils.appendNewElement(doc, errandElement, "Fastighetsbeteckning", fastighetsbeteckning);
		String address = attributeHandler.getString("Fastighet.address");
		XMLUtils.appendNewElement(doc, errandElement, "Adress", address);

		String omradeID = attributeHandler.getString("OmradeID");
		String omrade = attributeHandler.getString("Omrade");
		Element areaElement = XMLUtils.appendNewElement(doc, errandElement, "omrade");
		XMLUtils.appendNewElement(doc, areaElement, "Id", omradeID);
		XMLUtils.appendNewElement(doc, areaElement, "Varde", omrade);

		PersonType personType = EnumUtils.toEnum(PersonType.class, attributeHandler.getString("Persontyp"));

		if (personType != null) {

			Element personElement = XMLUtils.appendNewElement(doc, errandElement, "person");

			Element personRoleElement = XMLUtils.appendNewElement(doc, personElement, "personroll");
			XMLUtils.appendNewElement(doc, personRoleElement, "Id", 1);
			XMLUtils.appendNewElement(doc, personRoleElement, "Varde", "Byggherre");

			Element personTypeElement = XMLUtils.appendNewElement(doc, personElement, "persontyp");
			XMLUtils.appendNewElement(doc, personTypeElement, "Id", personType.getId());
			XMLUtils.appendNewElement(doc, personTypeElement, "Varde", personType.name());

			if (personType.equals(PersonType.Juridisk)) {

				XMLUtils.appendNewElement(doc, personElement, "Organisationsnummer", attributeHandler.getString("Organisationsnummer"));
				XMLUtils.appendNewElement(doc, personElement, "Namn", attributeHandler.getString("Namn"));
				XMLUtils.appendNewElement(doc, personElement, "Adressrad1", attributeHandler.getString("Adress"));
				XMLUtils.appendNewElement(doc, personElement, "Postnummer", attributeHandler.getString("Postnummer"));
				XMLUtils.appendNewElement(doc, personElement, "Ort", attributeHandler.getString("Ort"));

			} else {

				AttributeHandler userAttributeHandler = event.getPoster().getAttributeHandler();

				XMLUtils.appendNewElement(doc, personElement, "Organisationsnummer", CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(event.getPoster()));
				XMLUtils.appendNewElement(doc, personElement, "Namn", event.getPoster().getFirstname() + " " + event.getPoster().getLastname());
				XMLUtils.appendNewElement(doc, personElement, "Adressrad1", userAttributeHandler.getString("address"));
				XMLUtils.appendNewElement(doc, personElement, "Postnummer", userAttributeHandler.getString("zipCode"));
				XMLUtils.appendNewElement(doc, personElement, "Ort", userAttributeHandler.getString("postalAddress"));

			}

		} else {

			log.warn("Invalid Persontype found. Persontype is required skipping import for " + event + " belogning to flow instance " + flowInstance);

			return;
		}

		String xml;

		try {

			xml = XMLUtils.toString(doc, "UTF-8", true);

		} catch (TransformerFactoryConfigurationError | TransformerException e) {

			log.error("Error generating XML file for event " + event + " belonging to flow instance " + flowInstance, e);
			return;
		}

		if (logXML && log.isInfoEnabled()) {

			log.info("Generated XML of event " + event + " from flow instance " + flowInstance + ":\n" + xml);
		}

		if (disablePosting) {

			log.info("Posting is disabled, skipping posting of XML for event " + event + " belogning to flow instance " + flowInstance);

			return;
		}

		String endpoint = apiURL + "/lovarende";

		String externalID = null;
		String status = null;

		try {

			log.info("Sending generated XML of event " + event + " from flow instance " + flowInstance + " to " + endpoint);

			SimpleRequest request = new SimpleRequest(endpoint);

			request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
			request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

			request.addParameter("Behorig_UUID", behorigUUID);

			String response = HTTPUtils.sendHTTPPostRequest(request, new ByteArrayInputStream(xml.getBytes(UTF8)), UTF8);

			log.info("XML successfully sent with response: " + response);

			try {

				Document responseDoc = XMLUtils.parseXML(response, "UTF-8", false, false);

				XMLParser responseParser = new XMLParser(responseDoc);

				externalID = responseParser.getString("/Lovarende/UUID");

				if (StringUtils.isEmpty(externalID)) {

					log.warn("Got empty external ID for event " + event + " belonging to flow instance " + flowInstance + ", response:\n" + response);
					externalID = null;
				}

				status = responseParser.getString("/Lovarende/Lovarendestatus/Varde");

			} catch (Exception e) {

				log.warn("Error parsing response and getting external ID and for event " + event + " belonging to flow instance " + flowInstance, e);

			}

		} catch (IOException e) {

			log.warn("Error sending XML for event " + event + " belonging to flow instance " + flowInstance + " to URL " + endpoint, e);
		}

		if (externalID != null) {

			log.info("Got external ID: " + externalID);

			if (addExternalIDAttribute) {

				addOrUpdateFlowInstanceAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, externalID, flowInstance);

				attributeHandler.setAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, externalID);
			}

			addOrUpdateFlowInstanceAttribute(BOVERKET_EXTERNAL_ID_ATTRIBUTE, externalID, flowInstance);

			attributeHandler.setAttribute(BOVERKET_EXTERNAL_ID_ATTRIBUTE, externalID);

			addFlowInstanceEvent(flowInstance, EventType.OTHER_EVENT, eventDetailText + externalID, event.getPoster());

			if (status != null) {

				addOrUpdateFlowInstanceAttribute(BOVERKET_STATUS_ATTRIBUTE, status, flowInstance);

				attributeHandler.setAttribute(BOVERKET_STATUS_ATTRIBUTE, externalID);

			}

		}

	}

	public void addOrUpdateFlowInstanceAttribute(String attributeName, String attributeValue, ImmutableFlowInstance flowInstance) throws SQLException {

		FlowInstanceAttribute attribute = new FlowInstanceAttribute(attributeName, attributeValue);
		attribute.setFlowInstance((FlowInstance) flowInstance);

		flowAdminModule.getDAOFactory().getFlowInstanceAttributeDAO().addOrUpdate(attribute, null);

	}

	@WebPublic(alias = "toggle")
	public ForegroundModuleResponse toggleChangeOwner(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));

		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		BoverketIntegrationSettings settings = getIntegrationSettings(flow);

		if (settings == null && req.getParameter("enable") != null) {

			settings = new BoverketIntegrationSettings();
			settings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			settingsDAO.add(settings);

		} else if (settings != null && req.getParameter("disable") != null) {

			settingsDAO.delete(settings);
		}

		flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#boverketintegrationsettings");

		return null;

	}

	protected String getBehorigUUID(Integer behorigID) {

		SimpleRequest request = null;

		try {

			log.info("Getting behorig from Boverket API");

			request = new SimpleRequest(apiURL + "/behorig");

			request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
			request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

			request.addParameter("behorighetstyp", "1");

			String response = HTTPUtils.sendHTTPGetRequest(request, UTF8).getValue();

			Document responseDoc = XMLUtils.parseXML(response, "UTF-8", false, false);

			XMLParser responseParser = new XMLParser(responseDoc);

			XMLParser behorig = responseParser.getNode("/ArrayOfBehorig/Behorig[ID = " + behorigID + "]");

			if (behorig != null) {

				return behorig.getString("UUID");
			}

		} catch (Exception e) {

			log.warn("Unable get behorig from Boverket API using endpoint " + request.getEndpoint(), e);

		}

		return null;

	}

	protected FlowInstanceEvent addFlowInstanceEvent(ImmutableFlowInstance flowInstance, EventType eventType, String details, User user) {

		FlowInstanceEvent flowInstanceEvent = new FlowInstanceEvent();
		flowInstanceEvent.setFlowInstance((FlowInstance) flowInstance);
		flowInstanceEvent.setEventType(eventType);
		flowInstanceEvent.setDetails(details);
		flowInstanceEvent.setPoster(user);
		flowInstanceEvent.setStatus(flowInstance.getStatus().getName());
		flowInstanceEvent.setStatusDescription(flowInstance.getStatus().getDescription());

		flowInstanceEvent.getAttributeHandler().setAttribute("integration", true);

		flowInstanceEvent.setAdded(TimeUtils.getCurrentTimestamp());

		try {

			flowAdminModule.getDAOFactory().getFlowInstanceEventDAO().add(flowInstanceEvent, FLOW_INSTANCE_EVENT_ATTRIBUTE_RELATION_QUERY);

		} catch (Exception e) {

			throw new RuntimeException(e);
		}

		return flowInstanceEvent;
	}

	@Override
	public String getExtensionViewTitle() {

		return moduleDescriptor.getName();
	}

	@Override
	public int getPriority() {

		return priority;
	}

	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {

		return moduleDescriptor;
	}

	@Override
	public List<LinkTag> getLinkTags() {

		return links;
	}

	@Override
	public List<ScriptTag> getScriptTags() {

		return scripts;
	}

	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {

		if (flowAdminModule == null && this.flowAdminModule != null) {

			this.flowAdminModule.removeExtensionViewProvider(this);
		}

		this.flowAdminModule = flowAdminModule;

		if (this.flowAdminModule != null) {

			this.flowAdminModule.addExtensionViewProvider(this);
		}
	}

	public String getApiURL() {

		return apiURL;
	}

	public Integer getConnectionTimeout() {

		return connectionTimeout;
	}

	public Integer getRequestTimeout() {

		return requestTimeout;
	}

	@Override
	public int getModuleID() {

		return moduleDescriptor.getModuleID();
	}
}
