package com.nordicpeak.flowengine.evaluators.childcountquerystateevaluator;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.childrelationprovider.ChildRelationProvider;
import com.nordicpeak.childrelationprovider.ChildrenResponse;
import com.nordicpeak.childrelationprovider.exceptions.ChildRelationProviderException;
import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.evaluators.calculatequerystateevaluator.CalculatedQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;

public class ChildCountQueryStateEvaluationProviderModule extends CalculatedQueryStateEvaluationProviderModule<ChildCountQueryStateEvaluator> {

	protected AnnotatedDAO<ChildCountQueryStateEvaluator> evaluatorDAO;

	protected QueryParameterFactory<ChildCountQueryStateEvaluator, Integer> evaluatorIDParamFactory;

	protected ChildCountQueryStateEvaluatorCRUD evaluatorCRUD;

	@InstanceManagerDependency
	protected ChildRelationProvider childRelationProvider;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, ChildCountQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(ChildCountQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new ChildCountQueryStateEvaluatorCRUD(ChildCountQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		ChildCountQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<ChildCountQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	@Override
	public ChildCountQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		ChildCountQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);

		evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}

	@Override
	protected boolean supportsQueryInstance(QueryInstance queryInstance) {

		return true;
	}

	@Override
	protected Class<ChildCountQueryStateEvaluator> getEvaluatorClass() {

		return ChildCountQueryStateEvaluator.class;
	}

	@Override
	protected AnnotatedDAO<ChildCountQueryStateEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	@Override
	protected QueryParameterFactory<ChildCountQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return this.evaluatorIDParamFactory;
	}

	@Override
	protected List<Integer> getTargetQueryIDs(ChildCountQueryStateEvaluator evaluator, EvaluationCallback callback) {

		return Collections.singletonList(callback.getQueryInstance().getQueryInstanceDescriptor().getQueryDescriptor().getQueryID());
	}

	@Override
	protected BigDecimal getCalculatedValue(QueryInstance queryInstance, ChildCountQueryStateEvaluator evaluator, User poster) throws ValidationException {
		
		if (childRelationProvider == null) {
			
			log.warn("User " + poster + " tried to get child count but no child relation provider is configured!");
			return null;
		}

		if (poster != null) {

			try {

				String citizenIdentifier = poster.getAttributeHandler().getString(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE);
				
				if (citizenIdentifier == null) {
					
					return null;
				}
				
				ChildrenResponse childrenResponse = childRelationProvider.getChildren(citizenIdentifier, false, false);

				if (childrenResponse != null && childrenResponse.getChildrenTotalCount() != null) {

					return new BigDecimal(childrenResponse.getChildrenTotalCount());
				}
				
				return BigDecimal.ZERO;

			} catch (ChildRelationProviderException e) {

				log.error("Error getting children for user " + poster, e);
			}
		}
		
		return null;
	}

}
