package com.nordicpeak.flowengine.coloredtag.cruds;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.GenericCRUD;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.populators.HexColorPopulator;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.coloredtag.ColoredTagProviderModule;
import com.nordicpeak.flowengine.coloredtag.beans.ColoredTag;
import com.nordicpeak.flowengine.coloredtag.beans.ColoredTagSettings;
import com.nordicpeak.flowengine.populators.FlowAdminFragmentExtensionViewCRUDIDParser;

public class ColoredTagSettingsCRUD extends ModularCRUD<ColoredTagSettings, Integer, User, ColoredTagProviderModule> {

	private static final HexColorPopulator HEX_COLOR_POPULATOR = new HexColorPopulator();

	public ColoredTagSettingsCRUD(AnnotatedDAOWrapper<ColoredTagSettings, Integer> crudDAO, ColoredTagProviderModule callback) {

		super(FlowAdminFragmentExtensionViewCRUDIDParser.getInstance(), crudDAO, new AnnotatedRequestPopulator<ColoredTagSettings>(ColoredTagSettings.class), "Settings", "settings", "", callback);
	}

	@Override
	protected void checkUpdateAccess(ColoredTagSettings settings, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		checkAccess(req, uriParser);
	}

	@Override
	protected void checkDeleteAccess(ColoredTagSettings settings, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		checkAccess(req, uriParser);
	}

	private void checkAccess(HttpServletRequest req, URIParser uriParser) throws URINotFoundException {

		Flow flow = getFlow(req);

		if (flow == null) {

			throw new URINotFoundException(uriParser);
		}
	}

	@Override
	protected void appendUpdateFormData(ColoredTagSettings settings, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		XMLUtils.append(doc, updateTypeElement, getFlow(req));
		appendURLs(doc, getExtensionRequestURL(req));
	}

	@Override
	protected ColoredTagSettings populateFromUpdateRequest(ColoredTagSettings settings, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		List<ValidationError> validationErrors = new ArrayList<>();

		try {

			settings = super.populateFromUpdateRequest(settings, req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (settings.isEnabled()) {

			String[] tagIDs = req.getParameterValues("tagID");

			if (tagIDs != null) {

				List<ColoredTag> tags = populateTags(settings, req, validationErrors, tagIDs);

				settings.setTags(tags);

			} else {

				settings.setTags(null);
			}

			if (CollectionUtils.isEmpty(settings.getTags())) {

				validationErrors.add(new ValidationError("AtLeastOneTagRequired"));
			}
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return settings;
	}

	private List<ColoredTag> populateTags(ColoredTagSettings settings, HttpServletRequest req, List<ValidationError> validationErrors, String[] tagIDs) {

		List<ColoredTag> tags = new ArrayList<>(tagIDs.length);

		for (String tagID : tagIDs) {

			String name = ValidationUtils.validateParameter("name-" + tagID, callback.getTagNameDisplayName(), req, true, 1, 40, StringPopulator.getPopulator(), validationErrors);
			String hexColor = ValidationUtils.validateParameter("hexColor-" + tagID, callback.getTagHexColorDisplayName(), req, true, 1, 255, HEX_COLOR_POPULATOR, validationErrors);

			if (validationErrors.isEmpty()) {

				ColoredTag tag = settings.getTag(NumberUtils.toInt(tagID));

				if (tag == null) {

					tag = new ColoredTag();
				}

				tag.setName(name);
				tag.setHexColor(hexColor);

				tags.add(tag);
			}
		}

		return tags;
	}

	@Override
	protected void updateBean(ColoredTagSettings settings, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		if (settings.getSettingsID() == null) {

			crudDAO.add(settings);

		} else {

			crudDAO.update(settings);
		}
	}

	@Override
	protected ForegroundModuleResponse beanUpdated(ColoredTagSettings bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return null;
	}

	@Override
	protected ForegroundModuleResponse beanDeleted(ColoredTagSettings bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return null;
	}

	public FlowAdminExtensionShowView getShowView(String extensionRequestURL, Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws SQLException, TransformerException {

		Document doc = callback.createDocument(req, uriParser, user);

		Element showViewElement = doc.createElement("ShowSettings");
		doc.getDocumentElement().appendChild(showViewElement);

		appendURLs(doc, extensionRequestURL);

		ColoredTagSettings settings = crudDAO.get(flow.getFlowFamily().getFlowFamilyID());

		boolean enabled = false;

		if (settings != null && settings.isEnabled()) {

			enabled = true;

			XMLUtils.append(doc, showViewElement, settings);
		}

		return new FlowAdminExtensionShowView(callback.getViewFragmentTransformer().createViewFragment(doc, true), enabled);
	}

	private void appendURLs(Document doc, String extensionRequestURL) {

		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "extensionRequestURL", extensionRequestURL);
		XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "StaticContentURL", callback.getStaticContentURL());
	}

	@Override
	public ColoredTagSettings getRequestedBean(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, String getMode) throws SQLException, AccessDeniedException {

		Flow flow = getFlow(req);
		ColoredTagSettings settings = crudDAO.get(flow.getFlowFamily().getFlowFamilyID());

		if (settings == null && !getMode.equals(GenericCRUD.DELETE)) {

			settings = new ColoredTagSettings();
			settings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
		}

		return settings;
	}

	private Flow getFlow(HttpServletRequest req) {

		return (Flow) req.getAttribute("flow");
	}

	private String getExtensionRequestURL(HttpServletRequest req) {

		return (String) req.getAttribute("extensionRequestURL");
	}

}
