package com.nordicpeak.flowengine.api.culling;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryOperators;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.api.APIModule;
import com.nordicpeak.flowengine.culling.beans.DeletedFlowInstance;
import com.nordicpeak.flowengine.culling.listeners.DeletedFlowInstanceTimestampMillisecondElementableListener;

public class CullingAPIModule extends APIModule {

	private final DeletedFlowInstanceTimestampMillisecondElementableListener MILLISECOND_LISTENER = new DeletedFlowInstanceTimestampMillisecondElementableListener();

	@ModuleSetting
	@CheckboxSettingDescriptor(name="Force UTF-8 encoding", description="Forces UTF-8 encoding to be used instead of the system default encoding.")
	protected boolean forceUTF8;
	
	private AnnotatedDAO<DeletedFlowInstance> deletedFlowInstanceDAO;

	protected QueryParameterFactory<DeletedFlowInstance, Integer> flowInstanceIDParamFactory;
	protected QueryParameterFactory<DeletedFlowInstance, Integer> flowFamilyIDParamFactory;
	protected QueryParameterFactory<DeletedFlowInstance, Integer> flowIDParamFactory;
	protected QueryParameterFactory<DeletedFlowInstance, Timestamp> deletedParamFactory;
	protected QueryParameterFactory<DeletedFlowInstance, Integer> cullingIDParamFactory;

	@Override
	protected void moduleConfigured() throws Exception {

		String encoding;
		
		if(forceUTF8) {
			
			encoding = "UTF-8";
			
		}else {
			
			encoding = sectionInterface.getSystemInterface().getEncoding();
		}
		
		setResponseHandlers(CollectionUtils.getGenericSingletonList(new XMLResponseHandler(encoding)));
	}	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		deletedFlowInstanceDAO = daoFactory.getDAO(DeletedFlowInstance.class);

		flowInstanceIDParamFactory = deletedFlowInstanceDAO.getParamFactory("flowInstanceID", Integer.class);
		flowFamilyIDParamFactory = deletedFlowInstanceDAO.getParamFactory("flowFamilyID", Integer.class);
		flowIDParamFactory = deletedFlowInstanceDAO.getParamFactory("flowID", Integer.class);
		deletedParamFactory = deletedFlowInstanceDAO.getParamFactory("deleted", Timestamp.class);
		cullingIDParamFactory = deletedFlowInstanceDAO.getParamFactory("cullingID", Integer.class);
	}

	@RESTMethod(alias = "getevents/family/{flowFamilyID}", method = "get")
	public Document getDeletedFlowInstancesByFamily(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowFamilyID") Integer flowFamilyID) throws Throwable {

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);
		
		Integer fromCullingID = NumberUtils.toInt(req.getParameter("fromCullingID"));
		Integer toCullingID = NumberUtils.toInt(req.getParameter("toCullingID"));
		
		log.info("User " + user + " requesting deleted flow instances for flow family id " + flowFamilyID + " from date " + fromDate + " to date " + toDate + " from culling ID " + fromCullingID + " to culling ID " + toCullingID);

		List<DeletedFlowInstance> deletedFlowInstances = getDeletedFlowInstances(flowFamilyID, fromDate, toDate, fromCullingID, toCullingID, flowFamilyIDParamFactory);

		return createDeletedFlowInstancesDocument(deletedFlowInstances);
	}

	@RESTMethod(alias = "getevents/flow/{flowID}", method = "get")
	public Document getDeletedFlowInstancesByFlow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowID") Integer flowID) throws Throwable {

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromCullingID = NumberUtils.toInt(req.getParameter("fromCullingID"));
		Integer toCullingID = NumberUtils.toInt(req.getParameter("toCullingID"));
		
		log.info("User " + user + " requesting deleted flow instances for flowID " + flowID + " from date " + fromDate + " to date " + toDate + " from culling ID " + fromCullingID + " to culling ID " + toCullingID);

		List<DeletedFlowInstance> deletedFlowInstances = getDeletedFlowInstances(flowID, fromDate, toDate, fromCullingID, toCullingID, flowIDParamFactory);

		return createDeletedFlowInstancesDocument(deletedFlowInstances);
	}

	@RESTMethod(alias = "getevents/{flowInstanceID}", method = "get")
	public Document getDeletedFlowInstance(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "flowInstanceID") Integer flowInstanceID) throws Throwable {

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);

		Integer fromCullingID = NumberUtils.toInt(req.getParameter("fromCullingID"));
		Integer toCullingID = NumberUtils.toInt(req.getParameter("toCullingID"));
		
		log.info("User " + user + " requesting deleted flow instance with flowInstanceID " + flowInstanceID + " from date " + fromDate + " to date " + toDate + " from culling ID " + fromCullingID + " to culling ID " + toCullingID);

		List<DeletedFlowInstance> deletedFlowInstances = getDeletedFlowInstances(flowInstanceID, fromDate, toDate, fromCullingID, toCullingID, flowInstanceIDParamFactory);

		return createDeletedFlowInstancesDocument(deletedFlowInstances);
	}

	protected List<DeletedFlowInstance> getDeletedFlowInstances(Integer id, Date fromDate, Date toDate, Integer fromCullingID, Integer toCullingID, QueryParameterFactory<DeletedFlowInstance, Integer> paramFactory) throws SQLException {

		HighLevelQuery<DeletedFlowInstance> query = new HighLevelQuery<>();

		query.addParameter(paramFactory.getParameter(id));

		if (fromDate != null) {

			query.addParameter(deletedParamFactory.getParameter(new Timestamp(fromDate.getTime()), QueryOperators.BIGGER_THAN_OR_EQUALS));
		}

		if (toDate != null) {

			query.addParameter(deletedParamFactory.getParameter(new Timestamp(toDate.getTime()), QueryOperators.SMALLER_THAN_OR_EQUALS));
		}

		if (fromCullingID != null) {

			query.addParameter(cullingIDParamFactory.getParameter(fromCullingID, QueryOperators.BIGGER_THAN_OR_EQUALS));
		}

		if (toCullingID != null) {

			query.addParameter(cullingIDParamFactory.getParameter(toCullingID, QueryOperators.SMALLER_THAN_OR_EQUALS));
		}
		
		return deletedFlowInstanceDAO.getAll(query);
	}

	private Document createDeletedFlowInstancesDocument(List<DeletedFlowInstance> deletedFlowInstances) {

		Document doc = XMLUtils.createDomDocument();

		Element deletedFlowInstancesElement = doc.createElement("DeletedFlowInstances");
		doc.appendChild(deletedFlowInstancesElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addRootElementableListener(DeletedFlowInstance.class, MILLISECOND_LISTENER);

		XMLUtils.append(generatorDocument, deletedFlowInstancesElement, deletedFlowInstances);

		return doc;
	}

}
