package com.nordicpeak.flowengine.documentsigner.cruds;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.io.FilenameUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.emailutils.populators.EmailPopulator;
import se.unlogic.fileuploadutils.MultipartRequest;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.hierarchy.core.validationerrors.FileCountExceededValidationError;
import se.unlogic.hierarchy.core.validationerrors.FileNameLengthLimitExceededValidationError;
import se.unlogic.hierarchy.core.validationerrors.FileSizeLimitExceededValidationError;
import se.unlogic.hierarchy.core.validationerrors.InvalidFileExtensionValidationError;
import se.unlogic.hierarchy.core.validationerrors.UnableToParseFileValidationError;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.standardutils.dao.CRUDDAO;
import se.unlogic.standardutils.io.BinarySizeFormater;
import se.unlogic.standardutils.io.BinarySizes;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.populators.EnumPopulator;
import se.unlogic.standardutils.populators.NonNegativeStringIntegerPopulator;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.populators.SwedishSocialSecurity12DigitsWithoutMinusPopulator;
import se.unlogic.standardutils.random.RandomUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.documentsigner.DocumentSignerAdminModule;
import com.nordicpeak.flowengine.documentsigner.beans.DocumentSigner;
import com.nordicpeak.flowengine.documentsigner.beans.SigningDocument;
import com.nordicpeak.flowengine.documentsigner.enums.SignerType;
import com.nordicpeak.flowengine.documentsigner.enums.SigningDocumentStatus;
import com.nordicpeak.flowengine.documentsigner.listeners.DocumentSignerUserTitleElementableListener;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class SigningDocumentCRUD extends ModularCRUD<SigningDocument, Integer, User, DocumentSignerAdminModule> {

	private static final AnnotatedRequestPopulator<SigningDocument> POPULATOR = new AnnotatedRequestPopulator<>(SigningDocument.class);
	private static final EnumPopulator<SignerType> SIGNER_TYPE_POPULATOR = new EnumPopulator<>(SignerType.class);

	public SigningDocumentCRUD(CRUDDAO<SigningDocument, Integer> crudDAO, DocumentSignerAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), crudDAO, POPULATOR, "Document", "document", "", callback);
		
		setRequirePostForDelete(true);
	}

	@Override
	protected List<SigningDocument> getAllBeans(User user) throws SQLException {

		throw new UnsupportedOperationException();
	}
	
	@Override
	protected void appendBean(SigningDocument bean, Element targetElement, Document doc, User user) {
		
		XMLGeneratorDocument genDoc = new XMLGeneratorDocument(doc);
		genDoc.addAssignableFieldElementableListener(DocumentSigner.class, new DocumentSignerUserTitleElementableListener(callback.getTitleAttribute()));
		
		targetElement.appendChild(bean.toXML(genDoc));
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);

		appendFormData(doc, addTypeElement);
	}

	@Override
	protected void appendUpdateFormData(SigningDocument signingDocument, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendUpdateFormData(signingDocument, doc, updateTypeElement, user, req, uriParser);

		appendFormData(doc, updateTypeElement);
	}

	protected void appendFormData(Document doc, Element typeElement) {

		XMLUtils.appendNewElement(doc, typeElement, "FormatedMaxFileSize", BinarySizeFormater.getFormatedSize(callback.getMaxDocumentSize() * BinarySizes.MegaByte));
		XMLUtils.appendNewElement(doc, typeElement, "DefaultCancelInXDays", callback.getDefaultCancelInXDays());
		XMLUtils.appendNewElement(doc, typeElement, "DefaultDeleteAfterXDays", callback.getDefaultDeleteAfterXDays());
		XMLUtils.appendNewElement(doc, typeElement, "DefaultSendReminderInXDays", callback.getDefaultSendReminderInXDays());
		XMLUtils.appendNewElement(doc, typeElement, "SendEmailWhenSigningPartiallyDone", callback.sendEmailWhenSigningPartiallyDone());
	}

	@Override
	protected SigningDocument populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		List<ValidationError> validationErrors = new ArrayList<>();

		SigningDocument document = null;

		try {
			document = super.populateFromAddRequest(req, user, uriParser);
			document.setAdded(TimeUtils.getCurrentTimestamp());
			document.setPoster(user);
			document.setStatus(SigningDocumentStatus.SAVED);
			document.setHashAlgorithm(callback.getHashAlgorithm());

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		validatePopulation(document, req, user, uriParser, validationErrors);

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return document;
	}

	@Override
	protected SigningDocument populateFromUpdateRequest(SigningDocument document, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		List<ValidationError> validationErrors = new ArrayList<>();

		try {
			document = super.populateFromUpdateRequest(document, req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		validatePopulation(document, req, user, uriParser, validationErrors);

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return document;
	}

	private void validatePopulation(SigningDocument document, HttpServletRequest req, User user, URIParser uriParser, List<ValidationError> validationErrors) {

		List<DocumentSigner> signers = new ArrayList<>();

		String[] signerIDs = req.getParameterValues("signerID");

		if (signerIDs != null) {
			for (String signerID : signerIDs) {

				SignerType type = ValidationUtils.validateParameter("signer-type-" + signerID, req, true, SIGNER_TYPE_POPULATOR, validationErrors);
				Integer sortIndex = ValidationUtils.validateParameter("signer-sortIndex-" + signerID, req, true, NonNegativeStringIntegerPopulator.getPopulator(), validationErrors);

				if (type != null && sortIndex != null) {

					DocumentSigner signer = new DocumentSigner();
					signer.setType(type);
					signer.setSortIndex(sortIndex);

					if (type == SignerType.POSTER || type == SignerType.INTERNAL_USER) {

						Integer userID = ValidationUtils.validateParameter("signer-userID-" + signerID, req, true, NonNegativeStringIntegerPopulator.getPopulator(), validationErrors);

						if (userID != null) {

							User signerUser = callback.getUserHandler().getUser(userID, false, true);

							if (signerUser == null || !signerUser.isEnabled()) {

								validationErrors.add(new ValidationError("signer-userID-" + signerID, ValidationErrorType.InvalidFormat));
								continue;
							}
							
							if (StringUtils.isEmpty(CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(signerUser))) {

								validationErrors.add(new ValidationError("NoCitizenIDOnInternalUser", "", "signer-userID-" + signerID));
								continue;
							}
							
							signer.setUser(signerUser);

							signers.add(signer);
						}

					} else if (type == SignerType.EXTERNAL_USER) {

						int errorCount = validationErrors.size();

						String firstname = ValidationUtils.validateParameter("signer-firstname-" + signerID, req, true, StringPopulator.getPopulator(), validationErrors);
						String lastname = ValidationUtils.validateParameter("signer-lastname-" + signerID, req, false, StringPopulator.getPopulator(), validationErrors);
						String citizenID = ValidationUtils.validateParameter("signer-citizenID-" + signerID, req, false, SwedishSocialSecurity12DigitsWithoutMinusPopulator.getPopulator(), validationErrors);
						String email = ValidationUtils.validateParameter("signer-email-" + signerID, req, false, EmailPopulator.getPopulator(), validationErrors);
						String mobilePhone = ValidationUtils.validateParameter("signer-mobilePhone-" + signerID, req, false, StringPopulator.getPopulator(), validationErrors);

						if (errorCount == validationErrors.size()) {

							signer.setFirstname(firstname);
							signer.setLastname(lastname);
							signer.setEmail(email);
							signer.setMobilePhone(mobilePhone);

							if (email == null && mobilePhone == null) {

								validationErrors.add(new ValidationError("EmailOrPhoneRequired", signerID));
								continue;
							}

							if (citizenID == null) {

								signer.setUnsecureKey(RandomUtils.getRandomString(16, RandomUtils.MIXED_CASE_CHARACTERS_AND_DIGITS));

							} else {

								signer.setCitizenID(citizenID);
							}

							signers.add(signer);
						}
					}
				}
			}
		}

		if (signers.isEmpty()) {

			validationErrors.add(new ValidationError("signerID", ValidationErrorType.RequiredField));

		} else if (document != null) {

			document.setSigners(signers);
		}

		if (document != null) {

			SiteProfile profile = callback.getCurrentSiteProfile(req, user, uriParser);

			if (profile != null) {

				document.setProfileID(profile.getProfileID());
			}

			if ("true".equals(req.getParameter("sendOnSave")) && (document.getStatus() == SigningDocumentStatus.SAVED || document.getStatus() == SigningDocumentStatus.CANCELLED)) {

				document.setStatus(SigningDocumentStatus.SIGNING);
				document.setStartDate(TimeUtils.getCurrentTimestamp());
				document.setCompletedDate(null);
				document.setReminderSent(false);
				document.setExpiringSent(false);

				req.setAttribute("documentStatusChanged", true);
			}

			if (document.getDeleteAfterXDays() > 90) {
				validationErrors.add(new ValidationError("DeleteAfterXDaysTooLarge"));
			}
		}

		MultipartRequest mReq = (MultipartRequest) req;
		List<FileItem> fileItems = mReq.getFiles("file");

		File unsignedPDF = null;

		if (document != null) {

			unsignedPDF = callback.getOriginalPDF(document);
		}

		if (fileItems != null) {

			if (fileItems.size() > 1) {

				validationErrors.add(new FileCountExceededValidationError(1));
				return;
			}

			FileItem fileItem = fileItems.get(0);

			if (StringUtils.isEmpty(fileItem.getName()) || fileItem.getSize() == 0) {

				if (document != null && (document.getFilesize() == 0 || !unsignedPDF.exists())) {

					validationErrors.add(new ValidationError("file", ValidationErrorType.RequiredField));
				}

				return;
			}

			String fileExtension = FileUtils.getFileExtension(fileItem.getName());

			if (fileExtension == null || !"pdf".equalsIgnoreCase(fileExtension)) {

				validationErrors.add(new InvalidFileExtensionValidationError(FilenameUtils.getName(fileItem.getName())));
				return;
			}

			long maxFileSizeInBytes = callback.getMaxDocumentSize() * BinarySizes.MegaByte;

			if (fileItem.getSize() > maxFileSizeInBytes) {

				validationErrors.add(new FileSizeLimitExceededValidationError(FilenameUtils.getName(fileItem.getName()), fileItem.getSize(), maxFileSizeInBytes));
				return;
			}

			String filename = FilenameUtils.getName(fileItem.getName());

			if (filename.length() > 255) {

				validationErrors.add(new FileNameLengthLimitExceededValidationError(filename, filename.length(), 255));
				return;
			}
			
			try {
				File tmpFile1 = File.createTempFile("user-" + user.getUserID() + "-document-signing-upload-test1-", ".pdf", new File(callback.getTempDir()));
				File tmpFile2 = File.createTempFile("user-" + user.getUserID() + "-document-signing-upload-test2-", ".pdf", new File(callback.getTempDir()));

				try {
					FileUtils.writeFile(tmpFile1, fileItem.getInputStream(), true); // Could be removed if generatePDF is rewritten to handle both input file or stream
					
					callback.generatePDF(document, tmpFile1, tmpFile2, false, false);
					callback.writePDFA(tmpFile2, tmpFile1);
	
				} catch (Exception e) {
					
					log.warn("Exception " + e.getMessage() + " during early testing of uploaded PDF\"" + filename + "\" for document " + document);
					validationErrors.add(new UnableToParseFileValidationError(filename));
					return;
					
				} finally {
	
					FileUtils.deleteFile(tmpFile1);
					FileUtils.deleteFile(tmpFile2);
				}
				
			} catch (IOException e) {
				
				log.error("Unable to create temp files for early testing of uploaded PDF");
			}

			if (document != null) {

				document.setFilename(filename);
				document.setFilesize(fileItem.getSize());
			}

		} else if (document != null && (document.getFilesize() == 0 || !unsignedPDF.exists())) {

			validationErrors.add(new ValidationError("file", ValidationErrorType.RequiredField));
		}
	}

	@Override
	protected void beforeAddFilteredBeanCommit(SigningDocument signingDocument, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		super.beforeAddFilteredBeanCommit(signingDocument, req, user, uriParser);

		writeUploadedFile(signingDocument, req);
	}

	@Override
	protected void beforeUpdateFilteredBeanCommit(SigningDocument signingDocument, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		super.beforeUpdateFilteredBeanCommit(signingDocument, req, user, uriParser);

		writeUploadedFile(signingDocument, req);
	}

	protected void writeUploadedFile(SigningDocument document, HttpServletRequest req) {

		MultipartRequest mReq = (MultipartRequest) req;

		List<FileItem> fileItems = mReq.getFiles("file");

		if (fileItems != null) {

			FileItem fileItem = fileItems.get(0);

			if (fileItem.getSize() > 0) {

				File unsignedPDF = callback.getOriginalPDF(document);

				try {
					log.info("Writing new unsigned PDF to " + unsignedPDF + " for " + document);

					fileItem.write(unsignedPDF);

				} catch (Exception e) {

					log.error("Error saving file item " + fileItem.getName() + " as file " + unsignedPDF.getAbsolutePath(), e);
					//					validationErrors.add(new UnableToSaveFileValidationError(fileItem.getName()));
				}
			}
		}
	}

	@Override
	protected void checkShowAccess(SigningDocument signingDocument, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		callback.checkAccess(signingDocument, user);
	}

	@Override
	protected void checkUpdateAccess(SigningDocument signingDocument, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		callback.checkAccess(signingDocument, user);
		
		if (signingDocument.getStatus() == SigningDocumentStatus.SIGNING || signingDocument.getStatus() == SigningDocumentStatus.SIGNED) {
			throw new AccessDeniedException("May not modify signing document that is being signed");
		}
	}

	@Override
	protected void checkDeleteAccess(SigningDocument signingDocument, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		callback.checkAccess(signingDocument, user);
	}
	
	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {

		return callback.listDocuments(req, res, user, uriParser, validationErrors);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(SigningDocument signingDocument, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (signingDocument.getStatus() == SigningDocumentStatus.SIGNING) {

			callback.sendSigningNotifications(signingDocument, uriParser, true);

			if (redirectToSigning(signingDocument, res, uriParser)) {
				return null;
			}
		}

		return super.filteredBeanAdded(signingDocument, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(SigningDocument signingDocument, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Boolean documentStatusChanged = (Boolean) req.getAttribute("documentStatusChanged");

		if (signingDocument.getStatus() == SigningDocumentStatus.SIGNING && documentStatusChanged.booleanValue()) {

			callback.sendSigningNotifications(signingDocument, uriParser, true);

			if (redirectToSigning(signingDocument, res,uriParser)) {
				return null;
			}
		}

		return super.filteredBeanUpdated(signingDocument, req, res, user, uriParser);
	}

	private boolean redirectToSigning(SigningDocument signingDocument, HttpServletResponse res, URIParser uriParser) throws IOException {

		if (signingDocument.isUseSequentialSigning()) {

			DocumentSigner firstSigner = signingDocument.getSigners().get(0);

			if (firstSigner.getType() == SignerType.POSTER) {

				res.sendRedirect(callback.getSigningURL(signingDocument, firstSigner, uriParser));
				return true;
			}

		} else {

			for (DocumentSigner signer : signingDocument.getSigners()) {
				if (signer.getType() == SignerType.POSTER) {

					res.sendRedirect(callback.getSigningURL(signingDocument, signer, uriParser));
					return true;
				}
			}
		}

		return false;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(SigningDocument signingDocument, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (signingDocument.getStatus() == SigningDocumentStatus.SIGNING) {

			callback.sendSigningCancelledNotifications(signingDocument, null, uriParser);
		}

		FileUtils.deleteFile(callback.getOriginalPDF(signingDocument));
		FileUtils.deleteFile(callback.getSignedPDF(signingDocument));
		FileUtils.deleteFile(callback.getSignReceiptPDF(signingDocument));

		return super.filteredBeanDeleted(signingDocument, req, res, user, uriParser);
	}

}
