package com.nordicpeak.flowengine.integration.easit.callback;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jws.WebService;

import com.easit.bpsservice.proxy.BPS;
import com.easit.bpsservice.proxy.GetItemsRequest;
import com.easit.bpsservice.proxy.GetItemsResponse;
import com.easit.bpsservice.proxy.ImportItemsRequest;
import com.easit.bpsservice.proxy.ImportItemsRequest.ItemToImport;
import com.easit.bpsservice.proxy.ImportItemsResponse;
import com.easit.bpsservice.proxy.ImportItemsResponse.ImportItemResult;
import com.easit.bpsservice.proxy.ImportItemsResponse.ImportItemResult.ReturnValues;
import com.easit.bpsservice.proxy.ImportItemsResponse.ImportItemResult.ReturnValues.ReturnValue;
import com.easit.bpsservice.proxy.Item;
import com.easit.bpsservice.proxy.Item.Property;
import com.easit.bpsservice.proxy.PingResponse;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.StatusChangedByManagerEvent;
import com.nordicpeak.flowengine.integration.easit.beans.IntegrationSettings;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;

import se.unlogic.hiearchy.foregroundmodules.jaxws.BaseWSModuleService;
import se.unlogic.hiearchy.foregroundmodules.jaxws.WSModuleCallback;
import se.unlogic.hiearchy.foregroundmodules.jaxws.WSModuleInstanceResolver;
import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.interfaces.instances.InstanceListener;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.time.TimeUtils;

@WebService(endpointInterface = "com.easit.bpsservice.proxy.BPS", name = "BPSService", serviceName = "BPSService", targetNamespace="http://www.easit.com/bps/schemas")
@WSModuleInstanceResolver
public class EasitIntegrationCallback extends BaseWSModuleService implements BPS, InstanceListener<FlowAdminModule> {

	protected static final RelationQuery FLOW_INSTANCE_EVENT_ATTRIBUTE_RELATION_QUERY = new RelationQuery(FlowInstanceEvent.ATTRIBUTES_RELATION);

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Debug logging", description = "Toggles debug logging when calling esit web service")
	protected boolean enableDebugLoggning;

	protected FlowAdminModule flowAdminModule;

	protected FlowEngineDAOFactory daoFactory;
	protected QueryParameterFactory<FlowInstance, Integer> flowInstanceIDParamFactory;
	protected QueryParameterFactory<Status, Integer> statusIDParamFactory;
	protected QueryParameterFactory<Status, String> statusNameParamFactory;
	protected QueryParameterFactory<Status, Flow> statusFlowParamFactory;

	protected QueryParameterFactory<IntegrationSettings, Integer> flowFamilyIDParamFactory;

	protected AnnotatedDAO<IntegrationSettings> settingsDAO;

	@Override
	public void init(WSModuleCallback callback) throws Exception {

		super.init(callback);

		daoFactory = new FlowEngineDAOFactory(callback.getDataSource(), callback.getSystemInterface().getUserHandler(), callback.getSystemInterface().getGroupHandler());

		flowInstanceIDParamFactory = daoFactory.getFlowInstanceDAO().getParamFactory("flowInstanceID", Integer.class);
		statusIDParamFactory = daoFactory.getStatusDAO().getParamFactory("statusID", Integer.class);
		statusNameParamFactory = daoFactory.getStatusDAO().getParamFactory("name", String.class);
		statusFlowParamFactory = daoFactory.getStatusDAO().getParamFactory("flow", Flow.class);

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(callback.getDataSource());

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);

		ModuleUtils.setModuleSettings(this, BaseWSModuleService.class, callback.getModuleDescriptor().getMutableSettingHandler(), callback.getSystemInterface());

		callback.getSystemInterface().getInstanceHandler().addInstanceListener(FlowAdminModule.class, this);
	}

	@Override
	public void unload() throws Exception {

		callback.getSystemInterface().getInstanceHandler().removeInstanceListener(FlowAdminModule.class, this);

		super.unload();
	}

	protected FlowInstance getFlowInstance(Integer flowInstanceID) {

		try {

			return flowAdminModule.getFlowInstance(flowInstanceID, null, FlowInstance.FLOW_RELATION, FlowInstance.STATUS_RELATION, Flow.FLOW_FAMILY_RELATION, FlowInstance.ATTRIBUTES_RELATION, FlowInstance.MANAGERS_RELATION);

		} catch (SQLException e) {

			throw new RuntimeException(e);
		}

	}

	@Override
	public ImportItemsResponse importItems(ImportItemsRequest request) {

		if (request.getImportHandlerIdentifier() == null) {
			return sendErrorResponse("Import handler identifier cannot be null");
		}

		List<ItemToImport> items = request.getItemToImport();

		if (items.isEmpty()) {

			logDebug("No import items specified in request from user " + callback.getUser());

			return sendErrorResponse("No import items specified in request");
		}

		if (request.getImportHandlerIdentifier().equalsIgnoreCase("SetStatus")) {

			for (ItemToImport item : items) {

				Map<String, String> properties = getItemProperties(item);

				if (CollectionUtils.isEmpty(properties)) {

					logDebug("No properties for item " + item.getId() + " specified in request from user " + callback.getUser());

					return sendErrorResponse("No properties specified in item");
				}

				Integer flowInstanceID = NumberUtils.toInt(properties.get(Constants.FLOWINSTANCEID));

				if (flowInstanceID == null) {

					logDebug("No valid property '" + Constants.FLOWINSTANCEID + "' found in item " + item.getId() + " for request from user " + callback.getUser());

					return sendErrorResponse("No valid property for '" + Constants.FLOWINSTANCEID + "' specified in item " + item.getId());
				}

				String statusAlias = properties.get(Constants.STATUS_ALIAS);

				if (statusAlias == null) {

					logDebug("No valid property '" + Constants.STATUS_ALIAS + "' found in item " + item.getId() + " for request from user " + callback.getUser());

					return sendErrorResponse("No valid property for '" + Constants.STATUS_ALIAS + "' specified in item " + item.getId());
				}

				return setStatus(flowInstanceID, statusAlias);

			}

		} else if (request.getImportHandlerIdentifier().equalsIgnoreCase("AddEvent")) {

			for (ItemToImport item : items) {

				Map<String, String> properties = getItemProperties(item);

				if (CollectionUtils.isEmpty(properties)) {

					logDebug("No properties for item " + item.getId() + " specified in request from user " + callback.getUser());

					return sendErrorResponse("No properties specified in item");
				}

				Integer flowInstanceID = NumberUtils.toInt(properties.get(Constants.FLOWINSTANCEID));

				if (flowInstanceID == null) {

					logDebug("No valid property '" + Constants.FLOWINSTANCEID + "' found in item " + item.getId() + " for request from user " + callback.getUser());

					return sendErrorResponse("No valid property for '" + Constants.FLOWINSTANCEID + "' specified in item " + item.getId());
				}

				String message = properties.get(Constants.EVENT_MESSAGE);

				Date date = null;

				if (properties.get(Constants.EVENT_DATE) != null) {
					date = DateUtils.getDate(DateUtils.DATE_TIME_FORMATTER, properties.get(Constants.EVENT_DATE));
				}

				if (date == null) {
					date = new Date();
				}

				return addEvent(flowInstanceID, new Timestamp(date.getTime()), message);
			}

		} else {

			log.info("Unknown import handler identifier " + request.getImportHandlerIdentifier() + " in request from user " + callback.getUser());

			return sendErrorResponse("Unkown import handler identifier " + request.getImportHandlerIdentifier());
		}

		return null;
	}

	private ImportItemsResponse addEvent(Integer flowInstanceID, Timestamp date, String message) {

		try {

			checkDependencies();

			FlowInstance flowInstance = getFlowInstance(flowInstanceID);

			if (flowInstance == null) {

				logDebug("Flow instance with ID " + flowInstanceID + " not found, requested by user " + callback.getUser());

				return sendErrorResponse("Flow instance with ID " + flowInstanceID + " not found");
			}

			if (!integrationActivated(flowInstance.getFlow())) {

				return sendErrorResponse("Integration for flowinstance " + flowInstance + " not activated");
			}

			log.info("User " + callback.getUser() + " requested add event for flow instance " + flowInstance);

			FlowInstanceEvent flowInstanceEvent = addFlowInstanceEvent(flowInstance, EventType.OTHER_EVENT, message, callback.getUser(), date);

			ImportItemResult result = new ImportItemResult();
			result.setResult("EventAdded");

			ReturnValue returnValue = new ReturnValue();
			returnValue.setName("EventID");
			returnValue.setContent(flowInstanceEvent.getEventID().toString());

			ReturnValues returnValues = new ReturnValues();
			returnValues.getReturnValue().add(returnValue);

			result.setReturnValues(returnValues);

			return sendResponse(result);

		} catch (RuntimeException e) {

			log.error("Error adding event", e);

			throw e;
		}

	}

	private ImportItemsResponse setStatus(Integer flowInstanceID, String statusAlias) {

		try {

			checkDependencies();

			FlowInstance flowInstance = getFlowInstance(flowInstanceID);

			if (flowInstance == null) {

				logDebug("Flow instance with ID " + flowInstanceID + " not found, requested by user " + callback.getUser());

				return sendErrorResponse("Flow instance with ID " + flowInstanceID + " not found");
			}

			if (!integrationActivated(flowInstance.getFlow())) {

				return sendErrorResponse("Integration for flowinstance " + flowInstance + " not activated");
			}

			log.info("User " + callback.getUser() + " requested set status for flow instance " + flowInstance);

			Status status = getStatus(flowInstance.getFlow(), statusAlias);

			if (status == null) {

				log.info("Status with alias " + statusAlias + " not found for flow " + flowInstance.getFlow());

				return sendErrorResponse("Status not found");

			} else if (flowInstance.getStatus().equals(status)) {

				log.info("Flow instance " + flowInstance + " already has the requested status of " + status + ", ignoring request from user " + callback.getUser());

				return sendErrorResponse("Status already set");

			} else {

				log.info("User " + callback.getUser() + " changing status of instance " + flowInstance + " from " + flowInstance.getStatus() + " to " + status);

				Status previousStatus = flowInstance.getStatus();

				flowInstance.setStatus(status);
				flowInstance.setLastStatusChange(TimeUtils.getCurrentTimestamp());

				try {

					this.daoFactory.getFlowInstanceDAO().update(flowInstance);

				} catch (SQLException e) {

					throw new RuntimeException(e);
				}

				FlowInstanceEvent flowInstanceEvent = addFlowInstanceEvent(flowInstance, EventType.STATUS_UPDATED, null, callback.getUser(), null);

				callback.getSystemInterface().getEventHandler().sendEvent(FlowInstance.class, new CRUDEvent<>(CRUDAction.UPDATE, flowInstance), EventTarget.ALL);

				callback.getSystemInterface().getEventHandler().sendEvent(FlowInstance.class, new StatusChangedByManagerEvent(flowInstance, flowInstanceEvent, flowAdminModule.getSiteProfile(flowInstance), previousStatus, callback.getUser()), EventTarget.ALL);

				ImportItemResult result = new ImportItemResult();
				result.setResult("StatusChanged");

				ReturnValue returnValue = new ReturnValue();
				returnValue.setName("EventID");
				returnValue.setContent(flowInstanceEvent.getEventID().toString());

				ReturnValues returnValues = new ReturnValues();
				returnValues.getReturnValue().add(returnValue);

				result.setReturnValues(returnValues);

				return sendResponse(result);
			}

		} catch (RuntimeException e) {

			log.error("Error changing status", e);

			throw e;
		}
	}

	private ImportItemsResponse sendErrorResponse(String message) {

		ImportItemsResponse response = new ImportItemsResponse();

		ImportItemResult result = new ImportItemResult();
		result.setSkippedOrExceptionMessage(message);

		response.getImportItemResult().add(result);

		return response;
	}

	private ImportItemsResponse sendResponse(ImportItemResult... result) {

		ImportItemsResponse response = new ImportItemsResponse();

		response.getImportItemResult().addAll(CollectionUtils.getList(result));

		return response;
	}

	private Map<String, String> getItemProperties(Item item) {

		List<Property> properties = item.getProperty();

		if (properties.isEmpty()) {

			return null;
		}

		Map<String, String> propertyMap = new HashMap<>(properties.size());

		for (Property property : properties) {

			propertyMap.put(property.getName(), property.getContent());
		}

		return propertyMap;
	}

	@Override
	public GetItemsResponse getItems(GetItemsRequest request) {

		throw new RuntimeException("Get items method not implemented");
	}

	@Override
	public PingResponse ping(Object pingRequest) {

		log.info("User " + callback.getUser() + " sent at ping request from " + callback.getRequest().getRemoteAddr());
		
		PingResponse response = new PingResponse();
		response.setMessage("Pong!");

		return response;
	}

	protected void checkDependencies() {

		if (flowAdminModule == null) {

			throw new RuntimeException("Missing required dependency " + FlowAdminModule.class.getSimpleName());
		}
	}

	protected Status getStatus(Flow flow, String statusName) {

		HighLevelQuery<Status> query = new HighLevelQuery<>();

		query.addParameter(statusNameParamFactory.getParameter(statusName));
		query.addParameter(statusFlowParamFactory.getParameter(flow));

		try {
			return daoFactory.getStatusDAO().get(query);

		} catch (SQLException e) {

			throw new RuntimeException(e);
		}
	}

	protected FlowInstanceEvent addFlowInstanceEvent(ImmutableFlowInstance flowInstance, EventType eventType, String details, User user, Timestamp timestamp) {

		FlowInstanceEvent flowInstanceEvent = new FlowInstanceEvent();
		flowInstanceEvent.setFlowInstance((FlowInstance) flowInstance);
		flowInstanceEvent.setEventType(eventType);
		flowInstanceEvent.setDetails(details);
		flowInstanceEvent.setPoster(user);
		flowInstanceEvent.setStatus(flowInstance.getStatus().getName());
		flowInstanceEvent.setStatusDescription(flowInstance.getStatus().getDescription());

		flowInstanceEvent.getAttributeHandler().setAttribute("integration", true);

		if (timestamp == null) {

			flowInstanceEvent.setAdded(TimeUtils.getCurrentTimestamp());

		} else {

			flowInstanceEvent.setAdded(timestamp);
		}

		try {
			daoFactory.getFlowInstanceEventDAO().add(flowInstanceEvent, FLOW_INSTANCE_EVENT_ATTRIBUTE_RELATION_QUERY);

		} catch (Exception e) {

			throw new RuntimeException(e);
		}

		return flowInstanceEvent;
	}

	protected void logDebug(String text) {

		//if (enableDebugLoggning) {

			log.info(text);
		//}

	}

	protected boolean integrationActivated(Flow flow) {

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		if (integrationSettings != null && integrationSettings.isPopulated()) {

			return true;
		}

		return false;
	}

	protected IntegrationSettings getIntegrationSettings(Flow flow) {

		try {

			HighLevelQuery<IntegrationSettings> query = new HighLevelQuery<IntegrationSettings>();

			query.addParameter(flowFamilyIDParamFactory.getParameter(flow.getFlowFamily().getFlowFamilyID()));

			return settingsDAO.get(query);

		} catch (Exception e) {

			throw new RuntimeException(e);
		}

	}

	@Override
	public <InstanceType extends FlowAdminModule> void instanceAdded(Class<FlowAdminModule> key, InstanceType instance) {

		this.flowAdminModule = instance;
	}

	@Override
	public <InstanceType extends FlowAdminModule> void instanceRemoved(Class<FlowAdminModule> key, InstanceType instance) {

		this.flowAdminModule = null;
	}

}
