package com.nordicpeak.flowengine.integration.efecte;

import java.io.IOException;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.TimeUtils;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.FlowInstanceAdminModule;
import com.nordicpeak.flowengine.MessageHandler;
import com.nordicpeak.flowengine.beans.ExternalMessage;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.ManagersChangedEvent;
import com.nordicpeak.flowengine.events.StatusChangedByManagerEvent;
import com.nordicpeak.flowengine.integration.efecte.beans.IntegrationSettings;


public class EfecteCallbackModule extends AnnotatedForegroundModule {

	private static final RelationQuery MANAGER_UPDATED_RELATION_QUERY = new RelationQuery(FlowInstance.MANAGERS_RELATION);
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name="Log XML", description="Controls if the received XML is logged")
	private boolean logXML = false;
	
	@InstanceManagerDependency(required = true)
	protected FlowAdminModule flowAdminModule;
	
	@InstanceManagerDependency(required = true)
	protected FlowInstanceAdminModule flowInstanceAdminModule;
	
	@InstanceManagerDependency(required = true)
	protected MessageHandler messageHandler;

	protected FlowEngineDAOFactory daoFactory;
	
	protected QueryParameterFactory<FlowInstance, Integer> flowInstanceIDParamFactory;
	
	protected QueryParameterFactory<Status, String> statusNameParamFactory;
	protected QueryParameterFactory<Status, Flow> statusFlowParamFactory;
	
	private AnnotatedDAO<IntegrationSettings> settingsDAO;
	private QueryParameterFactory<IntegrationSettings, Integer> flowFamilyIDParamFactory;
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);
		
		this.daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());
		
		flowInstanceIDParamFactory = daoFactory.getFlowInstanceDAO().getParamFactory("flowInstanceID", Integer.class);
		
		statusNameParamFactory = daoFactory.getStatusDAO().getParamFactory("name", String.class);
		statusFlowParamFactory = daoFactory.getStatusDAO().getParamFactory("flow", Flow.class);
		
		settingsDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(IntegrationSettings.class);
		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
	}
	
	@Override
	protected ForegroundModuleResponse processForegroundRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws IOException, SQLException {
		
		log.info("Parsing " + req.getMethod() + " request from " + req.getRemoteAddr() + " by user " + user);
		
		Document doc;
		
		try {
			doc = XMLUtils.parseXML(req.getInputStream(), false, false, "UTF-8");
			
		} catch (SAXException | IOException | ParserConfigurationException e) {
			
			log.error("Error parsing " + req.getMethod() + " request from " + req.getRemoteAddr() + " by user " + user, e);
			sendError(res, "Unable to parse request");
			return null;
		}
		
		if (logXML) {
			
			try {
				log.info("Received XML: " + XMLUtils.toString(doc, systemInterface.getEncoding(), true));
				
			} catch (TransformerFactoryConfigurationError | TransformerException e) {
				
				log.error("Error logging XML", e);
			}
		}
		
		XMLParser xmlParser = new XMLParser(doc);
		
		Integer flowInstanceID = xmlParser.getInteger("substring(/entityset/entity/attribute[@code='externalId']/value,14)");
		
		if (flowInstanceID == null) {
			
			log.warn("Unable to find valid 'externalId' attribute in XML submitted by user " + user);
			sendError(res, "Unable to find valid 'externalId' attribute in submitted XML.");
			return null;
		}
		
		String efecteID = xmlParser.getString("/entityset/entity/attribute[@code='efecte_id']/value");
		
		if (StringUtils.isEmpty(efecteID)) {
			
			log.warn("Unable to find valid 'efecte_id' attribute in XML submitted by user " + user);
			sendError(res, "Unable to find valid 'efecte_id' attribute in submitted XML.");
			return null;
		}
		
		String statusName = xmlParser.getString("/entityset/entity/attribute[@code='status']/value");
		
		if (StringUtils.isEmpty(statusName)) {
			
			log.warn("Unable to find valid 'status' attribute in XML submitted by user " + user);
			sendError(res, "Unable to find valid 'status' attribute in submitted XML.");
			return null;
		}
		
		FlowInstance flowInstance = getFlowInstance(flowInstanceID);
		
		if (flowInstance == null) {
			
			log.warn("Unable to find flow instance " + flowInstanceID + " referenced in XML submitted by user " + user);
			sendError(res, "Unable to find flow instance " + flowInstanceID);
			return null;
		}
		
		String externalID = flowInstance.getAttributeHandler().getString(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE);
		
		if (externalID == null || !externalID.equals(efecteID)) {
			
			log.warn("Submitted efecteID '" + efecteID + "' does not match value " + externalID + " set in " + Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE + " attribute for flow instance " + flowInstance + " referenced in XML submitted by user " + user);
			sendError(res, "Efecte integration is not enabled for flow instance " + flowInstance);
			return null;
		}
		
		Status status = getStatus(flowInstance.getFlow(), statusName);
		
		Status previousStatus = null;
		
		boolean statusUpdated = false;
		
		if (status == null) {
			
			log.warn("Submitted status " + statusName + " does not match any statuses for flow " + flowInstance.getFlow() + ", unable to update flow instance " + flowInstance);
			sendError(res, "Unable to find status " + statusName);
			return null;
			
		} else if (!flowInstance.getStatus().equals(status)) {
			
			previousStatus = flowInstance.getStatus();
			
			flowInstance.setStatus(status);
			flowInstance.setLastStatusChange(TimeUtils.getCurrentTimestamp());
			
			statusUpdated = true;
			
			log.info("Updating status of flow instance " + flowInstance + " from " + previousStatus + " to " + status);
		}
		
		boolean managerUpdated = false;
		
		String detailString = null;
		
		String managerUsername = xmlParser.getString("/entityset/entity/attribute[@code='support_person_ad']/value");
		
		List<User> previousManagers = flowInstance.getManagers();
		
		if (StringUtils.isEmpty(managerUsername)) {
			
			if (flowInstance.getManagers() != null) {
				
				flowInstance.setManagers(null);
				managerUpdated = true;
				
				detailString = flowInstanceAdminModule.getNoManagersSelectedMessage();
			}
			
		} else {
			
			User manager = systemInterface.getUserHandler().getUserByUsername(managerUsername, false, true);
			
			if (manager == null) {
				
				log.error("Submitted managed username " + managerUsername + " does not match any local user, unable to update manager for flow instance " + flowInstance);
				
				if (flowInstance.getManagers() != null) {
					
					flowInstance.setManagers(null);
					managerUpdated = true;
					
					detailString = flowInstanceAdminModule.getNoManagersSelectedMessage();
				}
				
			} else if (flowInstance.getManagers() == null || flowInstance.getManagers().size() > 1 || !flowInstance.getManagers().get(0).equals(manager)) {
				
				flowInstance.setManagers(Collections.singletonList(manager));
				managerUpdated = true;
				
				detailString = manager.getFirstname() + " " + manager.getLastname();
			}
		}
		
		if (statusUpdated || managerUpdated) {
			
			if (managerUpdated) {
				
				log.info("User " + user + " setting managers of instance " + flowInstance + " to " + flowInstance.getManagers());
				
				daoFactory.getFlowInstanceDAO().update(flowInstance, MANAGER_UPDATED_RELATION_QUERY);
				
			} else {
				
				daoFactory.getFlowInstanceDAO().update(flowInstance);
			}
			
			SiteProfile profile = flowAdminModule.getSiteProfile(flowInstance);
			
			if (statusUpdated) {
				
				FlowInstanceEvent flowInstanceEvent = flowAdminModule.getFlowInstanceEventGenerator().addFlowInstanceEvent(flowInstance, EventType.STATUS_UPDATED, null, null);
				flowInstance.getAttributeHandler().setAttribute("integrationSource", "Efecte");
				
				systemInterface.getEventHandler().sendEvent(FlowInstance.class, new CRUDEvent<FlowInstance>(CRUDAction.UPDATE, flowInstance), EventTarget.ALL);
				systemInterface.getEventHandler().sendEvent(FlowInstance.class, new StatusChangedByManagerEvent(flowInstance, flowInstanceEvent, profile, previousStatus, user), EventTarget.ALL);
				
				//TODO check if solved status is enabled, set and a solution is provided
			}
			
			if (managerUpdated) {
				
				FlowInstanceEvent flowInstanceEvent = flowAdminModule.getFlowInstanceEventGenerator().addFlowInstanceEvent(flowInstance, EventType.MANAGERS_UPDATED, detailString, user);
				
				systemInterface.getEventHandler().sendEvent(FlowInstance.class, new CRUDEvent<FlowInstance>(CRUDAction.UPDATE, flowInstance), EventTarget.ALL);
				systemInterface.getEventHandler().sendEvent(FlowInstance.class, new ManagersChangedEvent(flowInstance, flowInstanceEvent, profile, previousManagers, flowInstance.getManagerGroups(), user), EventTarget.ALL);
			}
		}
		
		boolean solutionMessageAdded = false;
		
		IntegrationSettings integrationSettings = getIntegrationSettings(flowInstance);
		
		if (integrationSettings != null && integrationSettings.isEnabled() && integrationSettings.isSolutionSupportEnabled() && integrationSettings.getSolutionStatusName().equalsIgnoreCase(statusName) && statusUpdated) {
			
			String efecteSolutionMessage = xmlParser.getString("/entityset/entity/attribute[@code='resolution']/value");
			
			if (!StringUtils.isEmpty(efecteSolutionMessage)) {
				
				efecteSolutionMessage = efecteSolutionMessage.replace("\n", "\r\n");
				
				ExternalMessage externalMessage = new ExternalMessage();
				
				externalMessage.setAdded(TimeUtils.getCurrentTimestamp());
				externalMessage.setFlowInstance(flowInstance);
				externalMessage.setMessage(integrationSettings.getSolutionMessage().replace("$solutionMessage", efecteSolutionMessage));
				externalMessage.setPostedByManager(true);
				externalMessage.setPoster(null);
				externalMessage.setSystemMessage(true);
				
				messageHandler.add(externalMessage);
			}
		}
		
		res.setStatus(HttpServletResponse.SC_OK);
		
		PrintWriter writer = res.getWriter();
		
		if (statusUpdated) {
			writer.println("Status updated");
			
			if (solutionMessageAdded) {
				
				writer.println("Solutions message added");
			}
		}
		
		if (managerUpdated) {
			writer.println("Manager updated");
		}
		
		if (!statusUpdated && !managerUpdated) {
			
			writer.println("No changes found");
		}
		
		writer.flush();
		writer.close();
		
		return null;
	}

	protected FlowInstance getFlowInstance(Integer flowInstanceID) throws SQLException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<>(FlowInstance.FLOW_RELATION, FlowInstance.STATUS_RELATION, FlowInstance.MANAGERS_RELATION, FlowInstance.ATTRIBUTES_RELATION, Flow.FLOW_FAMILY_RELATION, FlowInstance.OWNERS_RELATION);
		
		query.addParameter(flowInstanceIDParamFactory.getParameter(flowInstanceID));
		
		return daoFactory.getFlowInstanceDAO().get(query);
	}

	protected void sendError(HttpServletResponse res, String message) throws IOException{
		
		res.setStatus(HttpServletResponse.SC_BAD_REQUEST);
		
		PrintWriter writer = res.getWriter();
		
		writer.println(message);
		writer.flush();
		writer.close();
	}

	private Status getStatus(Flow flow, String name) {

		HighLevelQuery<Status> query = new HighLevelQuery<>();

		query.addParameter(statusNameParamFactory.getParameter(name));
		query.addParameter(statusFlowParamFactory.getParameter(flow));

		try {
			return daoFactory.getStatusDAO().get(query);
			
		} catch (SQLException e) {

			throw new RuntimeException(e);
		}
	}
	
	private IntegrationSettings getIntegrationSettings(FlowInstance flowInstance) throws SQLException {

		HighLevelQuery<IntegrationSettings> query = new HighLevelQuery<>();

		query.addParameter(flowFamilyIDParamFactory.getParameter(flowInstance.getFlow().getFlowFamily().getFlowFamilyID()));

		return settingsDAO.get(query);
	}
}
