package com.nordicpeak.flowengine.integration.evolution;

import java.io.File;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

import javax.net.ssl.SSLContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.jaxws.WSUtils;
import se.unlogic.standardutils.ssl.SSLUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.beans.Contact;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.events.StatusChangedByManagerEvent;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.integration.base.BaseIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationSettingsException;
import com.nordicpeak.flowengine.integration.evolution.beans.IntegrationEndpoint;
import com.nordicpeak.flowengine.integration.evolution.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.evolution.beans.QueuedIntegrationRequest;
import com.nordicpeak.flowengine.integration.evolution.proxy.EService;
import com.nordicpeak.flowengine.integration.evolution.proxy.IEService;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.CustomerDetails;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.FileDetails;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.GetNextDiaryNumberRequest;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.GetNextDiaryNumberResponse;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.InsertDocumentRequest;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservice.InsertDocumentResponse;
import com.nordicpeak.flowengine.integration.evolution.proxy.schemas._2004._07.evolution_eservices.DefaultDistributionType;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.notifications.StandardFlowNotificationHandler;
import com.nordicpeak.flowengine.utils.FlowInstanceUtils;
import com.nordicpeak.flowengine.utils.PDFByteAttachment;

public class EvolutionIntegrationProviderModule extends BaseIntegrationProviderModule<QueuedIntegrationRequest, IntegrationSettings, IntegrationEndpoint> {

	private static final String SUPPORTED_TAGS;

	static {
		ArrayList<String> tagList = new ArrayList<>(40);

		tagList.add(StandardFlowNotificationHandler.CONTACT_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.FLOW_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.FLOWINSTANCE_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.STATUS_TAG_SOURCE_FACTORY.getAvailableTags());

		SUPPORTED_TAGS = StringUtils.toCommaSeparatedString(tagList);
	}
	
	private static final AnnotatedRequestPopulator<IntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<>(IntegrationSettings.class);
	private static final String STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME = "evolutionIntegrationStatusRequestEventID";
	private static final RelationQuery UPDATE_FLOW_INSTANCE_RELATIONS = new RelationQuery(FlowInstance.ATTRIBUTES_RELATION);

	@XSLVariable(name = "i18n.QueryName")
	protected String queryNameDisplayName;

	@XSLVariable(name = "i18n.AttachmentTypeID")
	protected String attachmentTypeIDDisplayName;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Error email subject", description = "The subject of emails sent when requests fail")
	@XSLVariable(prefix = "java.")
	private String errorEmailSubject = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Error email message", description = "The message of emails sent when errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String errorEmailMessage = "Not set";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Disable certificate validation", description = "Controls whether certificate should be validated when using SSL or not")
	private boolean disableCertificateValidation = false;

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, this.getClass().getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		endpointDAO = daoFactory.getDAO(IntegrationEndpoint.class);
		queuedRequestDAO = daoFactory.getDAO(QueuedIntegrationRequest.class);

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
		endpointIDParamFactory = endpointDAO.getParamFactory("endpointID", Integer.class);
		lastRetryParamFactory = queuedRequestDAO.getParamFactory("lastRetry", Timestamp.class);
		eventIDParamFactory = queuedRequestDAO.getParamFactory("eventID", Integer.class);
		flowInstanceIDParamFactory = queuedRequestDAO.getParamFactory("flowInstanceID", Integer.class);
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws Exception {

		FlowInstanceManager flowInstanceManager = event.getFlowInstanceManager();
		ImmutableFlowInstance flowInstance = flowInstanceManager.getFlowInstance();
		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !shouldProcessEvent(event)) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = event.getEvent();

		processEvent(flowInstanceEvent, settings);
	}

	private void processEvent(FlowInstanceEvent flowInstanceEvent, IntegrationSettings settings) throws Exception {

		FlowInstance flowInstance = flowInstanceEvent.getFlowInstance();

		try {

			processIntegrationEvent(flowInstance, flowInstanceEvent);

		} catch (IntegrationRequestException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue");

				addQueuedEvent(flowInstanceEvent, flowInstance);

			} else if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationSettingsException e) {

			log.error("Settings error while processing event " + flowInstanceEvent, e);
		}
	}

	@Override
	public void processIntegrationEvent(FlowInstance flowInstance, FlowInstanceEvent event) throws Exception {

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow());

		if (settings == null) {

			throw new IntegrationSettingsException("Settings are null");
		}

		if (!settings.isEnabled()) {

			throw new IntegrationSettingsException("Settings are disabled");
		}

		if (shouldDelayRequest(flowInstance, event, settings)) {

			return;
		}

		EService proxy = new EService();
		IEService service = proxy.getEServiceEndpoint();
		
		String url = settings.getEndpoint().getAddress() + "?apikey=" + settings.getEndpoint().getApiKey();
		
		@SuppressWarnings("rawtypes")
		List<Handler> handlerChain = ((BindingProvider)service).getBinding().getHandlerChain();
		handlerChain.add(new WSAddressHandler(url));
		((BindingProvider) service).getBinding().setHandlerChain(handlerChain);
		
		if(disableCertificateValidation) {
			
			SSLContext sslContext = SSLContext.getInstance("SSL");
			sslContext.init(null, SSLUtils.INSECURE_TRUST_MANGERS, new java.security.SecureRandom());
			WSUtils.setSSLSocketFactory((BindingProvider)service, sslContext.getSocketFactory());
		}
		
		WSUtils.setEndpoint((BindingProvider)service, url);
		WSUtils.setTimeouts((BindingProvider)service, connectionTimeout, requestTimeout);

		TagReplacer tagReplacer = new TagReplacer();

		tagReplacer.addTagSource(StandardFlowNotificationHandler.FLOWINSTANCE_TAG_SOURCE_FACTORY.getTagSource(flowInstance));
		tagReplacer.addTagSource(StandardFlowNotificationHandler.FLOW_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getFlow()));
		tagReplacer.addTagSource(StandardFlowNotificationHandler.STATUS_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getStatus()));

		Contact posterContact = FlowInstanceUtils.getPosterContact(flowInstance);

		if (posterContact != null) {
			tagReplacer.addTagSource(StandardFlowNotificationHandler.CONTACT_TAG_SOURCE_FACTORY.getTagSource(posterContact));
		}

		File pdfFile = pdfProvider.getPDF(flowInstance.getFlowInstanceID(), event.getEventID());

		List<PDFByteAttachment> pdfAttachments = null;

		if (settings.isAddAttachmentsSeparately()) {

			pdfAttachments = pdfProvider.getPDFAttachments(pdfFile, true);
		}
		
		MutableAttributeHandler attributeHandler = flowInstance.getAttributeHandler();

		String unitCode = replaceTags(settings.getUnitCode(), attributeHandler, tagReplacer);
		
		String description = replaceTags(settings.getDescription(), attributeHandler, tagReplacer);
		String notes = replaceTags(settings.getNotes(), attributeHandler, tagReplacer);
		
		String firstname = replaceTags(settings.getFirstname(), attributeHandler, tagReplacer);
		String lastname = replaceTags(settings.getLastname(), attributeHandler, tagReplacer);
		String address = replaceTags(settings.getAddress(), attributeHandler, tagReplacer);
		String coAddress = replaceTags(settings.getCoAddress(), attributeHandler, tagReplacer);
		String zipCode = replaceTags(settings.getZipCode(), attributeHandler, tagReplacer);
		String region = replaceTags(settings.getRegion(), attributeHandler, tagReplacer);
		String email = replaceTags(settings.getEmail(), attributeHandler, tagReplacer);
		String phone = replaceTags(settings.getPhone(), attributeHandler, tagReplacer);
		String mobilePhone = replaceTags(settings.getMobilePhone(), attributeHandler, tagReplacer);
		
		try {

			log.info("Sending next diary number request for flow instance " + flowInstance);
			
			GetNextDiaryNumberRequest nextDiaryNumberRequest = new GetNextDiaryNumberRequest();
			
			nextDiaryNumberRequest.setUnitCode(unitCode);
			
			GetNextDiaryNumberResponse nextDiaryNumberResponse = service.getNextDiarynumber(nextDiaryNumberRequest);
			
			log.info("Got diary number " + nextDiaryNumberResponse.getDiaryNumber() + " for flow instance " + flowInstance);
			
			InsertDocumentRequest insertDocumentRequest = new InsertDocumentRequest();
			
			insertDocumentRequest.setDiaryNumber(nextDiaryNumberResponse.getDiaryNumber());
			insertDocumentRequest.setUnitCode(unitCode);
			
			insertDocumentRequest.setDescription(description);
			insertDocumentRequest.setNote(notes);
			
			if(!StringUtils.isEmpty(firstname, lastname, address, coAddress, zipCode, region, email, phone, mobilePhone)) {
				
				CustomerDetails customerDetails = new CustomerDetails();
				customerDetails.setFirstName(firstname);
				customerDetails.setLastName(lastname);
				customerDetails.setAdress(address);
				customerDetails.setCoAdress(coAddress);
				customerDetails.setZipCode(zipCode);
				customerDetails.setRegion(region);
				customerDetails.setEmail(email);
				customerDetails.setPhone(phone);
				customerDetails.setPhoneMobile(mobilePhone);
				customerDetails.setDefaultDistributionType(DefaultDistributionType.EMAIL);
				
				insertDocumentRequest.setCustomerDetails(customerDetails);
			}
			
			FileDetails fileDetails = new FileDetails();
			fileDetails.setFileName(pdfFile.getName());
			fileDetails.setExtension(".pdf");
			fileDetails.setBuffer(FileUtils.getRawBytes(pdfFile));
			insertDocumentRequest.setFileDetails(fileDetails);
			
			log.info("Sending insert document request for flow instance " + flowInstance);
			
			InsertDocumentResponse insertDocumentResponse = service.insertDocument(insertDocumentRequest);

			FlowInstanceAttribute attribute = new FlowInstanceAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, nextDiaryNumberResponse.getDiaryNumber());
			attribute.setFlowInstance(flowInstance);

			flowAdminModule.getDAOFactory().getFlowInstanceAttributeDAO().add(attribute);
			
			log.info("Successfully created Evolution case with diary number " + nextDiaryNumberResponse.getDiaryNumber() + " and ID " + insertDocumentResponse.getId() + " for event " + event + " from flow instance " + flowInstance);
			
			
			if(pdfAttachments != null) {
				
				for(PDFByteAttachment attachment : pdfAttachments) {
					
					log.info("Sending insert document request for attachemnt " + attachment.getFilename() + " belonging to flow instance " + flowInstance);

					FileDetails attachmentFileDetails = new FileDetails();
					attachmentFileDetails.setFileName(attachment.getFilename());
					attachmentFileDetails.setExtension(FileUtils.getFileExtension(attachment.getFilename()));
					
					if(attachmentFileDetails.getExtension() != null) {
						
						attachmentFileDetails.setExtension("." + attachmentFileDetails.getExtension());
					}
					
					attachmentFileDetails.setBuffer(attachment.getData());
					
					InsertDocumentRequest attachmentRequest = new InsertDocumentRequest();
					attachmentRequest.setFileDetails(attachmentFileDetails);
					attachmentRequest.setDiaryNumber(nextDiaryNumberResponse.getDiaryNumber());
					attachmentRequest.setUnitCode(unitCode);
					
					InsertDocumentResponse attachmentInsertDocumentResponse = service.insertDocument(attachmentRequest);
					
					log.info("Successfully added attachment " + attachment.getFilename() + " to Evolution case with diary number " + nextDiaryNumberResponse.getDiaryNumber() + " and ID " + attachmentInsertDocumentResponse.getId() + " for event " + event + " from flow instance " + flowInstance);
				}
			}	

		} catch (Exception e) {

			boolean shouldLogAsError = maxRetryCount == 0 && !settings.isSendErrorEmail();
			
			logResponse("Error creating Evolution case for event " + event + " from flow instance " + flowInstance + " using endpoint " + settings.getEndpoint(), e, shouldLogAsError);

			throw new IntegrationRequestException();
		}
	}

	private boolean shouldDelayRequest(FlowInstance flowInstance, FlowInstanceEvent event, IntegrationSettings settings) throws SQLException {

		if (!settings.isCreateCaseOnSpecificStatus() || flowInstance.getStatus().getName().equalsIgnoreCase(settings.getCaseStatusName())) {

			return false;
		}

		log.info("Delaying integration request for event " + event + ", flow instance status " + flowInstance.getStatus() + " is not equal to set status " + settings.getCaseStatusName());

		flowInstance.getAttributeHandler().setAttribute(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME, event.getEventID());

		updateFlowInstance(flowInstance);

		return true;
	}

	@EventListener(channel = FlowInstance.class)
	public void processEvent(StatusChangedByManagerEvent event, EventSource eventSource) throws Exception {

		FlowInstance flowInstance = event.getFlowInstance();
		MutableAttributeHandler attributeHandler = flowInstance.getAttributeHandler();
		Integer eventID = attributeHandler.getInt(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME);

		if (eventID == null) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = flowAdminModule.getFlowInstanceEvent(eventID, getFlowInstanceRelations());

		if (flowInstanceEvent == null) {

			return;
		}

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !settings.isCreateCaseOnSpecificStatus()) {

			return;
		}

		if (!flowInstance.getStatus().getName().equalsIgnoreCase(settings.getCaseStatusName())) {

			return;
		}

		log.info("Processing integration request for event " + flowInstanceEvent + " due to changed status (" + flowInstance.getStatus() + ")");

		processEvent(flowInstanceEvent, settings);

		attributeHandler.removeAttribute(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME);

		updateFlowInstance(flowInstance);
	}

	@WebPublic(alias = "update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));
		}
		
		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			try {

				updateIntegrationSettings(integrationSettings, flow, user, req);

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + getShowFlowModuleAnchorLink());

			} catch (ValidationException e) {

				validationException = e;
			}
		}

		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);
		XMLUtils.append(doc, integrationSettingsElement, "Endpoints", endpointDAO.getAll());

		XMLUtils.appendNewElement(doc, integrationSettingsElement, "SupportedTags", SUPPORTED_TAGS);
		
		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}

	@WebPublic(alias = "statuses")
	public ForegroundModuleResponse getStatuses(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return flowAdminModule.getStatuses(req, res, user, uriParser);
	}

	private void updateIntegrationSettings(IntegrationSettings integrationSettings, Flow flow, User user, HttpServletRequest req) throws SQLException, ValidationException {

		if (integrationSettings == null) {

			integrationSettings = new IntegrationSettings();
			integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
		}

		List<ValidationError> validationErrors = new ArrayList<>();

		POPULATOR.populate(integrationSettings, req);

		IntegrationEndpoint endpoint = getEndpoint(req);

		if (endpoint == null) {

			validationErrors.add(new ValidationError("NoEndpointSelected"));

		} else {

			integrationSettings.setEndpoint(endpoint);
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		log.info("User " + user + " updating integration settings for flow " + flow);

		if (integrationSettings.isPopulated()) {

			settingsDAO.addOrUpdate(integrationSettings, null);

		} else {

			settingsDAO.delete(integrationSettings);
		}
	}

	@Override
	protected QueuedIntegrationRequest getNewQueuedIntegrationRequest() {

		return new QueuedIntegrationRequest();
	}

	@Override
	protected void handleMaxRetryCountExceeded(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler, IntegrationRequestException lastRequestException) throws SQLException {

		super.handleMaxRetryCountExceeded(flowInstance, event, queuedRequest, transactionHandler, lastRequestException);

		IntegrationSettings settings;

		if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

			if (settings.isSendErrorEmail()) {

				log.info("Sending error emails due to max retries exceeded");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));

			} else {

				log.error(queuedRequest + " hit its max retry count and error email is disabled");
			}
		}
	}

	@Override
	protected String getShowFlowModuleAnchorLink() {

		return "#evolutionintegrationsettings";
	}

	private void updateFlowInstance(FlowInstance flowInstance) throws SQLException {

		flowAdminModule.getDAOFactory().getFlowInstanceDAO().update(flowInstance, UPDATE_FLOW_INSTANCE_RELATIONS);
	}
}
