package com.nordicpeak.flowengine.externalmessageoffer.views;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Supplier;

import javax.servlet.http.HttpServletRequest;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.standardutils.bool.BooleanUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.populators.BooleanPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.externalmessageoffer.ExternalMessageOfferProviderModule;
import com.nordicpeak.flowengine.externalmessageoffer.beans.ExternalMessageOfferSettings;
import com.nordicpeak.flowengine.interfaces.StatusFormExtensionProvider;

public class ExternalMessageOfferSettingsFormProvider implements StatusFormExtensionProvider {

	private static final AnnotatedRequestPopulator<ExternalMessageOfferSettings> POPULATOR = new AnnotatedRequestPopulator<>(ExternalMessageOfferSettings.class);

	private static final Logger log = Logger.getLogger(ExternalMessageOfferSettingsFormProvider.class);

	private final ExternalMessageOfferProviderModule callback;

	public ExternalMessageOfferSettingsFormProvider(ExternalMessageOfferProviderModule callback) {

		this.callback = callback;
	}

	@Override
	public String getProviderID() {

		return callback.getProviderID();
	}

	@Override
	public ViewFragment getViewFragment(Status status, Flow flow, HttpServletRequest req, User user, URIParser uriParser, ValidationException validationException) throws SQLException, TransformerException {

		if (flow == null) {

			return null;
		}

		Document doc = callback.createDocument(req, uriParser, user);
		Element showElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "UpdateOfferSettings");

		if (validationException != null) {

			showElement.appendChild(RequestUtils.getRequestParameters(req, doc));
			XMLUtils.append(doc, showElement, validationException);
		}

		List<Status> availableStatuses = callback.getStatuses(flow);

		XMLUtils.append(doc, showElement, "Statuses", availableStatuses);

		if (status != null) {

			ExternalMessageOfferSettings settings = callback.getSettingsDAO().getByStatusID(status.getStatusID());

			if (settings != null) {

				setStatusNames(settings, availableStatuses);

				XMLUtils.append(doc, showElement, settings);
			}
		}

		return callback.getViewFragmentTransformer().createViewFragment(doc);
	}

	private void setStatusNames(ExternalMessageOfferSettings settings, List<Status> availableStatuses) throws SQLException {

		setStatusName(settings::getExpiredStatusID, settings::setExpiredStatusName, settings::setExpiredStatusID, availableStatuses);
		setStatusName(settings::getAcceptedStatusID, settings::setAcceptedStatusName, settings::setAcceptedStatusID, availableStatuses);
		setStatusName(settings::getDeclinedStatusID, settings::setDeclinedStatusName, settings::setDeclinedStatusID, availableStatuses);
	}

	private void setStatusName(Supplier<Integer> statusIDGetter, Consumer<String> statusNameSetter, Consumer<Integer> statusIDSetter, List<Status> availableStatuses) throws SQLException {

		if (statusIDGetter.get() == null) {

			return;
		}

		for (Status status : availableStatuses) {

			if (status.getStatusID().equals(statusIDGetter.get())) {

				statusNameSetter.accept(status.getName());
				return;
			}
		}

		statusIDSetter.accept(null);
	}

	@Override
	public void validateRequest(HttpServletRequest req, User user, URIParser uriParser) throws ValidationException {

		if (!BooleanUtils.toBoolean(req.getParameter("offerEnabled"))) {

			return;
		}

		List<ValidationError> validationErrors = new ArrayList<>();

		ExternalMessageOfferSettings settings = null;

		try {

			settings = POPULATOR.populate(req);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (settings != null && settings.isEnabled()) {

			ValidationUtils.validateParameter("addExternalMessage", req, true, BooleanPopulator.getPopulator(), validationErrors);
		}

		if (!CollectionUtils.isEmpty(validationErrors)) {

			throw new ValidationException(validationErrors);
		}
	}

	@Override
	public void processRequest(Status status, HttpServletRequest req, User user, URIParser uriParser) throws SQLException, ValidationException {

		if (status == null) {

			return;
		}

		boolean enabled = BooleanUtils.toBoolean(req.getParameter("offerEnabled"));
		ExternalMessageOfferSettings settings = callback.getSettingsDAO().getByStatusID(status.getStatusID());

		if (settings == null && !enabled) {

			return;
		}

		if (settings == null) {

			settings = new ExternalMessageOfferSettings();
			settings.setStatusID(status.getStatusID());
		}

		if (enabled) {

			settings = POPULATOR.populate(settings, req);

		} else {

			settings.setEnabled(false);
		}

		log.info("User " + user + " updating offer settings for settings " + status);

		callback.getSettingsDAO().addOrUpdate(settings, null);
	}

}
