package com.nordicpeak.flowengine.externalmessageoffer.views;

import java.io.IOException;
import java.sql.SQLException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.ExternalMessage;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.externalmessageoffer.ExternalMessageOfferProviderModule;
import com.nordicpeak.flowengine.externalmessageoffer.beans.ExternalMessageOffer;
import com.nordicpeak.flowengine.externalmessageoffer.enums.ExternalMessageOfferStatus;
import com.nordicpeak.flowengine.interfaces.ExternalMessageExtensionProvider;
import com.nordicpeak.flowengine.interfaces.FlowInstanceOverviewExtensionProvider;

public class ExternalMessageOfferShowProvider implements ExternalMessageExtensionProvider, FlowInstanceOverviewExtensionProvider {

	private static final Logger log = Logger.getLogger(ExternalMessageOfferShowProvider.class);

	private final ExternalMessageOfferProviderModule callback;

	public ExternalMessageOfferShowProvider(ExternalMessageOfferProviderModule callback) {

		this.callback = callback;
	}

	@Override
	public ViewFragment getViewFragment(ExternalMessage message, FlowInstance flowInstance, HttpServletRequest req, User user, URIParser uriParser, boolean manager) throws SQLException, TransformerException {

		if (message == null || callback.getMessageIDCache() == null || !callback.getMessageIDCache().contains(message.getMessageID())) {

			return null;
		}

		ExternalMessageOffer offer = callback.getOfferDAO().getByMessageID(message.getMessageID(), ExternalMessageOffer.FIELD_SETTINGS);

		if (offer == null) {

			return null;
		}

		if (offer.getStatus() == ExternalMessageOfferStatus.ACTIVE) {

			ExternalMessageOfferStatus currentOfferStatus = callback.calculateOfferStatus(offer, flowInstance);

			if (currentOfferStatus == null) {

				return null;
			}

			offer.setStatus(currentOfferStatus);
		}

		Document doc = callback.createDocument(req, uriParser, user);

		Element showElement = appendShowElement(offer, doc);

		XMLUtils.appendNewElement(doc, showElement, "OverviewExtensionURL", RequestUtils.getFullContextPathURL(req) + callback.getUserFlowInstanceModule().getOverviewExtensionRequestMethodAlias(flowInstance, getOverviewExtensionProviderID()));
		XMLUtils.appendNewElement(doc, showElement, "Manager", manager);
		XMLUtils.append(doc, showElement, offer);

		return callback.getViewFragmentTransformer().createViewFragment(doc);
	}

	private Element appendShowElement(ExternalMessageOffer offer, Document doc) {

		String elementName;

		switch (offer.getStatus()) {

			case ACCEPTED:
				elementName = "ShowAcceptedOffer";
				break;

			case DECLINED:
				elementName = "ShowDeclinedOffer";
				break;

			case ACTIVE:
				elementName = "ShowActiveOffer";
				break;

			case WITHDRAWN:
				elementName = "ShowWithdrawnOffer";
				break;

			case EXPIRED:
				elementName = "ShowExpiredOffer";
				break;

			default:
				throw new UnsupportedOperationException();
		}

		return XMLUtils.appendNewElement(doc, doc.getDocumentElement(), elementName);
	}

	@Override
	public ForegroundModuleResponse processOverviewExtensionRequest(FlowInstance flowInstance, HttpServletRequest req, HttpServletResponse res, URIParser uriParser, User user) throws Exception {

		if(!HTTPUtils.isPost(req)) {
			
			throw new AccessDeniedException("Offer accept/decline requests using method " + req.getMethod() + " are not allowed.");
		}		
		
		String method = uriParser.get(5);
		Integer offerID = NumberUtils.toInt(uriParser.get(6));

		if (method == null || offerID == null) {

			return redirect(flowInstance, req, res);
		}

		ExternalMessageOfferStatus targetStatus = ExternalMessageOfferStatus.fromMethod(method);
		ExternalMessageOffer offer = callback.getOfferDAO().get(offerID, ExternalMessageOffer.FIELD_SETTINGS);

		if (targetStatus == null || offer == null) {

			return redirect(flowInstance, req, res);
		}

		ExternalMessageOfferStatus currentStatus = callback.calculateOfferStatus(offer, flowInstance);

		if (currentStatus != ExternalMessageOfferStatus.ACTIVE) {

			log.warn("User " + user + " tried to " + method + " offer " + offer + " which is not active");
			return redirect(flowInstance, req, res);
		}

		log.info("User " + user + " changing status of offer " + offer + " to " + targetStatus);

		offer.setStatus(targetStatus);

		callback.getOfferDAO().update(offer);

		callback.handleOfferTermination(offer, flowInstance, user);

		return redirect(flowInstance, req, res);
	}

	private ForegroundModuleResponse redirect(FlowInstance flowInstance, HttpServletRequest req, HttpServletResponse res) throws IOException {

		callback.getUserFlowInstanceModule().redirectToMethod(req, res, "/messages/" + flowInstance.getFlow().getFlowID() + "/" + flowInstance.getFlowInstanceID());
		return null;
	}

	@Override
	public String getOverviewExtensionProviderID() {

		return callback.getProviderID();
	}

	@Override
	public ExtensionLink getOverviewTabHeaderExtensionLink(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws Exception {

		return null;
	}

	@Override
	public ViewFragment getOverviewTabContentsViewFragment(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws Exception {

		return null;
	}

}
