package com.nordicpeak.flowengine.queries.fixedtablequery;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.beans.QueryDescriptor;
import com.nordicpeak.flowengine.beans.Step;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableQueryDescriptor;
import com.nordicpeak.flowengine.interfaces.MutableQueryInstanceDescriptor;
import com.nordicpeak.flowengine.queries.basetablequery.TableQueryProviderModule;

public class FixedTableQueryProviderModule extends TableQueryProviderModule<DropDownAlternative, FixedTableColumn, FixedTableQuery, FixedTableRow, FixedTableCellValue, FixedTableQueryInstance, FixedTableQueryProviderModule> {

	@XSLVariable(prefix = "java.")
	protected String eventRowAddedMessage = "This variable should be set by your stylesheet";

	@XSLVariable(prefix = "java.")
	protected String eventRowUpdatedMessage = "This variable should be set by your stylesheet";

	@XSLVariable(prefix = "java.")
	protected String eventRowDeletedMessage = "This variable should be set by your stylesheet";

	private FixedTableRowCRUD rowCRUD;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, FixedTableQueryProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		queryDAO = daoFactory.getDAO(FixedTableQuery.class);
		queryInstanceDAO = daoFactory.getDAO(FixedTableQueryInstance.class);
		columnDAO = daoFactory.getDAO(FixedTableColumn.class);

		AnnotatedDAOWrapper<FixedTableColumn, Integer> columnDAOWrapper = columnDAO.getWrapper(Integer.class);
		columnDAOWrapper.addRelations(FixedTableColumn.ROWSUM_ADD_COLUMNS_RELATION, FixedTableColumn.ROWSUM_SUBTRACT_COLUMNS_RELATION);
		columnDAOWrapper.setUseRelationsOnAdd(true);
		columnDAOWrapper.setUseRelationsOnUpdate(true);

		queryCRUD = new FixedTableQueryCRUD(queryDAO.getWrapper(Integer.class), new AnnotatedRequestPopulator<>(FixedTableQuery.class), "FixedTableQuery", "query", null, this);
		columnCRUD = new FixedTableColumnCRUD(columnDAOWrapper, new AnnotatedRequestPopulator<>(FixedTableColumn.class), "FixedTableColumn", "fixedtablecolumn", null, this);
		rowCRUD = new FixedTableRowCRUD(daoFactory.getDAO(FixedTableRow.class).getWrapper(Integer.class), new AnnotatedRequestPopulator<>(FixedTableRow.class), "FixedTableRow", "fixedtablerow", null, this);

		queryIDParamFactory = queryDAO.getParamFactory("queryID", Integer.class);
		queryInstanceIDParamFactory = queryInstanceDAO.getParamFactory("queryInstanceID", Integer.class);
		columnQueryParamFactory = columnDAO.getParamFactory("query", FixedTableQuery.class);
		columnListNameParamFactory = columnDAO.getParamFactory("listName", String.class);

		FlowEngineDAOFactory flowDAOFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		queryDescriptorDAOWrapper = flowDAOFactory.getQueryDescriptorDAO().getWrapper(Integer.class);
		queryDescriptorDAOWrapper.addRelations(QueryDescriptor.STEP_RELATION, Step.FLOW_RELATION, Flow.FLOW_TYPE_RELATION, Flow.CATEGORY_RELATION);
		queryDescriptorDAOWrapper.setUseRelationsOnGet(true);

	}

	@Override
	public FixedTableQueryInstance getQueryInstance(MutableQueryInstanceDescriptor descriptor, String instanceManagerID, HttpServletRequest req, User user, User poster, InstanceMetadata instanceMetadata) throws SQLException {

		FixedTableQueryInstance queryInstance = super.getQueryInstance(descriptor, instanceManagerID, req, user, poster, instanceMetadata);

		if (queryInstance != null) {

			calculateQueryInstanceRowSums(queryInstance, queryInstance.getRows());
			calculateQueryInstanceSummary(queryInstance, queryInstance.getQuery().getRows());
		}

		return queryInstance;
	}

	@Override
	public void populate(FixedTableQueryInstance queryInstance, HttpServletRequest req, User user, User poster, boolean allowPartialPopulation, MutableAttributeHandler attributeHandler, InstanceRequestMetadata requestMetadata) throws ValidationException {

		List<FixedTableColumn> columns = queryInstance.getQuery().getColumns();
		List<FixedTableRow> rows = queryInstance.getQuery().getRows();

		if (CollectionUtils.isEmpty(columns) || CollectionUtils.isEmpty(rows)) {

			//If the parent query doesn't have any columns or rows then there is no population to do
			queryInstance.reset(attributeHandler);
			return;
		}

		List<ValidationError> validationErrors = new ArrayList<>(columns.size() * rows.size());

		List<FixedTableCellValue> cellValues = new ArrayList<>(columns.size() * rows.size());

		for (FixedTableColumn column : columns) {

			for (FixedTableRow row : rows) {

				String cellID = "cell_" + row.getRowIdentifier() + "_" + column.getColumnID().toString();

				populateValue(req, cellValues, column, row, queryInstance, cellID, allowPartialPopulation, validationErrors);
			}
		}

		if (!validationErrors.isEmpty()) {

			resetQueryInstanceSummary(queryInstance);

			throw new ValidationException(validationErrors);
		}

		if (cellValues.isEmpty()) {

			queryInstance.setValues(null);
			queryInstance.getQueryInstanceDescriptor().setPopulated(false);
			resetQueryInstanceRowSums(queryInstance);
			resetQueryInstanceSummary(queryInstance);

		} else {

			queryInstance.setValues(cellValues);
			queryInstance.getQueryInstanceDescriptor().setPopulated(true);
			calculateQueryInstanceRowSums(queryInstance, queryInstance.getRows());
			calculateQueryInstanceSummary(queryInstance, rows);
		}

	}

	@WebPublic(alias = "addrow")
	public ForegroundModuleResponse addRow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		FixedTableQuery query = this.queryCRUD.getRequestedBean(req, res, user, uriParser, null);

		if (query != null) {

			this.checkUpdateQueryAccess(user, query);

			req.setAttribute("FixedTableQuery", query);

			return rowCRUD.add(req, res, user, uriParser);

		}

		throw new URINotFoundException(uriParser);
	}

	@WebPublic(alias = "updaterow")
	public ForegroundModuleResponse updateRow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return rowCRUD.update(req, res, user, uriParser);
	}

	@WebPublic(alias = "deleterow")
	public ForegroundModuleResponse deleteRow(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return rowCRUD.delete(req, res, user, uriParser);
	}

	@WebPublic(alias = "sortrows")
	public ForegroundModuleResponse sortRows(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return ((FixedTableQueryCRUD) queryCRUD).sortRows(req, res, user, uriParser);
	}

	@Override
	public void copyQuery(MutableQueryDescriptor sourceQueryDescriptor, MutableQueryDescriptor copyQueryDescriptor, TransactionHandler transactionHandler, Map<Integer, ImmutableStatus> statusConversionMap) throws SQLException {

		FixedTableQuery query = getQuery(sourceQueryDescriptor.getQueryID(), transactionHandler);

		if (query.getRows() != null) {

			for (FixedTableRow row : query.getRows()) {

				row.setRowID(null);
			}
		}

		super.copyQuery(query, sourceQueryDescriptor, copyQueryDescriptor, transactionHandler, statusConversionMap);
	}

	@Override
	protected Field[] getQueryInstanceRelations() {

		return new Field[] { FixedTableQueryInstance.VALUES_RELATION, FixedTableCellValue.COLUMN_RELATION, FixedTableCellValue.ROW_RELATION, FixedTableQueryInstance.QUERY_RELATION, FixedTableQuery.COLUMNS_RELATION, FixedTableQuery.ROWS_RELATION };
	}

	@Override
	protected RelationQuery getSaveQueryInstanceRelationQuery() {

		return new RelationQuery(FixedTableQueryInstance.VALUES_RELATION);
	}

	@Override
	protected Field[] getTableQueryRelations() {

		return new Field[] { FixedTableQuery.COLUMNS_RELATION, FixedTableQuery.ROWS_RELATION, FixedTableColumn.ROWSUM_ADD_COLUMNS_RELATION, FixedTableColumn.ROWSUM_SUBTRACT_COLUMNS_RELATION };
	}

	@Override
	protected FixedTableQueryInstance getNewQueryInstance() {

		return new FixedTableQueryInstance();
	}

	@Override
	protected FixedTableQuery getNewQuery() {

		return new FixedTableQuery();
	}

	@Override
	protected FixedTableCellValue getNewCellValue(FixedTableRow row, FixedTableColumn column, String value) {

		return new FixedTableCellValue(row, column, value);
	}

	@Override
	protected FixedTableCellValue getNewCellValue(FixedTableRow row, FixedTableColumn column, String value, String displayValue, String xmlValue) {

		return new FixedTableCellValue(row, column, value, displayValue, xmlValue);
	}

	@Override
	protected FixedTableColumn getNewTableColumn() {

		return new FixedTableColumn();
	}

	@Override
	protected Class<FixedTableQueryInstance> getQueryInstanceClass() {

		return FixedTableQueryInstance.class;
	}

	@Override
	protected Field getColumnQueryRelation() {

		return FixedTableColumn.QUERY_RELATION;
	}

	public String getEventRowAddedMessage() {

		return eventRowAddedMessage;
	}

	public String getEventRowUpdatedMessage() {

		return eventRowUpdatedMessage;
	}

	public String getEventRowDeletedMessage() {

		return eventRowDeletedMessage;
	}
}
