package com.nordicpeak.flowengine.flowinstanceadmin2;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.enums.SystemStatus;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.time.TimeUtils;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.events.ExpiredFlowInstanceNotificationEvent;

public class ExpiredFlowInstanceNotificationChecker implements Runnable {

	private final Logger log = Logger.getLogger(this.getClass());

	private final FlowInstanceAdmin2Module callback;

	private boolean running;

	public ExpiredFlowInstanceNotificationChecker(FlowInstanceAdmin2Module callback) {

		this.callback = callback;
	}

	@Override
	public void run() {

		if (callback.getSystemInterface().getSystemStatus() == SystemStatus.STARTED && !running) {

			try {

				running = true;

				try {

					checkForExpiredFlowInstances();

				} catch (Exception e) {

					log.error("Error checking for expired flow instances", e);
				}

			} finally {

				running = false;
			}
		}
	}

	private void checkForExpiredFlowInstances() throws SQLException {

		Collection<FlowFamily> flowFamilies = callback.getFlowAdminModule().getCachedFlowFamilies();

		if (CollectionUtils.isEmpty(flowFamilies)) {

			return;
		}

		Timestamp now = TimeUtils.getCurrentTimestamp();
		int expiredCount = 0;

		for (FlowFamily flowFamily : flowFamilies) {

			if (!shouldCheckForExpiration(flowFamily)) {

				continue;
			}

			List<FlowInstance> flowInstances = callback.getPrioritizedFlowInstances(flowFamily.getFlowFamilyID());

			if (flowInstances == null) {

				continue;
			}

			List<FlowInstance> expiredFlowInstances = CollectionUtils.filter(flowInstances, this::flowInstanceHasExpired);

			if (expiredFlowInstances == null) {

				continue;
			}

			expiredCount += expiredFlowInstances.size();

			expiredFlowInstances.forEach(flowInstance -> flowInstance.setLastExpiredNotificationSent(now));

			callback.getDAOFactory().getFlowInstanceDAO().update(expiredFlowInstances, null);

			List<Integer> expiredFlowInstanceIDs = CollectionUtils.map(expiredFlowInstances, FlowInstance::getFlowInstanceID);

			Map<Integer, SiteProfile> flowInstanceIDToSiteProfileMap = getFlowInstanceIDToSiteProfileMap(expiredFlowInstances);

			callback.getEventHandler().sendEvent(Flow.class, new ExpiredFlowInstanceNotificationEvent(flowFamily, expiredFlowInstanceIDs, flowInstanceIDToSiteProfileMap), EventTarget.ALL);
		}

		if (expiredCount > 0) {

			log.info("Found " + expiredCount + " expired flow instance(s)");
		}
	}

	private boolean shouldCheckForExpiration(FlowFamily flowFamily) throws SQLException {

		if (!anyStatusHasManagingTimeEnabled(flowFamily)) {

			return false;
		}

		if (!callback.getFlowNotificationHandler().isAnyFlowInstanceExpiredNotificationsEnabled(flowFamily)) {

			return false;
		}

		return true;
	}

	private boolean anyStatusHasManagingTimeEnabled(FlowFamily flowFamily) {

		List<Flow> flows = callback.getFlowAdminModule().getFlowVersions(flowFamily);

		if (flows != null) {

			for (Flow flow : flows) {

				if (flow.getStatuses() != null) {

					for (Status status : flow.getStatuses()) {

						if (status.getManagingTime() != null) {

							return true;
						}
					}
				}
			}
		}

		return false;
	}

	private boolean flowInstanceHasExpired(FlowInstance flowInstance) {

		return flowInstance.getLastExpiredNotificationSent() == null || flowInstance.getLastExpiredNotificationSent().before(flowInstance.getLastStatusChange());
	}

	private Map<Integer, SiteProfile> getFlowInstanceIDToSiteProfileMap(List<FlowInstance> flowInstances) {

		Map<Integer, SiteProfile> flowInstanceIDToSiteProfileMap = new HashMap<>();

		for (FlowInstance flowInstance : flowInstances) {

			SiteProfile siteProfile = callback.getFlowAdminModule().getSiteProfile(flowInstance);

			if (siteProfile != null) {

				flowInstanceIDToSiteProfileMap.put(flowInstance.getFlowInstanceID(), siteProfile);
			}
		}

		return flowInstanceIDToSiteProfileMap;
	}

}
