package com.nordicpeak.flowengine.tags.httpsharing;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.PasswordSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.base64.Base64;
import se.unlogic.standardutils.bool.BooleanUtils;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.tags.TextTag;
import com.nordicpeak.flowengine.tags.TextTagAdminModule;


public class HTTPTagReceiverModule extends AnnotatedForegroundModule {

	@ModuleSetting
	@TextFieldSettingDescriptor(name="Username", description="Username for tag upload")
	private String username;
	
	@ModuleSetting
	@PasswordSettingDescriptor(name="Password", description="Password for tag upload")
	private String password;
	
	@InstanceManagerDependency(required=true)
	private TextTagAdminModule tagAdminModule;
	
	@Override
	protected ForegroundModuleResponse processForegroundRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		checkAccess(req);
		
		if(!req.getMethod().equalsIgnoreCase("POST")) {
			
			log.warn("User " + user + " submitted tag request with method " + req.getMethod());
			
			throw new URINotFoundException(uriParser);
		}
		
		XMLParser xmlParser;
		
		try {
			String xml = req.getParameter("xml");
			
			Document doc = XMLUtils.parseXML(xml, systemInterface.getEncoding(), false, false);
			
			xmlParser = new XMLParser(doc);
			
		} catch (Exception e) {

			log.error("Error parsing tag XML posted by user " + user + " from " + req.getRemoteAddr(), e);
			sendError(res, "Unable to parse XML");
			return null;
		}
		
		List<XMLParser> tagElements = xmlParser.getNodes("/Tags/TextTag", true);
		
		if(tagElements.isEmpty()) {
			
			sendError(res, "No tags found");
			return null;
		}
		
		List<TextTag> tags = new ArrayList<TextTag>(tagElements.size()); 
		
		try {
			for(XMLParser tagElement : tagElements) {
				
				TextTag tag = new TextTag();
				
				tag.populate(tagElement);
				
				tags.add(tag);
			}
			
		} catch (Exception e) {
			
			log.error("Error populating tags from XML posted by user " + user + " from " + req.getRemoteAddr(), e);
			sendError(res, "Unable to parse XML");
			return null;
		}
		
		
		boolean overwrite = BooleanUtils.toBoolean(req.getParameter("overwrite"));

		log.info("Received tags " + tags + " with overwrite set to " + overwrite);
		
		this.tagAdminModule.importTags(tags, overwrite);
		
		sendResponse(res, "Tags imported.");
		
		return null;
	}

	private void checkAccess(HttpServletRequest req) throws AccessDeniedException, UnsupportedEncodingException, IOException {

		String authHeader = req.getHeader("Authorization");

		String username = null;
		
		if (authHeader != null) {
			
			StringTokenizer st = new StringTokenizer(authHeader);
			
			if (st.hasMoreTokens()) {
				
				String basic = st.nextToken();
				
				if (basic.equalsIgnoreCase("Basic")) {
					
					String decodedCredentials = new String(Base64.decode(st.nextToken()), "UTF-8");
					
					int index = decodedCredentials.indexOf(":");
					
					if (index != -1) {
						
						username = decodedCredentials.substring(0, index).trim();
						String password = decodedCredentials.substring(index + 1).trim();
						
						if(ObjectUtils.compare(this.username, username) && ObjectUtils.compare(this.password, password)) {
							
							return;
						}
						
					}
				}
			}
			
		}
		
		log.warn("Failed login attempt using username " + username + " from address " + req.getRemoteHost());
		
		throw new AccessDeniedException("Invalid credentials");
	}

	protected void sendResponse(HttpServletResponse res, String message) throws IOException{
		
		res.setStatus(HttpServletResponse.SC_OK);
		
		PrintWriter writer = res.getWriter();
		
		writer.println(message);
		writer.flush();
		writer.close();
	}
	
	protected void sendError(HttpServletResponse res, String message) throws IOException{
		
		res.setStatus(HttpServletResponse.SC_BAD_REQUEST);
		
		PrintWriter writer = res.getWriter();
		
		writer.println(message);
		writer.flush();
		writer.close();
	}
}
