package com.nordicpeak.flowengine.queries.iconmapquery;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.rowset.serial.SerialBlob;
import javax.sql.rowset.serial.SerialException;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadBase.FileSizeLimitExceededException;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.io.FilenameUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.fileuploadutils.MultipartRequest;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.validationerrors.FileSizeLimitExceededValidationError;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.image.ImageUtils;
import se.unlogic.standardutils.io.BinarySizes;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.beans.QueryDescriptor;
import com.nordicpeak.flowengine.queries.basequery.BaseQueryCRUD;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQuery;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.MapIcon;
import com.nordicpeak.flowengine.remotegeometryendpointprovider.RemoteGeometryEndpointDescriptor;

public class IconMapQueryCRUD extends BaseQueryCRUD<IconMapQuery, IconMapQueryProviderModule> {

	AnnotatedDAOWrapper<IconMapQuery, Integer> queryDAO;

	public IconMapQueryCRUD(AnnotatedDAOWrapper<IconMapQuery, Integer> queryDAO, IconMapQueryProviderModule callback) {

		super(IconMapQuery.class, queryDAO, new AnnotatedRequestPopulator<>(IconMapQuery.class), "IconMapQuery", "general map query", null, callback);

		this.queryDAO = queryDAO;
	}

	@Override
	protected void appendUpdateFormData(IconMapQuery bean, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendUpdateFormData(bean, doc, updateTypeElement, user, req, uriParser);

		if (callback.getRemoteGeometryEndpointProvider() != null && callback.getRemoteGeometryEndpointProvider().getEndpoints() != null) {

			Element endpointsElement = XMLUtils.appendNewElement(doc, updateTypeElement, "RemoteGeometryEndpoints");

			for (RemoteGeometryEndpointDescriptor descriptor : callback.getRemoteGeometryEndpointProvider().getEndpoints()) {

				Element endpointElement = XMLUtils.appendNewElement(doc, endpointsElement, "Endpoint");

				XMLUtils.appendNewElement(doc, endpointElement, "Name", descriptor.getName());
				XMLUtils.appendNewElement(doc, endpointElement, "ID", descriptor.getName());

			}

			if (!StringUtils.isEmpty(bean.getGeometryEndpointName())) {

				RemoteGeometryEndpointDescriptor currentEndpoint = callback.getRemoteGeometryEndpointProvider().getEndpoint(bean.getGeometryEndpointName());

				if (currentEndpoint != null) {

					XMLUtils.appendNewElement(doc, updateTypeElement, "CurrentEndpoint", currentEndpoint.getName());

				} else {

					Element providerElement = XMLUtils.appendNewElement(doc, endpointsElement, "Endpoint");

					XMLUtils.appendNewElement(doc, providerElement, "Name", callback.getMissingEndpoint() + bean.getGeometryEndpointName());
					XMLUtils.appendNewElement(doc, providerElement, "ID", bean.getGeometryEndpointName());
				}

			}

		}

		if (callback.getReservationHandler() != null) {

			XMLUtils.appendNewElement(doc, updateTypeElement, "ReservationAvailable", true);
		}
	}

	@Override
	protected IconMapQuery populateFromUpdateRequest(IconMapQuery bean, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {

		List<ValidationError> errors = new ArrayList<>();

		IconMapQuery generalMapQuery = super.populateFromUpdateRequest(bean, req, user, uriParser);

		this.populateQueryDescriptor((QueryDescriptor) generalMapQuery.getQueryDescriptor(), req, errors);

		if (!StringUtils.isEmpty(bean.getGeometryEndpointName()) && bean.isUseSelectableGeometries()) {

			bean.setUseDetailedIconInfo(false);
			bean.setIcons(null);
			bean.setShowIconPUDInfo(false);

			return bean;
		}

		String[] mapIconIDs = req.getParameterValues("mapIconID");

		List<MapIcon> mapIcons = new ArrayList<>();

		if (mapIconIDs != null) {

			for (String mapIconID : mapIconIDs) {

				String name = ValidationUtils.validateParameter("mapIconName_" + mapIconID, req, false, 1, 255, errors);
				Integer minIcons = NumberUtils.toInt(req.getParameter("minIcons_" + mapIconID));
				Integer maxIcons = NumberUtils.toInt(req.getParameter("maxIcons_" + mapIconID));

				String sortOrder = req.getParameter("sortOrder_" + mapIconID);

				if (NumberUtils.isInt(sortOrder)) {

					Integer sortIndex = NumberUtils.toInt(sortOrder);

					if (StringUtils.isEmpty(name)) {

						errors.add(new ValidationError("mapIconName_" + mapIconID, sortIndex + 1 + "", ValidationErrorType.RequiredField));
					}

					if (maxIcons != null && minIcons != null && maxIcons < minIcons) {
						errors.add(new ValidationError("MaxIconsSmallerThanMinIcons", sortIndex + 1 + ""));
					}

					MapIcon mapIcon = new MapIcon();

					mapIcon.setName(name);
					mapIcon.setSortIndex(sortIndex);
					mapIcon.setMinIcons(minIcons);
					mapIcon.setMaxIcons(maxIcons);

					populateIconImage(req, mapIcon, mapIconID, errors);

					if (NumberUtils.isInt(mapIconID)) {

						this.checkForExistingMapIcons(bean.getIcons(), mapIcon, NumberUtils.toInt(mapIconID), errors);
					}

					if (mapIcon.getIconImage() == null) {
						errors.add(new ValidationError("NoMapIcon", mapIcon.getSortIndex() + 1 + ""));
					}

					mapIcons.add(mapIcon);

				}

			}

		}

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

		generalMapQuery.setIcons(mapIcons);

		return generalMapQuery;
	}

	private void populateIconImage(HttpServletRequest req, MapIcon mapIcon, String mapIconID, List<ValidationError> errors) throws SerialException, SQLException {

		if (!(req instanceof MultipartRequest)) {

			return;
		}

		MultipartRequest multipartReq = (MultipartRequest) req;

		FileItem file = multipartReq.getFile("icon_" + mapIconID);

		if (file == null || file.getSize() == 0) {

			return;
		}

		long maxFileSizeInBytes = callback.getMaxFileSize() * BinarySizes.MegaByte;

		if (file.getSize() > maxFileSizeInBytes) {

			errors.add(new FileSizeLimitExceededValidationError(FilenameUtils.getName(file.getName()), file.getSize(), maxFileSizeInBytes));
			return;
		}

		String lowerCasefileName = file.getName().toLowerCase();

		if (!lowerCasefileName.endsWith(".png") && !lowerCasefileName.endsWith(".jpg") && !lowerCasefileName.endsWith(".gif") && !lowerCasefileName.endsWith(".bmp")) {

			errors.add(new ValidationError("InvalidIconFileFormat"));

		} else {

			try {

				BufferedImage image = ImageUtils.getImage(file.get());

				if (image == null) {

					errors.add(new ValidationError("UnableToParseIcon"));

				} else {

					String filename = FilenameUtils.getName(file.getName());

					if (image.getWidth() > callback.getMaxMapIconWidth() || image.getHeight() > callback.getMaxMapIconHeight()) {

						image = ImageUtils.scaleImage(image, callback.getMaxMapIconHeight(), callback.getMaxMapIconWidth(), Image.SCALE_SMOOTH, BufferedImage.TYPE_INT_ARGB);

						ByteArrayOutputStream iconStream = new ByteArrayOutputStream();

						ImageIO.write(image, "png", iconStream);

						mapIcon.setIconImage(new SerialBlob(iconStream.toByteArray()));
						mapIcon.setIconImageFilename(FileUtils.replaceFileExtension(FilenameUtils.getName(filename), "png"));

					} else {

						mapIcon.setIconImage(new SerialBlob(file.get()));
						mapIcon.setIconImageFilename(FilenameUtils.getName(filename));
					}

				}

			} catch (IOException e) {

				errors.add(new ValidationError("UnableToParseIcon"));
			}
		}

	}

	protected void checkForExistingMapIcons(List<MapIcon> currentMapIcons, MapIcon mapIcon, Integer mapIconID, List<ValidationError> errors) {

		if (!CollectionUtils.isEmpty(currentMapIcons)) {

			for (MapIcon queryMapIcon : currentMapIcons) {

				if (queryMapIcon.getMapIconID().equals(mapIconID)) {

					if (mapIcon.getIconImage() == null && queryMapIcon.getIconImage() != null) {

						mapIcon.setIconImageFilename(queryMapIcon.getIconImageFilename());
						mapIcon.setIconImage(queryMapIcon.getIconImage());
					}

					mapIcon.setMapIconID(mapIconID);
					break;

				}

			}

		}

	}

	@Override
	protected ForegroundModuleResponse beanUpdated(IconMapQuery bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		//callback.cacheIconMapQueryConfiguration(bean);

		return super.beanUpdated(bean, req, res, user, uriParser);
	}

	@Override
	protected HttpServletRequest parseRequest(HttpServletRequest req, User user) throws ValidationException, Exception {

		if (!req.getMethod().equalsIgnoreCase("POST")) {

			return super.parseRequest(req, user);
		}

		try {
			return callback.parseMultipartRequest(req, user);

		} catch (FileSizeLimitExceededException e) {

			throw new ValidationException(new ValidationError("FileSizeLimitExceeded"));

		} catch (FileUploadException e) {

			log.warn(e);
			throw new ValidationException(new ValidationError("UnableToParseRequest"));
		}
	}

	@Override
	protected void releaseRequest(HttpServletRequest req, User user) {

		if (req instanceof MultipartRequest) {
			((MultipartRequest) req).deleteFiles();
		}
	}

}
