package com.nordicpeak.flowengine.queries.iconmapquery.requestprocessors;

import java.io.IOException;
import java.sql.SQLException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.interfaces.QueryRequestProcessor;
import com.nordicpeak.flowengine.queries.iconmapquery.IconMapQueryReservationHandler;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQueryInstance;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQueryInstanceReservation;
import com.nordicpeak.flowengine.remotegeometryendpointprovider.RemoteGeometry;

public class GeometryReservationRequestProcessor implements QueryRequestProcessor {

	private final Logger log = Logger.getLogger(this.getClass());

	private IconMapQueryReservationHandler reservationHandler;
	private IconMapQueryInstance queryInstance;
	private InstanceRequestMetadata requestMetadata;
	private User poster;
	private String action;

	public GeometryReservationRequestProcessor(IconMapQueryReservationHandler reservationHandler, IconMapQueryInstance queryInstance, InstanceRequestMetadata requestMetadata, User poster, String action) {

		this.reservationHandler = reservationHandler;
		this.queryInstance = queryInstance;
		this.requestMetadata = requestMetadata;
		this.poster = poster;
		this.action = action;
	}

	@Override
	public void processRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String objectID = req.getParameter("objectid");

		if (objectID == null) {

			throw new URINotFoundException(uriParser);
		}

		RemoteGeometry geometry = CollectionUtils.find(queryInstance.getFetchedGeometries(), g -> g.getObjectID().equals(objectID));

		if (geometry == null) {

			throw new URINotFoundException(uriParser);
		}

		if (isCreateRequest(action)) {

			createReservation(objectID, res);

		} else if (isDeleteRequest(action)) {

			deleteReservation(objectID, res);
		}
	}

	private void createReservation(String objectID, HttpServletResponse res) throws SQLException, IOException {

		try {

			reservationHandler.validateForReservation(queryInstance, poster, requestMetadata);

		} catch (ValidationException e) {

			log.info("User " + poster + " tried to create reservation for object " + objectID + " on queryInstance " + queryInstance + " " + e);
			sendQueryInstanceNotValidForReservationError(res);
			return;
		}

		log.info("User " + poster + " trying to create reservation for object " + objectID + " on queryInstance " + queryInstance);

		String flowInstanceManagerID = requestMetadata.getFlowInstanceManagerID();

		IconMapQueryInstanceReservation reservation = reservationHandler.createReservation(objectID, queryInstance, flowInstanceManagerID);

		if (reservation != null) {

			queryInstance.addReservation(reservation);

			sendSuccessResponse(res);

		} else {

			sendReservationFailedError(res);
		}
	}

	private void deleteReservation(String objectID, HttpServletResponse res) throws Exception {

		try {

			reservationHandler.validateForReservation(queryInstance, poster, requestMetadata);

		} catch (ValidationException e) {

			log.info("User " + poster + " tried to delete reservation for object " + objectID + " on queryInstance " + queryInstance + " " + e);
			sendQueryInstanceNotValidForReservationError(res);
			return;
		}

		log.info("User " + poster + " trying to delete reservation for object " + objectID + " on queryInstance " + queryInstance);

		reservationHandler.deleteReservation(objectID, queryInstance);

		sendSuccessResponse(res);
	}

	@Override
	public void close() throws Exception {

	}

	public static boolean isGeometryReservationRequest(String action) {

		return isCreateRequest(action) || isDeleteRequest(action);
	}

	private static boolean isDeleteRequest(String action) {

		return "deletereservation".equals(action);
	}

	private static boolean isCreateRequest(String action) {

		return "createreservation".equals(action);
	}

	protected void sendSuccessResponse(HttpServletResponse res) throws IOException {

		sendJSONResponse("success", res);
	}

	protected void sendReservationFailedError(HttpServletResponse res) throws IOException {

		sendJSONError("reservationFailed", res);
	}

	protected void sendQueryInstanceNotValidForReservationError(HttpServletResponse res) throws IOException {

		sendJSONError("queryInstanceNotValidForReservation", res);
	}

	protected void sendJSONResponse(String status, HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(1);

		jsonObject.putField("status", status);

		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	public void sendJSONError(String error, HttpServletResponse res) throws IOException {

		res.setStatus(500); //INTERNAL SERVER ERROR

		JsonObject json = new JsonObject();

		json.putField("error", error);

		HTTPUtils.sendReponse(json.toJson(), JsonUtils.getContentType(), res);
	}

}
