package com.nordicpeak.lmsearch.utils;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.Calendar;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.interfaces.providers.OAuthCredentialsProvider;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.lmsearch.beans.LMAccessToken;

import flexjson.JSONDeserializer;

public class LMTokenHandler {

	protected Logger log = Logger.getLogger(this.getClass());

	private static final Charset UTF8 = Charset.forName("UTF-8");

	private static final String SERVICE_URL = "https://api.lantmateriet.se";

	private OAuthCredentialsProvider oAuthCredentialsProvider;

	private final int connectionTimeout;
	private final int requestTimeout;

	private LMAccessToken accessToken;

	public LMTokenHandler(OAuthCredentialsProvider oAuthCredentialsProvider, int connectionTimeout, int requestTimeout) {

		this.oAuthCredentialsProvider = oAuthCredentialsProvider;
		this.connectionTimeout = connectionTimeout;
		this.requestTimeout = requestTimeout;
	}

	public String getAccessToken() {

		if(oAuthCredentialsProvider == null) {
			
			log.warn("No oauth credentials provider is set, unable to request access token");
			
			return null;
		}
		
		if(StringUtils.isEmpty(oAuthCredentialsProvider.getClientID()) || StringUtils.isEmpty(oAuthCredentialsProvider.getClientSecret())) {
			
			log.warn("Consumer key or consumer secret is not set, unable to request access token");
			
			return null;
		}
		
		synchronized (LMTokenHandler.class) {
			
			if (accessToken != null && Calendar.getInstance().after(accessToken.getExpiresDate())) {
				
				log.info("Token " + accessToken.getToken() + " has expired, renewing...");
				
				revokeToken(accessToken.getToken());
				accessToken = renewToken();
				
			} else if (accessToken == null) {
				
				accessToken = renewToken();
			}
		}

		return accessToken != null ? accessToken.getToken() : null;

	}

	private LMAccessToken renewToken() {

		try {

			log.info("Requesting access token from token service");

			SimpleRequest request = new SimpleRequest(SERVICE_URL + "/token");
			request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
			request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);
			request.setUsername(oAuthCredentialsProvider.getClientID());
			request.setPassword(oAuthCredentialsProvider.getClientSecret());

			String body = "grant_type=client_credentials";

			String response = HTTPUtils.sendHTTPPostRequest(request, new ByteArrayInputStream(body.getBytes(UTF8)), UTF8);

			if (!StringUtils.isEmpty(response)) {

				LMAccessToken newAccessToken = new JSONDeserializer<LMAccessToken>().deserialize(response, LMAccessToken.class);

				if (newAccessToken == null) {

					log.error("Unable to deserialize response from token service, got response: " + response);

					return null;
				}

				Calendar expiresDate = Calendar.getInstance();

				if (newAccessToken.getExpiresIn() > Integer.MAX_VALUE) {
					expiresDate.add(Calendar.YEAR, 50);
				} else {
					expiresDate.add(Calendar.SECOND, (newAccessToken.getExpiresIn().intValue() - 3));
				}

				newAccessToken.setExpiresDate(expiresDate);

				log.info("Got access token " + newAccessToken.getToken() + " expiring in " + newAccessToken.getExpiresIn() + " seconds");

				return newAccessToken;
			}

		} catch (Exception e) {

			log.warn("Unable to create token using consumer key " + oAuthCredentialsProvider.getClientID() + " and consumer secret" + oAuthCredentialsProvider.getClientSecret(), e);
		}

		return null;
	}

	private boolean revokeToken(String token) {

		try {

			log.info("Revoking token " + token);

			SimpleRequest request = new SimpleRequest(SERVICE_URL + "/revoke");
			request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
			request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);
			request.setUsername(oAuthCredentialsProvider.getClientID());
			request.setPassword(oAuthCredentialsProvider.getClientSecret());

			String body = "token=" + token;

			HTTPUtils.sendHTTPPostRequest(request, new ByteArrayInputStream(body.getBytes(UTF8)), UTF8);

			return true;

		} catch (UnsupportedEncodingException e) {

		} catch (IOException e) {

			log.warn("Unable to revoke token " + token);
		}

		return false;

	}

}
