package com.nordicpeak.flowengine.paymentlog;
import java.io.Writer;
import java.sql.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.xml.UserAttributeElementableListener;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.hierarchy.foregroundmodules.rest.responsehandlers.XMLResponseHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.xml.XMLGeneratorDocument;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.paymentlog.beans.Account;
import com.nordicpeak.flowengine.paymentlog.beans.ChartOfAccounts;
import com.nordicpeak.flowengine.paymentlog.beans.Transaction;
import com.nordicpeak.flowengine.paymentlog.enums.AccountType;
import com.nordicpeak.flowengine.paymentlog.interfaces.TransactionFileExporter;


public class TransactionAPIModule extends AnnotatedRESTModule {

	private static final UserAttributeElementableListener ELEMENTABLE_LISTENER = new UserAttributeElementableListener();
	
	@InstanceManagerDependency(required=true)
	protected TransactionModule transactionModule; 

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		addResponseHandler(new XMLResponseHandler(sectionInterface.getSystemInterface().getEncoding()));

		super.init(moduleDescriptor, sectionInterface, dataSource);
	}

	@RESTMethod(alias = "account/{accountID}", method = "get")
	public Document getTransactionsByAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "accountID") Integer accountID) throws Throwable {

		Account account = transactionModule.getAccount(accountID);

		if (account == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting transactions for account " + account + " from date " + fromDate + " to date " + toDate);

		List<Transaction> transactions = transactionModule.getTransactions(account, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element transactionsElement = doc.createElement("Transactions");
		doc.appendChild(transactionsElement);

		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addAssignableFieldElementableListener(User.class, ELEMENTABLE_LISTENER);
		
		XMLUtils.append(generatorDocument, transactionsElement, transactions);

		return doc;
	}
	
	@RESTMethod(alias = "chart/{chartID}", method = "get")
	public Document getTransactionsByChartOfAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "chartID") Integer chartID) throws Throwable {

		ChartOfAccounts chart = transactionModule.getChart(chartID, ChartOfAccounts.ACCOUNTS_RELATION);

		if (chart == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);

		Date toDate = getDate(req.getParameter("toDate"), false);

		log.info("User " + user + " requesting transactions for chart " + chart + " from date " + fromDate + " to date " + toDate);

		List<Transaction> transactions = transactionModule.getTransactions(chart, fromDate, toDate);

		Document doc = XMLUtils.createDomDocument();

		Element transactionsElement = doc.createElement("Transactions");
		doc.appendChild(transactionsElement);
		
		XMLGeneratorDocument generatorDocument = new XMLGeneratorDocument(doc);
		generatorDocument.addAssignableFieldElementableListener(User.class, ELEMENTABLE_LISTENER);		

		XMLUtils.append(generatorDocument, transactionsElement, transactions);

		return doc;
	}	

	@RESTMethod(alias = "chart/{chartID}/fileexport", method = "get")
	public Document getTransactionsInFileExporterFormatByChartOfAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "chartID") Integer chartID) throws Throwable {

		TransactionFileExporter fileExporter = transactionModule.getTransactionFileExporter();
		ChartOfAccounts chart = transactionModule.getChartForTransactionFile(chartID);
		
		if (fileExporter == null || chart == null) {

			throw new URINotFoundException(uriParser);
		}

		Date fromDate = getDate(req.getParameter("fromDate"), true);
		Date toDate = getDate(req.getParameter("toDate"), false);
		Date balanceAccountDate = getDate(req.getParameter("balanceAccountDate"), false);

		log.info("User " + user + " requesting transactions in file export format for chart " + chart + " from date " + fromDate + " to date " + toDate);

		if (!CollectionUtils.isEmpty(chart.getAccounts())) {

			for (Account account : chart.getAccounts()) {

				if (account.getAccountType() != AccountType.BALANCE) {

					account.setTransactions(transactionModule.getTransactions(account, fromDate, toDate));
				}
			}
		}
		
		res.setContentType("text/plain");
		
		try (Writer writer = res.getWriter()){

			fileExporter.writeTransactions(chart, writer, user, balanceAccountDate);
		}

		return null;
	}
	
	protected Date getDate(String value, boolean start) {

		if(!StringUtils.isEmpty(value)){
			
			if(value.length() > 10){
				
				return DateUtils.getSQLDate(DateUtils.DATE_TIME_FORMATTER, value);
				
			}else{
				
				Date date = DateUtils.getSQLDate(DateUtils.DATE_FORMATTER, value);
		
				if (date != null) {

					if(start){
						
						DateUtils.setTimeToMidnight(date);

					}else{
					
						DateUtils.setTimeToMaximum(date);
					}
				}
				
				return date;
			}
		}
		
		return null;
	}
}
