package com.nordicpeak.flowengine.paymentlog.cruds;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.dao.CRUDDAO;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.paymentlog.AccountModule;
import com.nordicpeak.flowengine.paymentlog.beans.Account;
import com.nordicpeak.flowengine.paymentlog.beans.AccountNumberPart;
import com.nordicpeak.flowengine.paymentlog.beans.ChartOfAccounts;
import com.nordicpeak.flowengine.paymentlog.enums.AccountType;
import com.nordicpeak.flowengine.paymentlog.populators.AccountNumberPartPopulator;

public class AccountCRUD extends ModularCRUD<Account, Integer, User, AccountModule> {
	
	protected static AccountNumberPartPopulator ALTERNATIVES_POPLATOR = new AccountNumberPartPopulator();
	
	public AccountCRUD(CRUDDAO<Account, Integer> crudDAO, AccountModule callback) {
		super(IntegerBeanIDParser.getInstance(), crudDAO, new AnnotatedRequestPopulator<>(Account.class), "Account", "account", "/accounts", callback);
	}
	
	@Override
	protected void appendBean(Account account, Element targetElement, Document doc, User user) {
		
		Element element = account.toXML(doc);
		
		try {
			if (callback.isAccountInUse(account)) {
				XMLUtils.appendNewElement(doc, element, "inUse", true);
			}
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
		
		targetElement.appendChild(element);
	}
	
	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {
		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);
		
		AccountType type = (AccountType) req.getAttribute("accountType");
		ChartOfAccounts chart = (ChartOfAccounts) req.getAttribute("chart");
		
		addTypeElement.appendChild(chart.toXML(doc));
		XMLUtils.appendNewElement(doc, addTypeElement, "AccountType", type.name());
	}

	@Override
	protected void appendUpdateFormData(Account bean, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {
		super.appendUpdateFormData(bean, doc, updateTypeElement, user, req, uriParser);
	}
	
	@Override
	protected void checkAddAccess(User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {
		super.checkAddAccess(user, req, uriParser);
		
		AccountType type = (AccountType) req.getAttribute("accountType");
		ChartOfAccounts chart = (ChartOfAccounts) req.getAttribute("chart");
		
		if (type == AccountType.BALANCE && callback.getAccounts(chart, type) != null) {
			
			throw new AccessDeniedException("Only 1 balance account is allowed");
		}
	}

	@Override
	protected void checkDeleteAccess(Account account, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {
		super.checkDeleteAccess(account, user, req, uriParser);
		
		//TODO if used by flow family or has transactions deny, maybe allow if archived
		if (callback.isAccountInUse(account)) {
			throw new AccessDeniedException("Account in use");
		}
	}

	@Override
	protected Account populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {
		
		Account account = super.populateFromAddRequest(req, user, uriParser);
		
		AccountType type = (AccountType) req.getAttribute("accountType");
		ChartOfAccounts chart = (ChartOfAccounts) req.getAttribute("chart");
		
		account.setChart(chart);
		account.setAccountType(type);
		
		populate(account, req, user, uriParser);
		
		return account;
	}
	
	@Override
	protected Account populateFromUpdateRequest(Account oldAccount, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {
		
		Account account = super.populateFromUpdateRequest(oldAccount, req, user, uriParser);
		
		populate(account, req, user, uriParser);
		
		return account;
	}
	
	protected void populate(Account account,HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, SQLException, Exception {
		
		ArrayList<ValidationError> validationErrors = new ArrayList<ValidationError>();
		
		if (account.getAccountType() == AccountType.VAT) {
			
			Integer vat = ValidationUtils.validateParameter("vat", req, true, PositiveStringIntegerPopulator.getPopulator(), validationErrors);
			
			if (vat != null && vat >= 100) {
				
				validationErrors.add(new ValidationError("vat", ValidationErrorType.TooLong));
			}
			
			account.setVat(vat);
			
		} else {
			
			account.setVat(null);
		}
		
		List<AccountNumberPart> accountNumberParts = ALTERNATIVES_POPLATOR.populate(account.getChart().getAccountNumberPartDefinitions(), req, validationErrors);
		
		if (!validationErrors.isEmpty()) {
			
			throw new ValidationException(validationErrors);
		}
		
		account.setAccountNumberParts(accountNumberParts);
	}
	
	@Override
	protected ForegroundModuleResponse filteredBeanAdded(Account account, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showchart/" + account.getChart().getChartID());
		return null;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(Account account, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showchart/" + account.getChart().getChartID());
		return null;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(Account account, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/showchart/" + account.getChart().getChartID());
		return null;
	}

	@Override
	protected void redirectToListMethod(HttpServletRequest req, HttpServletResponse res, Account bean) throws Exception {

		throw new UnsupportedOperationException();
	}

	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {
		
		return callback.list(req, res, user, uriParser, validationErrors);
	}
	
}
