package com.nordicpeak.flowengine.queries.personquery;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.standardutils.annotations.RequiredIfSet;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Key;
import se.unlogic.standardutils.dao.annotations.OneToMany;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLValidationUtils;

import com.nordicpeak.flowengine.interfaces.CitizenIdentifierQuery;
import com.nordicpeak.flowengine.interfaces.ImmutableAlternative;
import com.nordicpeak.flowengine.interfaces.MultiSignQuery;
import com.nordicpeak.flowengine.queries.basequery.BaseQuery;
import com.nordicpeak.flowengine.queries.fixedalternativesquery.FixedAlternativesQuery;

@Table(name = "person_queries")
@XMLElement
public class PersonQuery extends BaseQuery implements CitizenIdentifierQuery, FixedAlternativesQuery, MultiSignQuery {

	private static final long serialVersionUID = 377885620576442774L;

	@DAOManaged
	@Key
	@XMLElement
	private Integer queryID;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean allowNoPersonFound;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean setAsAttribute;

	@DAOManaged
	@WebPopulate(maxLength = 255)
	@RequiredIfSet(paramNames = "setAsAttribute")
	@XMLElement
	private String attributeName;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean useForMultipartSigning;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean setMultipartsAsOwners;

	@DAOManaged
	@WebPopulate
	@XMLElement
	private boolean blockPosterSelection;

	@DAOManaged
	@OneToMany
	@XMLElement
	private List<PersonQueryInstance> instances;

	private PersonAlternative personSelectedAlternative;
	private PersonAlternative personNotFoundAlternative;

	@Override
	public Integer getQueryID() {

		return queryID;
	}

	public List<PersonQueryInstance> getInstances() {

		return instances;
	}

	public void setInstances(List<PersonQueryInstance> instances) {

		this.instances = instances;
	}

	public void setQueryID(int queryID) {

		this.queryID = queryID;
	}

	@Override
	public String toString() {

		if (this.queryDescriptor != null) {

			return queryDescriptor.getName() + " (queryID: " + queryID + ")";
		}

		return "ContactChannelQuery (queryID: " + queryID + ")";
	}

	@Override
	public String getXSDTypeName() {

		return "ContactChannelQuery" + queryID;
	}

	@Override
	public void toXSD(Document doc) {

		Element complexTypeElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexType");
		complexTypeElement.setAttribute("name", getXSDTypeName());

		Element complexContentElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:complexContent");
		complexTypeElement.appendChild(complexContentElement);

		Element extensionElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:extension");
		extensionElement.setAttribute("base", "Query");
		complexContentElement.appendChild(extensionElement);

		Element sequenceElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:sequence");
		extensionElement.appendChild(sequenceElement);

		Element nameElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:element");
		nameElement.setAttribute("name", "Name");
		nameElement.setAttribute("type", "xs:string");
		nameElement.setAttribute("minOccurs", "1");
		nameElement.setAttribute("maxOccurs", "1");
		nameElement.setAttribute("fixed", queryDescriptor.getName());
		sequenceElement.appendChild(nameElement);

		appendFieldDefenition("Firstname", true, doc, sequenceElement);
		appendFieldDefenition("Lastname", true, doc, sequenceElement);
		appendFieldDefenition("CareOf", false, doc, sequenceElement);
		appendFieldDefenition("Address", false, doc, sequenceElement);
		appendFieldDefenition("ZipCode", false, doc, sequenceElement);
		appendFieldDefenition("PostalAddress", false, doc, sequenceElement);
		appendFieldDefenition("SocialSecurityNumber", true, doc, sequenceElement);

		doc.getDocumentElement().appendChild(complexTypeElement);
	}

	private void appendFieldDefenition(String name, boolean required, Document doc, Element sequenceElement) {

		Element fieldElement = doc.createElementNS("http://www.w3.org/2001/XMLSchema", "xs:element");
		fieldElement.setAttribute("name", name);
		fieldElement.setAttribute("type", "xs:string");
		fieldElement.setAttribute("minOccurs", required ? "1" : "0");
		fieldElement.setAttribute("maxOccurs", "1");

		sequenceElement.appendChild(fieldElement);
	}

	@Override
	public void populate(XMLParser xmlParser) throws ValidationException {

		List<ValidationError> errors = new ArrayList<>();

		description = XMLValidationUtils.validateParameter("description", xmlParser, false, 1, 65535, StringPopulator.getPopulator(), errors);
		helpText = XMLValidationUtils.validateParameter("helpText", xmlParser, false, 1, 65535, StringPopulator.getPopulator(), errors);
		allowNoPersonFound = xmlParser.getPrimitiveBoolean("allowNoPersonFound");

		attributeName = XMLValidationUtils.validateParameter("attributeName", xmlParser, false, 1, 255, StringPopulator.getPopulator(), errors);

		if (attributeName != null) {

			setAsAttribute = xmlParser.getPrimitiveBoolean("setAsAttribute");
		}

		useForMultipartSigning = xmlParser.getPrimitiveBoolean("useForMultipartSigning");
		setMultipartsAsOwners = xmlParser.getPrimitiveBoolean("setMultipartsAsOwners");
		blockPosterSelection = xmlParser.getPrimitiveBoolean("blockPosterSelection");

		if (!errors.isEmpty()) {

			throw new ValidationException(errors);
		}

	}

	public boolean isSetAsAttribute() {

		return setAsAttribute;
	}

	public void setSetAsAttribute(boolean setAsAttribute) {

		this.setAsAttribute = setAsAttribute;
	}

	public String getAttributeName() {

		return attributeName;
	}

	public void setAttributeName(String attributeName) {

		this.attributeName = attributeName;
	}

	public PersonAlternative getPersonSelectedAlternative() {

		return personSelectedAlternative;
	}

	public void setPersonSelectedAlternative(PersonAlternative personSelectedAlternative) {

		this.personSelectedAlternative = personSelectedAlternative;
	}

	public PersonAlternative getPersonNotFoundAlternative() {

		return personNotFoundAlternative;
	}

	public void setPersonNotFoundAlternative(PersonAlternative personNotFoundAlternative) {

		this.personNotFoundAlternative = personNotFoundAlternative;
	}

	@Override
	public List<? extends ImmutableAlternative> getAlternatives() {

		return CollectionUtils.getList(personSelectedAlternative, personNotFoundAlternative);
	}

	@Override
	public String getFreeTextAlternative() {

		return null;
	}

	@Override
	public Map<Integer, Integer> getAlternativeConversionMap() {

		return null;
	}

	public boolean isAllowNoPersonFound() {

		return allowNoPersonFound;
	}

	public void setAllowNoPersonFound(boolean allowNoPersonFound) {

		this.allowNoPersonFound = allowNoPersonFound;
	}

	@Override
	public boolean requiresMultipartSigning() {

		return useForMultipartSigning;
	}

	public boolean isUseForMultipartSigning() {

		return useForMultipartSigning;
	}

	public void setUseForMultipartSigning(boolean useForMultipartSigning) {

		this.useForMultipartSigning = useForMultipartSigning;
	}

	public boolean isSetMultipartsAsOwners() {

		return setMultipartsAsOwners;
	}

	public void setSetMultipartsAsOwners(boolean setMultipartsAsOwners) {

		this.setMultipartsAsOwners = setMultipartsAsOwners;
	}

	public boolean isBlockPosterSelection() {

		return blockPosterSelection;
	}

	public void setBlockPosterSelection(boolean blockPosterSelection) {

		this.blockPosterSelection = blockPosterSelection;
	}

}
