package com.nordicpeak.flowengine.queries.plotquery.clients.tokenhandler;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.log4j.Logger;

import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.queries.plotquery.clients.PlotQueryClient;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.TokenRefreshCredentialsException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.TokenRefreshException;

import flexjson.JSONDeserializer;

public class ArcGisTokenHandler {

	private Logger log = Logger.getLogger(this.getClass());

	protected static final String GENERATE_TOKEN_REQUEST_CONTENT_TYPE = "application/x-www-form-urlencoded";

	private static final Integer ERROR_CODE_INVALID_TOKEN = 498;

	private final String url;
	private final String username;
	private final String password;
	private final String sourceIP;
	private final Integer connectionTimeout;
	private final Integer readTimeout;
	private final boolean loggingEnabled;

	protected String token = null;

	public ArcGisTokenHandler(String url, String username, String password, String sourceIP, Integer connectionTimeout, Integer readTimeout, boolean loggingEnabled) {

		this.url = url;
		this.username = username;
		this.password = password;
		this.sourceIP = sourceIP;
		this.connectionTimeout = connectionTimeout;
		this.readTimeout = readTimeout;
		this.loggingEnabled = loggingEnabled;
	}

	public String sendRequest(FunctionWithIOException<String, String> request, String loggingPrefix) throws IOException {

		loggingPrefix += "(TokenRequestID:" + request.hashCode() + ") ";

		return sendRequest(request, loggingPrefix, false);
	}

	protected String sendRequest(FunctionWithIOException<String, String> request, String loggingPrefix, boolean secondTry) throws IOException {

		if (token == null) {

			log(loggingPrefix + "Token is null, generating new token before sending request");

			token = generateToken(loggingPrefix);

			log(loggingPrefix + "Successfully generated new token");
		}

		log(loggingPrefix + "Sending request");

		String response = request.apply(token);

		InvalidTokenResponse parsedResponse = null;
		try {

			parsedResponse = new JSONDeserializer<InvalidTokenResponse>().deserialize(response, InvalidTokenResponse.class);

		} catch (Exception e) {}

		if (parsedResponse != null && parsedResponse.getError() != null) {

			if (ERROR_CODE_INVALID_TOKEN.equals(parsedResponse.getError().getCode())) {

				token = null;

				if (secondTry) {

					throw new TokenRefreshException(loggingPrefix + "Still received invalid token response after token refresh, is the specified token source ip correct? Response: " + response);
				}

				log(loggingPrefix + "Got invalid token response on first try, clearing token and retrying request");

				return sendRequest(request, loggingPrefix, true);
			}

			throw new TokenRefreshException(loggingPrefix + "Unknown error refreshing token " + parsedResponse.getError());
		}

		log(loggingPrefix + "Successfully authenticated request");

		return response;
	}

	protected String generateToken(String loggingPrefix) throws IOException {

		String data = "username=" + username + "&password=" + password + "&ip=" + sourceIP + "&client=ip&f=json";

		String response = sendGenerateTokenRequest(data, loggingPrefix);

		GenerateTokenResponse parsedResponse = null;
		try {

			parsedResponse = new JSONDeserializer<GenerateTokenResponse>().deserialize(response, GenerateTokenResponse.class);

		} catch (Exception e) {}

		if (parsedResponse == null || parsedResponse.getToken() == null) {

			throw new TokenRefreshCredentialsException(loggingPrefix + "Could not parse generate token response: " + response);
		}

		return parsedResponse.getToken();
	}

	protected String sendGenerateTokenRequest(String data, String loggingPrefix) throws IOException {

		InputStream inputStream = new ByteArrayInputStream(data.getBytes());

		SimpleRequest simpleRequest = new SimpleRequest(url);

		simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
		simpleRequest.setReadTimeout(readTimeout * 1000);

		simpleRequest.addHeader("Content-Type", GENERATE_TOKEN_REQUEST_CONTENT_TYPE);

		log(loggingPrefix + "Request URL: " + simpleRequest);
		log(loggingPrefix + "Request Data: " + data);

		String response = HTTPUtils.sendHTTPPostRequest(simpleRequest, inputStream, PlotQueryClient.INTEGRATION_CHARSET);

		log(loggingPrefix + "Response: " + response);

		return response;
	}

	private void log(Object message) {

		if (loggingEnabled) {

			log.info(message);
		}
	}

}
