package com.nordicpeak.flowengine.queue.cruds;

import java.io.IOException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.bool.BooleanUtils;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.LowLevelQuery;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.dao.querys.ObjectQuery;
import se.unlogic.standardutils.dao.querys.UpdateQuery;
import se.unlogic.standardutils.io.BinarySizeFormater;
import se.unlogic.standardutils.io.BinarySizes;
import se.unlogic.standardutils.populators.BooleanPopulator;
import se.unlogic.standardutils.populators.DatePopulator;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.populators.PositiveStringIntegerPopulator;
import se.unlogic.standardutils.populators.SocialSecurityPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.SimpleRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.enums.SenderType;
import com.nordicpeak.flowengine.queue.QueueSlotAdminModule;
import com.nordicpeak.flowengine.queue.QueueUtils;
import com.nordicpeak.flowengine.queue.beans.ExternalMessage;
import com.nordicpeak.flowengine.queue.beans.InternalMessage;
import com.nordicpeak.flowengine.queue.beans.PrimaryContact;
import com.nordicpeak.flowengine.queue.beans.Queue;
import com.nordicpeak.flowengine.queue.beans.QueueSlot;
import com.nordicpeak.flowengine.queue.beans.QueueSlotEvent;
import com.nordicpeak.flowengine.queue.enums.ExternalMessageType;
import com.nordicpeak.flowengine.queue.enums.QueueSlotEventType;
import com.nordicpeak.flowengine.queue.events.ExternalMessageAddedEvent;
import com.nordicpeak.flowengine.queue.events.QueueSlotManuallyAddedEvent;
import com.nordicpeak.personprovider.Person;

public class QueueSlotCRUD extends ModularCRUD<QueueSlot, Integer, User, QueueSlotAdminModule> {

	protected final ExternalMessageCRUD externalMessageCRUD;
	protected final InternalMessageCRUD internalMessageCRUD;
	protected final AdvancedAnnotatedDAOWrapper<QueueSlot, Integer> queueSlotDAOWrapper;

	public QueueSlotCRUD(AdvancedAnnotatedDAOWrapper<QueueSlot, Integer> crudDAO, InternalMessageCRUD internalMessageCRUD, ExternalMessageCRUD externalMessageCRUD, QueueSlotAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), crudDAO, new SimpleRequestPopulator<>(QueueSlot.class), "QueueSlot", "queue slot", "", callback);

		this.queueSlotDAOWrapper = crudDAO;
		this.internalMessageCRUD = internalMessageCRUD;
		this.externalMessageCRUD = externalMessageCRUD;

		setRequirePostForDelete(true);
	}

	@Override
	protected void appendBean(QueueSlot queueSlot, Element targetElement, Document doc, User user) {

		if (callback.isReverseOrderMessages() && queueSlot.getExternalMessages() != null) {

			Collections.reverse(queueSlot.getExternalMessages());
			XMLUtils.appendNewElement(doc, targetElement, "DefaultMessageOrder", "DESC");
		}

		targetElement.appendChild(queueSlot.toXML(doc, callback.getUserHandler()));
	}

	@Override
	protected void appendShowFormData(QueueSlot queueSlot, Document doc, Element showTypeElement, User user, HttpServletRequest req, HttpServletResponse res, URIParser uriParser) throws SQLException, IOException, Exception {

		super.appendShowFormData(queueSlot, doc, showTypeElement, user, req, res, uriParser);

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (req.getParameter("externalmessage") != null) {

				if (QueueUtils.isDigitalMember(queueSlot, callback.getUserHandler())) {

					//TODO append message or request parameters
					ExternalMessage externalMessage = externalMessageCRUD.add(req, res, uriParser, user, doc, showTypeElement, queueSlot, true);

					if (externalMessage != null) {

						QueueSlotEvent queueSlotEvent = callback.addQueueSlotEvent(queueSlot, QueueSlotEventType.MANAGER_MESSAGE_SENT, null, user, null, null);

						callback.getEventHandler().sendEvent(QueueSlot.class, new ExternalMessageAddedEvent(queueSlot, queueSlotEvent, externalMessage, SenderType.MANAGER), EventTarget.ALL);
						callback.getEventHandler().sendEvent(ExternalMessage.class, new CRUDEvent<>(CRUDAction.ADD, externalMessage), EventTarget.ALL);

						res.sendRedirect(req.getContextPath() + uriParser.getFormattedURI() + "#messages");
						return;
					}

				} else {

					log.warn("User " + user + " tried to add external message for queue slot " + queueSlot + " which has no digital user");
				}

			} else if (req.getParameter("internalmessage") != null) {

				//TODO append message or request parameters
				InternalMessage internalMessage = internalMessageCRUD.add(req, res, uriParser, user, doc, showTypeElement, queueSlot);

				if (internalMessage != null) {

					res.sendRedirect(req.getContextPath() + uriParser.getFormattedURI() + "#notes");

					callback.getEventHandler().sendEvent(InternalMessage.class, new CRUDEvent<>(CRUDAction.ADD, internalMessage), EventTarget.ALL);

					return;
				}
			}
		}

		if (callback.hasAdminAccess(user)) {

			XMLUtils.appendNewElement(doc, showTypeElement, "AdminAccess");
		}

		XMLUtils.appendNewElement(doc, showTypeElement, "FormatedMaxMessageFileSize", BinarySizeFormater.getFormatedSize(callback.getMaxFileSize() * BinarySizes.MegaByte));
	}

	@Override
	protected void appendListFormData(Document doc, Element listTypeElement, User user, HttpServletRequest req, URIParser uriParser, List<ValidationError> validationError) throws SQLException {

		super.appendListFormData(doc, listTypeElement, user, req, uriParser, validationError);

		appendFormData(doc, listTypeElement, user, req, uriParser);

		XMLUtils.appendNewElement(doc, listTypeElement, "PageSize", callback.getDefaultPageSize());
		XMLUtils.appendNewElement(doc, listTypeElement, "FullAlias", callback.getFullAlias());

		if (callback.hasAdminAccess(user)) {

			XMLUtils.appendNewElement(doc, listTypeElement, "AdminAccess");
		}
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);

		appendFormData(doc, addTypeElement, user, req, uriParser);
	}

	private void appendFormData(Document doc, Element element, User user, HttpServletRequest req, URIParser uriParser) {

		Queue queue = (Queue) req.getAttribute("queue");
		element.appendChild(queue.toXML(doc));
	}

	@Override
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {

		checkListAccess(user, req, uriParser);

		boolean archived = req.getParameter("archived") != null;

		log.info("User " + user + " listing " + (archived ? "archived " : "") + typeLogPluralName);

		String typeElementName = typeElementPluralName;

		if (archived) {
			typeElementName = "Archived" + typeElementName;
		}

		Document doc = this.callback.createDocument(req, uriParser, user);
		Element listTypeElement = doc.createElement("List" + typeElementName);
		doc.getFirstChild().appendChild(listTypeElement);

		//		appendAllBeans(doc, listTypeElement, user, req, uriParser, validationErrors);

		if (validationErrors != null) {
			XMLUtils.append(doc, listTypeElement, validationErrors);
			listTypeElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		appendListFormData(doc, listTypeElement, user, req, uriParser, validationErrors);

		SimpleForegroundModuleResponse moduleResponse = createListModuleResponse(doc, req, user, uriParser);

		moduleResponse.addBreadcrumbsLast(getListBreadcrumbs(req, user, uriParser, validationErrors));

		return moduleResponse;
	}

	@Override
	protected QueueSlot populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {

		Queue queue = (Queue) req.getAttribute("queue");
		Timestamp now = new Timestamp(System.currentTimeMillis());

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		QueueSlot queueSlot = null;

		try {
			queueSlot = super.populateFromAddRequest(req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		if (queueSlot != null) {

			try {
				super.validateAddPopulation(queueSlot, req, user, uriParser);

			} catch (ValidationException e) {

				validationErrors.addAll(e.getErrors());
			}

			java.sql.Date addedDate = ValidationUtils.validateParameter("addedDate", req, false, DatePopulator.getPopulator(), validationErrors);

			String citizenIdentifier = ValidationUtils.validateParameter("citizenIdentifier", req, true, SocialSecurityPopulator.getPopulator(), validationErrors);

			if (!StringUtils.isEmpty(citizenIdentifier)) {

				citizenIdentifier = citizenIdentifier.replace("-", "");

				LocalDate localDate = callback.getLatestWaitingDate(queue, citizenIdentifier);
				LocalDate chosenDate = addedDate == null ? LocalDate.now() : addedDate.toLocalDate();

				if (!queue.isAllowDuplicateQueueSlots() && callback.isActiveMemberOfQueue(queue, citizenIdentifier)) {

					validationErrors.add(new ValidationError("PersonAlreadyInQueue"));

				} 
				
				if (localDate != null && (localDate.equals(chosenDate) || localDate.isAfter(chosenDate))) {
					
					validationErrors.add(new ValidationError("PersonIsInWaitingPeriod", localDate.format(DateTimeFormatter.ofPattern("yyyy-MM-dd"))));
				
				} 
				
				Person person = callback.getPerson(citizenIdentifier);

				if (person == null) {

					validationErrors.add(new ValidationError("ErrorImportingCitizen"));

				} else {

					User personUser = callback.getUserHandler().getUserByAttribute("citizenIdentifier", citizenIdentifier, false, true);

					queueSlot.setPrimaryContact(new PrimaryContact(person, personUser));
				}
				
			}

			boolean payed = BooleanUtils.valueOf(ValidationUtils.validateParameter("payed", req, false, BooleanPopulator.getPopulator(), validationErrors));

			Date paymentDate = ValidationUtils.validateParameter("paymentDate", req, false, DatePopulator.getPopulator(), validationErrors);

			Integer paymentAmount = ValidationUtils.validateParameter("paymentAmount", req, payed, IntegerPopulator.getPopulator(), validationErrors);

			if (paymentDate != null && new Date().before(paymentDate)) {

				validationErrors.add(new ValidationError("PaymentDateInFuture"));
			}

			if (payed) {

				Timestamp timestamp;

				if (paymentDate != null) {

					timestamp = new Timestamp(paymentDate.getTime());

				} else {

					timestamp = now;
				}

				queueSlot.setLastPaymentDate(timestamp);

				queueSlot.addEvent(new QueueSlotEvent(timestamp, user, QueueSlotEventType.PAYED, callback.getEventPaymentManagerMessage() + paymentAmount + " kr", null));
			}

			String comment = ValidationUtils.validateParameter("comment", req, false, 0, 16777215, validationErrors);

			queueSlot.addEvent(new QueueSlotEvent(now, user, QueueSlotEventType.SUBMITTED, callback.getEventAdminSubmittedMessage() + (StringUtils.isEmpty(comment) ? "" : callback.getEventSeeExternalMessage()), null));

			if (!StringUtils.isEmpty(comment)) {

				req.setAttribute("comment", comment);
			}

			if (addedDate != null) {

				if (new Date().before(addedDate)) {

					validationErrors.add(new ValidationError("AddedDateInFuture"));

				} else {

					queueSlot.setAdded(new Timestamp(addedDate.getTime()));
				}

			} else {

				queueSlot.setAdded(new Timestamp(System.currentTimeMillis()));
			}
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		queueSlot.setQueue(queue);

		return queueSlot;
	}

	@Override
	protected QueueSlot populateFromUpdateRequest(QueueSlot queueSlot, HttpServletRequest req, User user, URIParser uriParser) throws ValidationException, Exception {

		int oldPosition = queueSlot.getCalculatedPosition();
		req.setAttribute("oldPosition", oldPosition);

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		Integer position = ValidationUtils.validateParameter("position", req, true, PositiveStringIntegerPopulator.getPopulator(), validationErrors);

		if (position != null) {

			queueSlot.setCalculatedPosition(position);
		}

		String comment = ValidationUtils.validateParameter("comment", req, false, 0, 16777215, validationErrors);

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		if (!StringUtils.isEmpty(comment)) {

			req.setAttribute("comment", comment);
		}

		return queueSlot;
	}

	@Override
	protected void addFilteredBean(QueueSlot queueSlot, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		Queue queue = (Queue) req.getAttribute("queue");

		TransactionHandler transactionHandler = getTransactionHandler(req);

		ObjectQuery<Integer> query = transactionHandler.getObjectQuery("SELECT MAX(queueIndex) FROM " + queueSlotDAOWrapper.getAnnotatedDAO().getTableName() + " WHERE queueID = ?", IntegerPopulator.getPopulator());
		query.setInt(1, queue.getQueueID());

		queueSlot.setQueueIndex(1 + query.executeQuery());

		super.addFilteredBean(queueSlot, req, user, uriParser);
	}

	@Override
	protected void updateFilteredBean(QueueSlot queueSlot, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		String comment = (String) req.getAttribute("comment");
		int oldPosition = (Integer) req.getAttribute("oldPosition");
		int queueID = queueSlot.getQueue().getQueueID();

		TransactionHandler transactionHandler = getTransactionHandler(req);

		// Get max rows
		ObjectQuery<Integer> rowsQuery = transactionHandler.getObjectQuery("SELECT COUNT(*) FROM " + queueSlotDAOWrapper.getAnnotatedDAO().getTableName() + " WHERE queueID = ? AND archived = 0", IntegerPopulator.getPopulator());
		rowsQuery.setInt(1, queueID);

		int tableRows = rowsQuery.executeQuery();

		int newPosition = queueSlot.getCalculatedPosition();

		if (newPosition > tableRows) {

			newPosition = tableRows;
		}

		if (oldPosition != newPosition) {

			// Add sorted event
			queueSlot.addEvent(new QueueSlotEvent(new Timestamp(System.currentTimeMillis()), user, QueueSlotEventType.MOVE, callback.getEventSortQueueMessage1() + oldPosition + callback.getEventSortQueueMessage2() + (newPosition) + (StringUtils.isEmpty(comment) ? "" : callback.getEventSeeExternalMessage()), null));

			LowLevelQuery<QueueSlot> query = new LowLevelQuery<QueueSlot>("SELECT * FROM " + queueSlotDAOWrapper.getAnnotatedDAO().getTableName() + " WHERE queueID = ? AND archived = 0 ORDER BY queueIndex ASC");
			query.addParameter(queueID);

			List<QueueSlot> queueSlots = queueSlotDAOWrapper.getAnnotatedDAO().getAll(query);

			// Set temp index
			queueSlot.setQueueIndex(-1);
			crudDAO.update(queueSlot, transactionHandler);

			if (newPosition < oldPosition) {

				for (int i = oldPosition - 1; i >= newPosition; i--) {

					QueueSlot slot = queueSlots.get(i - 1);
					QueueSlot slot2 = queueSlots.get(i - 1 + 1);

					UpdateQuery updateQuery = transactionHandler.getUpdateQuery("UPDATE " + queueSlotDAOWrapper.getAnnotatedDAO().getTableName() + " SET queueIndex = ? WHERE queueSlotID = ?");
					updateQuery.setInt(1, slot2.getQueueIndex());
					updateQuery.setInt(2, slot.getQueueSlotID());
					updateQuery.executeUpdate();
				}

			} else {

				for (int i = oldPosition + 1; i <= newPosition; i++) {

					QueueSlot slot = queueSlots.get(i - 1);
					QueueSlot slot2 = queueSlots.get(i - 1 - 1);

					UpdateQuery updateQuery = transactionHandler.getUpdateQuery("UPDATE " + queueSlotDAOWrapper.getAnnotatedDAO().getTableName() + " SET queueIndex = ? WHERE queueSlotID = ?");
					updateQuery.setInt(1, slot2.getQueueIndex());
					updateQuery.setInt(2, slot.getQueueSlotID());
					updateQuery.executeUpdate();
				}
			}

			// Set new index
			queueSlot.setQueueIndex(queueSlots.get(newPosition - 1).getQueueIndex());
		}

		super.updateFilteredBean(queueSlot, req, user, uriParser);
	}

	@Override
	public List<QueueSlot> getAllBeans(User user, HttpServletRequest req, URIParser uriParser) throws SQLException {

		return null;
	}

	@Override
	public QueueSlot getBean(Integer beanID, String getMode, HttpServletRequest req) throws SQLException, AccessDeniedException {

		QueueSlot slot = super.getBean(beanID, getMode, req);

		if (slot != null) {

			if (slot.getFlowInstanceID() != null) {
				slot.setFlowInstanceXMLOnly(callback.getFlowInstance(slot.getFlowInstanceID()));
			}

			slot.getPoster(callback.getUserHandler());

			if (!slot.isArchived()) {
				QueueUtils.getCalculatedQueuePosition(queueSlotDAOWrapper.getAnnotatedDAO().getDataSource(), slot);
			}
		}

		return slot;
	}

	@Override
	public void checkAddAccess(User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		Queue queue = (Queue) req.getAttribute("queue");

		if (!callback.hasQueueManagerAccess(queue, user)) {

			throw new AccessDeniedException(callback.getSectionDescriptor());
		}
	}

	@Override
	protected void checkListAccess(User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		Queue queue = (Queue) req.getAttribute("queue");

		if (queue == null) {

			throw new URINotFoundException(uriParser);
		}

		if (!callback.hasQueueManagerAccess(queue, user)) {

			throw new AccessDeniedException(callback.getSectionDescriptor());
		}
	}

	@Override
	public void checkShowAccess(QueueSlot queueSlot, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!callback.hasQueueManagerAccess(queueSlot.getQueue(), user)) {

			throw new AccessDeniedException(callback.getSectionDescriptor());
		}
	}

	@Override
	public void checkUpdateAccess(QueueSlot queueSlot, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!callback.hasQueueManagerAccess(queueSlot.getQueue(), user)) {

			throw new AccessDeniedException(callback.getSectionDescriptor());
		}
	}

	@Override
	protected void checkDeleteAccess(QueueSlot queueSlot, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!callback.hasAdminAccess(user) || !queueSlot.isArchived()) {

			throw new AccessDeniedException(callback.getSectionDescriptor());
		}
	}

	@Override
	protected ForegroundModuleResponse beanAdded(QueueSlot queueSlot, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String comment = (String) req.getAttribute("comment");

		if (!StringUtils.isEmpty(comment)) {
			callback.addQueueSlotExternalMessage(queueSlot, comment, true, user, ExternalMessageType.MANUAL_ENTRANCE, null);
		}

		callback.getEventHandler().sendEvent(QueueSlot.class, new QueueSlotManuallyAddedEvent(queueSlot, user), EventTarget.ALL);

		return super.beanAdded(queueSlot, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(QueueSlot bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Queue queue = (Queue) req.getAttribute("queue");

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/list/" + queue.getQueueID());
		return null;
	}

	@Override
	protected ForegroundModuleResponse beanUpdated(QueueSlot queueSlot, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String comment = (String) req.getAttribute("comment");

		if (!StringUtils.isEmpty(comment)) {
			callback.addQueueSlotExternalMessage(queueSlot, comment, true, user, ExternalMessageType.CHANGED_POSITION, null);
		}

		return super.beanUpdated(queueSlot, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(QueueSlot queueSlot, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/list/" + queueSlot.getQueue().getQueueID());
		return null;
	}

	@Override
	protected ForegroundModuleResponse filteredBeanDeleted(QueueSlot bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/list/" + bean.getQueue().getQueueID());
		return null;
	}

}
