package com.nordicpeak.flowengine.integration.sokigo;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.AccessUtils;
import se.unlogic.standardutils.base64.Base64;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.HTTPMethod;
import se.unlogic.webutils.http.exceptions.HTTPException;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.childrelationprovider.Child;
import com.nordicpeak.flowengine.beans.DefaultInstanceMetadata;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.enums.QueryState;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.exceptions.FlowEngineException;
import com.nordicpeak.flowengine.integration.base.BaseIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationSettingsException;
import com.nordicpeak.flowengine.integration.sokigo.beans.IntegrationEndpoint;
import com.nordicpeak.flowengine.integration.sokigo.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.sokigo.beans.QueuedIntegrationRequest;
import com.nordicpeak.flowengine.integration.sokigo.beans.SokigoSchoolTransportChildQueryFilterEndpoint;
import com.nordicpeak.flowengine.integration.sokigo.exceptions.IncompleteRequestException;
import com.nordicpeak.flowengine.interfaces.ImmutableFlow;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.managers.ImmutableFlowInstanceManager;
import com.nordicpeak.flowengine.queries.childquery.ChildQueryProviderModule;
import com.nordicpeak.flowengine.queries.childquery.beans.ChildQuery;
import com.nordicpeak.flowengine.queries.childquery.beans.ChildQueryInstance;
import com.nordicpeak.flowengine.queries.childquery.filterapi.FilterAPIChild;
import com.nordicpeak.flowengine.queries.childquery.interfaces.ChildQueryFilterProvider;

import flexjson.JSONDeserializer;

public class SokigoSchoolTransportIntegrationProviderModule extends BaseIntegrationProviderModule<QueuedIntegrationRequest, IntegrationSettings, IntegrationEndpoint> implements ChildQueryFilterProvider {

	private static final AnnotatedRequestPopulator<IntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<>(IntegrationSettings.class);

	private static final List<String> EXTRA_FIELDS = CollectionUtils.getList("Skola", "rskurs", "Klass");

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Error email subject", description = "The subject of emails sent when requests fail")
	@XSLVariable(prefix = "java.")
	private String errorEmailSubject;

	@ModuleSetting(allowsNull = true)
	@HTMLEditorSettingDescriptor(name = "Error email message", description = "The message of emails sent when errors are received")
	@XSLVariable(prefix = "java.")
	private String errorEmailMessage;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Endpoint Name", description = "", required = true)
	private String endpointName = "Sokigo skolskjuts";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Encoding", description = "", required = true)
	private String encoding = "UTF-8";

	@XSLVariable(prefix = "java.")
	private String providerName = "SokigoSchoolTransportIntegrationProviderModule";

	@InstanceManagerDependency
	private QueryHandler queryHandler;

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;

	private ChildQueryProviderModule childQueryProviderModule;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, SokigoSchoolTransportIntegrationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		endpointDAO = daoFactory.getDAO(IntegrationEndpoint.class);
		queuedRequestDAO = daoFactory.getDAO(QueuedIntegrationRequest.class);

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
		endpointIDParamFactory = endpointDAO.getParamFactory("endpointID", Integer.class);
		lastRetryParamFactory = queuedRequestDAO.getParamFactory("lastRetry", Timestamp.class);
		eventIDParamFactory = queuedRequestDAO.getParamFactory("eventID", Integer.class);
		flowInstanceIDParamFactory = queuedRequestDAO.getParamFactory("flowInstanceID", Integer.class);
	}

	@InstanceManagerDependency
	public void setChildQueryProviderModule(ChildQueryProviderModule childQueryProviderModule) throws SQLException {

		if (childQueryProviderModule != null) {

			childQueryProviderModule.addChildQueryFilterProvider(this);

		} else if (this.childQueryProviderModule != null) {

			this.childQueryProviderModule.removeChildQueryFilterProvider(this);
		}

		this.childQueryProviderModule = childQueryProviderModule;
	}

	public Map<String, FilterAPIChild> getChildren(SokigoSchoolTransportChildQueryFilterEndpoint endpoint, Map<String, Child> navetChildMap, User user, String parentCitizenID, ImmutableFlow flow) {

		IntegrationSettings settings;

		try {
			settings = getIntegrationSettings(flow);

		} catch (SQLException e) {

			log.error("Error getting integration settings for flow " + flow, e);
			return null;
		}

		if (settings == null) {

			log.warn("Found no integration settings for flow " + flow);
			return null;
		}

		Map<String, FilterAPIChild> children = new HashMap<String, FilterAPIChild>();

		for (Entry<String, Child> child : navetChildMap.entrySet()) {

			String childCitizenID = child.getKey();

			log.info("Looking up sokigo child " + childCitizenID);

			String serviceURL = settings.getEndpoint().getServiceAddress().replace("$replace", settings.getServiceName());

			SimpleRequest simpleRequest = new SimpleRequest(serviceURL + "/" + settings.getApiKey() + "/GetPupil/" + childCitizenID);
			setRequestSettings(simpleRequest, settings);

			try {
				HTTPResponse response = HTTPUtils.sendHTTPGetRequest(simpleRequest, Charset.forName(encoding));

				try {
					JSONDeserializer<Map<String, Object>> deserializer = new JSONDeserializer<Map<String, Object>>();

					Map<String, Object> parsedResponse = deserializer.deserialize(response.getValue(), Map.class);

					String className = (String) parsedResponse.get("className");
					String gradeName = (String) parsedResponse.get("gradeName");
					String schoolName = (String) parsedResponse.get("schoolName");

					if (!ObjectUtils.isNull(className, gradeName, schoolName)) {

						Map<String, String> fields = new HashMap<String, String>();

						fields.put("Klass", className);
						fields.put("rskurs", gradeName);
						fields.put("Skola", schoolName);

						children.put(child.getKey(), new FilterAPIChild(child.getKey(), fields));
					}

				} catch (Exception e) {

					log.warn("Error parsing children JSON \"" + response.getValue() + "\" for " + endpoint, e);
					return null;
				}

			} catch (HTTPException e) {

				if ("\"Pupil not found! Check the ID-number.\"".equals(e.getMessage())) {

					log.warn("No pupil found for child citizenID " + childCitizenID);

				} else {

					log.warn("Error connecting to " + endpoint, e);
					return null;
				}

			} catch (IOException e) {

				log.warn("Error connecting to " + endpoint, e);
				return null;
			}
		}

		return children;
	}

	private void setRequestSettings(SimpleRequest simpleRequest, IntegrationSettings settings) {

		simpleRequest.setFollowRedirects(false);
		simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
		simpleRequest.setReadTimeout(requestTimeout * 1000);
		simpleRequest.setUsername(settings.getUsername());
		simpleRequest.setPassword(settings.getPassword());
	}

	@Override
	public List<SokigoSchoolTransportChildQueryFilterEndpoint> getEndpoints() throws SQLException {

		return Collections.singletonList(getEndpoint(0));
	}

	public SokigoSchoolTransportChildQueryFilterEndpoint getEndpoint(Integer endpointID) throws SQLException {

		return new SokigoSchoolTransportChildQueryFilterEndpoint(endpointName, EXTRA_FIELDS, this);
	}

	@Override
	public SokigoSchoolTransportChildQueryFilterEndpoint getEndpoint(String name) throws SQLException {

		if (name.equals(endpointName)) {
			return getEndpoint(0);
		}

		return null;
	}

	public List<ChildQuery> getQueries(SokigoSchoolTransportChildQueryFilterEndpoint endpoint) throws SQLException {

		return childQueryProviderModule.getQueriesUsingFilterEndpoint(endpoint.getName());
	}

	@WebPublic(alias = "update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));

		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (integrationSettings == null) {

				integrationSettings = new IntegrationSettings();
				integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			}

			try {
				POPULATOR.populate(integrationSettings, req);

				List<ValidationError> validationErrors = new ArrayList<>();

				IntegrationEndpoint endpoint = getEndpoint(req);

				if (endpoint == null) {

					validationErrors.add(new ValidationError("NoEndpointSelected"));

				} else {

					integrationSettings.setEndpoint(endpoint);
				}

				if (!validationErrors.isEmpty()) {

					throw new ValidationException(validationErrors);
				}

				log.info("User " + user + " updating integration settings for flow " + flow);

				if (integrationSettings.isPopulated()) {

					settingsDAO.addOrUpdate(integrationSettings, null);

				} else {

					settingsDAO.delete(integrationSettings);
				}

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#sokigointegrationsettings");

				return null;

			} catch (ValidationException e) {

				validationException = e;
			}
		}

		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);
		XMLUtils.append(doc, integrationSettingsElement, "Endpoints", endpointDAO.getAll());

		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws Exception {

		FlowInstance flowInstance = (FlowInstance) event.getFlowInstanceManager().getFlowInstance();
		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled()) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = event.getEvent();

		try {

			processIntegrationEvent(flowInstance, flowInstanceEvent);

		} catch (IncompleteRequestException e) {

			if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationRequestException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue");

				addQueuedEvent(flowInstanceEvent, flowInstance);

			} else if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationSettingsException e) {

			log.error("Settings error while processing event " + event, e);
		}
	}

	@Override
	public void retryEvent(FlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler) throws Exception {

		try {

			super.retryEvent(flowInstance, event, queuedRequest, transactionHandler);

		} catch (IncompleteRequestException e) {

			log.info("Deleting " + queuedRequest + " due to previous error");

			queuedRequestDAO.delete(queuedRequest, transactionHandler);

			IntegrationSettings settings;

			if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null && settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}
		}
	}

	@Override
	protected Field[] getFlowInstanceRelations() {

		return new Field[] { FlowInstanceEvent.FLOW_INSTANCE_RELATION, FlowInstance.FLOW_RELATION, FlowInstance.STATUS_RELATION, Flow.FLOW_TYPE_RELATION, Flow.FLOW_FAMILY_RELATION };
	}

	@Override
	public void processIntegrationEvent(FlowInstance flowInstance, FlowInstanceEvent event) throws SQLException, IntegrationSettingsException, IntegrationRequestException {

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow());

		if (settings == null) {

			throw new IntegrationSettingsException("Settings are null");
		}

		if (!settings.isEnabled()) {

			throw new IntegrationSettingsException("Settings are disabled");
		}

		log.info("Triggering sokigo integration for event " + event + " belonging to flow instance " + flowInstance);

		Charset charset = Charset.forName(encoding);

		File pdfFile = pdfProvider.getPDF(flowInstance.getFlowInstanceID(), event.getEventID());

		if (pdfFile == null) {

			log.error("No PDF found for event " + event + " belonging to flow instance " + flowInstance);
			return;
		}

		String serviceURL = settings.getEndpoint().getServiceAddress().replace("$replace", settings.getServiceName());
		boolean shouldLogAsError = maxRetryCount == 0 && (!settings.isSendErrorEmail() || CollectionUtils.isEmpty(settings.getErrorEmailAddresses()));

		try {

			int flowInstanceID = flowInstance.getFlowInstanceID();

			FlowInstanceManager flowInstanceManager = flowAdminModule.getImmutableFlowInstanceManager(flowInstanceID);

			List<ChildQueryInstance> queryInstances = flowInstanceManager.getQueries(ChildQueryInstance.class);

			if (queryInstances != null) {
				for (ChildQueryInstance queryInstance : queryInstances) {

					if (endpointName.equals(queryInstance.getQuery().getFilterEndpointName())) {

						if (queryInstance.getQueryInstanceDescriptor().getQueryState() != QueryState.HIDDEN && queryInstance.getQueryInstanceDescriptor().isPopulated()) {

							String childCitizenID = queryInstance.getCitizenIdentifier();

							try {
								log.info("Registering shool transport application for " + childCitizenID + " from " + flowInstance);

								SimpleRequest simpleRequest = new SimpleRequest(serviceURL + "/" + settings.getApiKey() + "/RegisterSchoolTransportApplication/" + childCitizenID + "/" + flowInstanceID);
								simpleRequest.setRequestMethod(HTTPMethod.PUT);

								setRequestSettings(simpleRequest, settings);

								String response = HTTPUtils.sendHTTPPostRequest(simpleRequest, new ByteArrayInputStream(new byte[0]), charset);

								if (!"true".equalsIgnoreCase(response)) {

									logResponse("\"" + response + "\" response when registering school transport application for " + childCitizenID + " in " + flowInstance + " from " + event, null, shouldLogAsError);

									throw new IncompleteRequestException();
								}

							} catch (HTTPException e) {

								if ("\"No CAID found for pupil! Check the ID-number.\"".equals(e.getMessage())) {

									log.warn("Error registering school transport application, no pupil found for child citizenID, for " + childCitizenID + " in " + flowInstance + " from " + event);

									throw new IncompleteRequestException();

								} else if ("\"ApplicationID already exist on Pupil!\"".equals(e.getMessage())) { // Probably completion, continue with PDF upload

									log.warn("Error registering school transport application, application already exists with flowInstanceID " + flowInstanceID + ", for " + childCitizenID + " in " + flowInstance + " from " + event);

								} else {

									logResponse("Error registering school transport application for " + childCitizenID + " in " + flowInstance + " from " + event, e, shouldLogAsError);

									throw new IntegrationRequestException();
								}

							} catch (IOException e) {

								logResponse("Error registering school transport application for " + childCitizenID + " in " + flowInstance + " from " + event, e, shouldLogAsError);

								throw new IntegrationRequestException();
							}

							try {
								log.info("Sending PDF for " + flowInstance);

								SimpleRequest simpleRequest = new SimpleRequest(serviceURL + "/" + settings.getApiKey() + "/UploadFileToPupil/" + childCitizenID);
								setRequestSettings(simpleRequest, settings);

								List<Entry<String, String>> headerEntries = new ArrayList<Entry<String, String>>(1);
								headerEntries.add(new SimpleEntry<String, String>("Content-Type", "application/json"));

								simpleRequest.setHeaders(headerEntries);

								String filename = flowInstance.getFlow().getName() + " - " + flowInstanceID;

								JsonObject fileInfo = new JsonObject();
								fileInfo.putField("FileName", FileUtils.toValidHttpFilename(filename));
								fileInfo.putField("Description", "");
								fileInfo.putField("Name", filename);
								fileInfo.putField("FileDataBase64", Base64.encodeFromFile(pdfFile));
								fileInfo.putField("FileExtension", ".pdf");

								String response = HTTPUtils.sendHTTPPostRequest(simpleRequest, StringUtils.getInputStream(fileInfo.toJson(), charset), charset);

								if (!"true".equalsIgnoreCase(response)) {

									logResponse("\"" + response + "\" response when sending PDF for " + childCitizenID + " in " + flowInstance + " from " + event, null, shouldLogAsError);

									throw new IncompleteRequestException();
								}

							} catch (HTTPException e) {

								if ("\"No CAID found for pupil! Check the ID-number.\"".equals(e.getMessage())) {

									logResponse("Error sending PDF no application found for " + childCitizenID + " in " + flowInstance + " from " + event, null, shouldLogAsError);

									throw new IncompleteRequestException();

								} else {

									logResponse("Error sending PDF for " + childCitizenID + " in " + flowInstance + " from " + event, e, shouldLogAsError);

									throw new IntegrationRequestException();
								}

							} catch (IOException e) {

								logResponse("Error sending PDF for " + childCitizenID + " in " + flowInstance + " from " + event, e, shouldLogAsError);

								throw new IntegrationRequestException();
							}
						}
					}
				}
			}

		} catch (FlowEngineException e) {

			log.error("Error opening ImmutableFlowInstanceManager for " + flowInstance, e);
		}
	}

	@Override
	protected QueuedIntegrationRequest getNewQueuedIntegrationRequest() {

		return new QueuedIntegrationRequest();
	}

	@Override
	protected void handleMaxRetryCountExceeded(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler, IntegrationRequestException lastRequestException) throws SQLException {

		super.handleMaxRetryCountExceeded(flowInstance, event, queuedRequest, transactionHandler, lastRequestException);

		IntegrationSettings settings;

		if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

			if (settings.isSendErrorEmail() && !CollectionUtils.isEmpty(settings.getErrorEmailAddresses())) {

				log.info("Sending error emails due to max retries exceeded");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));

			} else {

				log.error(queuedRequest + " hit its max retry count and error email is disabled");
			}
		}
	}

	@Override
	protected String getShowFlowModuleAnchorLink() {

		return "#sokigointegrationsettings";
	}

	@WebPublic
	public ForegroundModuleResponse unregisterApplication(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		if (!AccessUtils.checkAccess(user, adminAccessInterface)) {

			throw new AccessDeniedException("Not admin");
		}

		if (flowAdminModule == null) {

			return new SimpleForegroundModuleResponse("Missing required dependency");
		}

		Integer flowInstanceID = uriParser.getInt(2);

		log.info("User " + user + " unregistering transport application from flowInstance ID: " + flowInstanceID);

		if (flowInstanceID == null) {

			throw new URINotFoundException(uriParser);
		}

		FlowInstance flowInstance = flowAdminModule.getFlowInstance(flowInstanceID);

		if (flowInstance == null) {

			throw new URINotFoundException(uriParser);
		}

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow());

		if (settings == null || !settings.isEnabled()) {

			return new SimpleForegroundModuleResponse("No settings or disabled");
		}

		String serviceURL = settings.getEndpoint().getServiceAddress().replace("$replace", settings.getServiceName());

		FlowInstanceManager flowInstanceManager = new ImmutableFlowInstanceManager(flowInstance, queryHandler, null, new DefaultInstanceMetadata(flowAdminModule.getSiteProfile(flowInstance)), null);

		List<ChildQueryInstance> queryInstances = flowInstanceManager.getQueries(ChildQueryInstance.class);

		if (queryInstances != null) {
			for (ChildQueryInstance queryInstance : queryInstances) {

				if (endpointName.equals(queryInstance.getQuery().getFilterEndpointName())) {

					if (queryInstance.getQueryInstanceDescriptor().getQueryState() != QueryState.HIDDEN && queryInstance.getQueryInstanceDescriptor().isPopulated()) {

						String childCitizenID = queryInstance.getCitizenIdentifier();

						try {
							log.info("Unregistering shool transport application for " + childCitizenID + " from " + flowInstance);

							SimpleRequest simpleRequest = new SimpleRequest(serviceURL + "/" + settings.getApiKey() + "/UnRegisterSchoolTransportApplication/" + childCitizenID);
							simpleRequest.setRequestMethod(HTTPMethod.PUT);

							setRequestSettings(simpleRequest, settings);

							String response = HTTPUtils.sendHTTPPostRequest(simpleRequest, new ByteArrayInputStream(new byte[0]), Charset.forName(encoding));

							log.info("Unregister result: " + response);

						} catch (Exception e) {

							log.warn("Error Unregistering shool transport application", e);
						}
					}
				}
			}
		}

		return new SimpleForegroundModuleResponse("Unregister attempted, check log.");
	}

	@Override
	public String getName() {

		return this.providerName;
	}

}