package com.nordicpeak.flowengine.integration.treserva;

import java.io.File;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.GroupMultiListSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.standardutils.base64.Base64;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.FlowInstanceAdminModule;
import com.nordicpeak.flowengine.beans.DefaultInstanceMetadata;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.beans.QueryDescriptor;
import com.nordicpeak.flowengine.beans.QueryInstanceDescriptor;
import com.nordicpeak.flowengine.beans.SigningParty;
import com.nordicpeak.flowengine.beans.Step;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.ManagersChangedEvent;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.integration.base.BaseIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationSettingsException;
import com.nordicpeak.flowengine.integration.treserva.beans.ApplicationDetails;
import com.nordicpeak.flowengine.integration.treserva.beans.CaseData;
import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationEndpoint;
import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationParameter;
import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.treserva.beans.QueuedIntegrationRequest;
import com.nordicpeak.flowengine.integration.treserva.clients.TreservaEkonomisktBistandClient;
import com.nordicpeak.flowengine.integration.treserva.clients.TreservaMinaIFOArendeClient;
import com.nordicpeak.flowengine.integration.treserva.exceptions.AutoAssignManagerException;
import com.nordicpeak.flowengine.integration.treserva.exceptions.BadResponseException;
import com.nordicpeak.flowengine.integration.treserva.exceptions.ExistingApplicationException;
import com.nordicpeak.flowengine.integration.treserva.exceptions.ProtectedPersonException;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.managers.ImmutableFlowInstanceManager;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;
import com.nordicpeak.flowengine.utils.FlowInstanceUtils;
import com.nordicpeak.flowengine.utils.MultiSignUtils;

import flexjson.JSONDeserializer;

public class TreservaIntegrationProviderModule extends BaseIntegrationProviderModule<QueuedIntegrationRequest, IntegrationSettings, IntegrationEndpoint> {

	private static final String JSON_CITIZEN_IDENTIFIER = "personNr";

	private static final AnnotatedRequestPopulator<IntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<>(IntegrationSettings.class);

	@XSLVariable(name = "i18n.Name")
	protected String paramNameDisplayName;

	@XSLVariable(name = "i18n.Value")
	protected String paramValueDisplayName;

	@XSLVariable(prefix = "java.")
	private String debugEmailReason = "Not set";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Log generated JSON", description = "Toggles logging of the generated JSON")
	protected boolean logJSON;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Log API request data and response for MinaIFOArende API", description = "Toggles logging of all API request data and response for MinaIFOArende API")
	protected boolean minaIFOArendeAPILoggingEnabled;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Existing application response message", description = "Responses containing this string will be treated as a existing application error", required = true)
	private String existingApplicationResponseMessage = "Det finns redan en obehandlad anskan fr aktuell person som  mste hanteras frst";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Protected person response message", description = "Responses containing this string will be treated as a protected person error", required = true)
	private String protectedPersonResponseMessage = "Protected person can not allowed to submit the application";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Error email subject", description = "The subject of emails sent when requests fail")
	@XSLVariable(prefix = "java.")
	private String errorEmailSubject = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Existing application error email message", description = "The message of emails sent when existing application errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String existingApplicationErrorEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Protected person error email message", description = "The message of emails sent when protected person errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String protectedPersonErrorEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Unknown error email message", description = "The message of emails sent when unknown errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String unknownErrorEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Max retries exceeded email message", description = "The message of emails sent when the max number of retries for a queued request is exceeded", required = true)
	@XSLVariable(prefix = "java.")
	private String maxRetriesExceededEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "No manager to auto assign email message", description = "The message of emails sent when no manager was found for auto assignment", required = true)
	@XSLVariable(prefix = "java.")
	private String noManagerToAutoAssignEmailMessage = "Not set";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "No manager to auto assign debug emails", description = "Enables sending log messages as email")
	protected boolean noManagerToAutoAssignDebugEmails;

	@ModuleSetting(allowsNull = true)
	@GroupMultiListSettingDescriptor(name = "Auto Assign Manager Groups", description = "Users in these groups are evaluated for auto assignment. If empty all users are evaluated.")
	private List<Integer> autoAssignManagerGroupIDs;

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;

	@InstanceManagerDependency(required = true)
	protected QueryHandler queryHandler;

	@InstanceManagerDependency(required = true)
	private FlowInstanceAdminModule flowInstanceAdminModule;

	private AnnotatedDAO<FlowInstance> flowInstanceDAO;

	protected QueryParameterFactory<FlowInstance, Integer> flowInstanceFlowInstanceIDParamFactory;
	protected QueryParameterFactory<QueryInstanceDescriptor, Integer> queryInstanceDescriptorFlowInstanceIDParamFactory;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		if (!systemInterface.getInstanceHandler().addInstance(TreservaIntegrationProviderModule.class, this)) {

			throw new RuntimeException("Unable to register module " + this.moduleDescriptor + " in global instance handler using key " + TreservaIntegrationProviderModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(TreservaIntegrationProviderModule.class, this);

		super.unload();
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, this.getClass().getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		endpointDAO = daoFactory.getDAO(IntegrationEndpoint.class);
		queuedRequestDAO = daoFactory.getDAO(QueuedIntegrationRequest.class);

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
		endpointIDParamFactory = endpointDAO.getParamFactory("endpointID", Integer.class);
		lastRetryParamFactory = queuedRequestDAO.getParamFactory("lastRetry", Timestamp.class);
		eventIDParamFactory = queuedRequestDAO.getParamFactory("eventID", Integer.class);
		flowInstanceIDParamFactory = queuedRequestDAO.getParamFactory("flowInstanceID", Integer.class);

		FlowEngineDAOFactory flowEngineDAOFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());

		flowInstanceDAO = flowEngineDAOFactory.getFlowInstanceDAO();

		flowInstanceFlowInstanceIDParamFactory = flowInstanceDAO.getParamFactory("flowInstanceID", Integer.class);
		queryInstanceDescriptorFlowInstanceIDParamFactory = flowEngineDAOFactory.getQueryInstanceDescriptorDAO().getParamFactory("flowInstanceID", Integer.class);

	}

	private FlowInstance getFlowInstanceForManager(int flowInstanceID) throws SQLException {

		HighLevelQuery<FlowInstance> query = new HighLevelQuery<>(FlowInstance.OWNERS_RELATION, FlowInstance.FLOW_RELATION, FlowInstance.STATUS_RELATION, Flow.STEPS_RELATION, Flow.FLOW_FAMILY_RELATION, Step.QUERY_DESCRIPTORS_RELATION, QueryDescriptor.QUERY_INSTANCE_DESCRIPTORS_RELATION, FlowInstance.ATTRIBUTES_RELATION, FlowInstance.EVENTS_RELATION, FlowInstanceEvent.ATTRIBUTES_RELATION);

		query.addParameter(flowInstanceFlowInstanceIDParamFactory.getParameter(flowInstanceID));
		query.addRelationParameter(QueryInstanceDescriptor.class, queryInstanceDescriptorFlowInstanceIDParamFactory.getParameter(flowInstanceID));

		return flowInstanceDAO.get(query);
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws Exception {

		FlowInstance flowInstance = (FlowInstance) event.getFlowInstanceManager().getFlowInstance();
		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !shouldProcessEvent(event)) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = event.getEvent();

		try {

			processIntegrationEvent(flowInstance, flowInstanceEvent);

		} catch (ExistingApplicationException e) {

			if (settings.isSendExistingApplicationErrorEmail()) {

				log.info("Sending existing application error emails for flow instance" + flowInstance);

				sendEmails(settings.getExistingApplicationErrorEmailAddresses(), errorEmailSubject, existingApplicationErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

			if (maxRetryCount > 0) {

				log.info("Adding event " + event + " to retry queue for flow instance" + flowInstance);

				addQueuedEvent(flowInstanceEvent, flowInstance);
			}

		} catch (ProtectedPersonException e) {

			if (settings.isSendProtectedPersonErrorEmail()) {

				log.info("Sending protected person error emails for flow instance" + flowInstance);

				sendEmails(settings.getProtectedPersonErrorEmailAddresses(), errorEmailSubject, protectedPersonErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (BadResponseException e) {

			if (settings.isSendUnknownErrorEmail()) {

				log.info("Sending unknown error emails for flow instance" + flowInstance);

				sendEmails(settings.getUnknownErrorEmailAddresses(), errorEmailSubject, unknownErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationRequestException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue for flow instance" + flowInstance);

				addQueuedEvent(flowInstanceEvent, flowInstance);
			}

		} catch (IntegrationSettingsException e) {

			log.error("Settings error while processing event " + event + " for flow instance" + flowInstance, e);
		}
	}

	@Override
	public void retryEvent(FlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler) throws Exception {

		try {

			super.retryEvent(flowInstance, event, queuedRequest, transactionHandler);

		} catch (ProtectedPersonException | BadResponseException e) {

			log.info("Deleting " + queuedRequest + " due to previous error");

			queuedRequestDAO.delete(queuedRequest, transactionHandler);

			IntegrationSettings settings;

			if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

				if (e instanceof ProtectedPersonException && settings.isSendProtectedPersonErrorEmail()) {

					log.info("Sending protected person error emails for flow instance" + flowInstance);

					sendEmails(settings.getProtectedPersonErrorEmailAddresses(), errorEmailSubject, protectedPersonErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));

				} else if (e instanceof BadResponseException && settings.isSendUnknownErrorEmail()) {

					log.info("Sending unknown error emails for flow instance" + flowInstance);

					sendEmails(settings.getUnknownErrorEmailAddresses(), errorEmailSubject, unknownErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
				}
			}
		}
	}

	@Override
	public void processIntegrationEvent(FlowInstance flowInstance, FlowInstanceEvent event) throws SQLException, IntegrationRequestException, IntegrationSettingsException {

		IntegrationSettings integrationSettings = getIntegrationSettings(flowInstance.getFlow());

		if (integrationSettings == null) {

			throw new IntegrationSettingsException("Settings are null");
		}

		if (!integrationSettings.isEnabled()) {

			throw new IntegrationSettingsException("Settings are disabled");
		}

		User poster = systemInterface.getUserHandler().getUser(event.getPoster().getUserID(), false, true);
		String citizenIdentifier = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(poster);

		log.info("Processing event " + event + " for flow instance " + flowInstance + " and poster " + poster);

		boolean applicationSaved = false;

		try {

			applicationSaved = saveApplicationRequest(flowInstance, event, integrationSettings, citizenIdentifier);

		} catch (ExistingApplicationException e) {

			if (integrationSettings.isAutoAssignManager()) {

				try {

					autoAssignManagers(flowInstance, citizenIdentifier, integrationSettings);

				} catch (AutoAssignManagerException e1) {

					log.warn("ExistingApplicationException - AutoAssignManagerRequest: " + e1.getMessage(), e1.getCause());
				}

			}

			throw e;
		}

		if (applicationSaved && integrationSettings.isAutoAssignManager()) {

			try {

				autoAssignManagers(flowInstance, citizenIdentifier, integrationSettings);

			} catch (AutoAssignManagerException e) {

				log.warn("AutoAssignManagerRequest: " + e.getMessage(), e.getCause());

				String message = noManagerToAutoAssignEmailMessage;

				if (noManagerToAutoAssignDebugEmails) {

					message += debugEmailReason + e.getMessage();
				}

				log.info("Sending no manager to auto assign emails for flow instance " + flowInstance);

				sendEmails(integrationSettings.getNoManagerToAutoAssignEmailAddresses(), errorEmailSubject, message, flowInstance, getTagReplacer(flowInstance));
			}
		}
	}

	private boolean saveApplicationRequest(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, IntegrationSettings integrationSettings, String citizenIdentifier) throws IntegrationRequestException {

		LocalDateTime latestApplicationDate = null;

		if (integrationSettings.isDuplicateCheckEnabled()) {

			latestApplicationDate = getLatestApplicationDateRequest(flowInstance, citizenIdentifier, integrationSettings);
		}

		String applicationDataJSON;

		try {

			applicationDataJSON = generateApplicationData(flowInstance, event, integrationSettings);

		} catch (IOException e) {

			log.error("Error generating application data for event " + event + " belonging to flow instance " + flowInstance, e);
			return false;
		}

		log.info("Sending save application request for event " + event + " from flow instance " + flowInstance + " to endpoint " + integrationSettings.getEndpoint());

		TreservaEkonomisktBistandClient client = new TreservaEkonomisktBistandClient(integrationSettings.getEndpoint(), connectionTimeout, requestTimeout);
		String response;

		try {

			response = client.saveApplication(applicationDataJSON);

		} catch (SocketTimeoutException e) {

			if (latestApplicationDate != null) {

				LocalDateTime newLatestApplicationDate = getLatestApplicationDateRequest(flowInstance, citizenIdentifier, integrationSettings);

				if (newLatestApplicationDate != null && !newLatestApplicationDate.equals(latestApplicationDate)) {

					log.info("Verified save application request success after request timeout for flow instance " + flowInstance);

					return true;
				}
			}

			boolean shouldLogAsError = maxRetryCount == 0;

			logResponse("Socket timeout during save application request for flow instance " + flowInstance, null, shouldLogAsError);

			throw new IntegrationRequestException();

		} catch (Exception e) {

			boolean shouldLogAsError = maxRetryCount == 0;

			logResponse("Error during save application request for flow instance " + flowInstance, e, shouldLogAsError);

			throw new IntegrationRequestException();
		}

		if (response != null && response.contains("\"message\":\"Success\"")) {

			log.info("Save application request successfully sent for flow instance " + flowInstance + ", response: " + response);

			return true;

		}

		if (response != null && response.contains(existingApplicationResponseMessage)) {

			boolean shouldLogError = maxRetryCount == 0 && (!integrationSettings.isSendExistingApplicationErrorEmail() || CollectionUtils.isEmpty(integrationSettings.getExistingApplicationErrorEmailAddresses()));

			logResponse("Existing application error on save application request for flow instance " + flowInstance + ", response: " + response, null, shouldLogError);

			throw new ExistingApplicationException();

		}

		if (response != null && response.contains(protectedPersonResponseMessage)) {

			boolean shouldLogError = !integrationSettings.isSendProtectedPersonErrorEmail() || CollectionUtils.isEmpty(integrationSettings.getProtectedPersonErrorEmailAddresses());

			logResponse("Protected person error on save application request for flow instance " + flowInstance + ", response: " + response, null, shouldLogError);

			throw new ProtectedPersonException();

		}

		boolean shouldLogError = !integrationSettings.isSendUnknownErrorEmail() || CollectionUtils.isEmpty(integrationSettings.getUnknownErrorEmailAddresses());

		logResponse("Bad response received on save application request for flow instance " + flowInstance + ", response: " + response, null, shouldLogError);

		throw new BadResponseException();
	}

	private String generateApplicationData(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, IntegrationSettings integrationSettings) throws IOException {

		MutableAttributeHandler attributeHandler = (MutableAttributeHandler) flowInstance.getAttributeHandler();

		TagReplacer tagReplacer = getTagReplacer(flowInstance);

		JsonObject jsonDocument = new JsonObject();

		jsonDocument.putField(JSON_CITIZEN_IDENTIFIER, replaceTags(integrationSettings.getPersonNr(), attributeHandler, tagReplacer));
		jsonDocument.putField("resubmit", "false");

		String stadsDelID = null;

		if (integrationSettings.isSetStadsDelIdFromAttribute()) {

			stadsDelID = attributeHandler.getString(integrationSettings.getStadsDelIdAttributeName());
		}

		if (StringUtils.isEmpty(stadsDelID)) {

			stadsDelID = integrationSettings.getStadsDelId();
		}

		jsonDocument.putField("stadsDelId", replaceTags(stadsDelID, attributeHandler, tagReplacer));

		JsonArray dataArray = new JsonArray();

		if (integrationSettings.getParameters() != null) {

			for (IntegrationParameter parameter : integrationSettings.getParameters()) {

				appendDataObject(dataArray, parameter.getName(), parameter.getValue(), attributeHandler, tagReplacer);
			}
		}

		if (integrationSettings.isCommentEnabled()) {

			appendDataObject(dataArray, integrationSettings.getCommentName(), integrationSettings.getCommentValue(), attributeHandler, tagReplacer);
		}

		if (!dataArray.isEmpty()) {

			jsonDocument.putField("data", dataArray);
		}

		File pdf = pdfProvider.getPDF(flowInstance.getFlowInstanceID(), event.getEventID());

		if (pdf == null) {

			throw new IOException("No PDF found for event " + event + " belonging to flow instance " + flowInstance);
		}

		jsonDocument.putField("pdf", Base64.encodeFromFile(pdf));

		String applicationDataJSON = jsonDocument.toJson();

		if (logJSON) {

			log.info("Generated JSON for event " + event + " belonging to flow instance " + flowInstance + "\n" + applicationDataJSON);
		}

		return applicationDataJSON;
	}

	private LocalDateTime getLatestApplicationDateRequest(ImmutableFlowInstance flowInstance, String citizenIdentifier, IntegrationSettings integrationSettings) {

		if (citizenIdentifier == null) {

			log.warn("Missing citizenIdentifier, can not get latest application date for flow instance " + flowInstance);
			return null;
		}

		log.info("Sending get aktualiseringar for person request for flow instance " + flowInstance + " to endpoint " + integrationSettings.getEndpoint());

		TreservaMinaIFOArendeClient client = getTreservaMinaIFOArendeClient(integrationSettings);

		String response;
		try {

			response = client.getAktualiseringarForPerson(generateCitizenIdentifierJSON(citizenIdentifier));

		} catch (InvalidKeyException | NoSuchAlgorithmException | IOException e) {

			log.warn("Error getting aktualiseringar for person for flow instance " + flowInstance, e);
			return null;
		}

		List<ApplicationDetails> applicationDetailsList = new JSONDeserializer<List<ApplicationDetails>>().use("values", ApplicationDetails.class).deserialize(response, ArrayList.class);

		if (CollectionUtils.isEmpty(applicationDetailsList)) {

			log.warn("Empty response getting aktualiseringar for person for flow instance " + flowInstance);
			return null;
		}

		ApplicationDetails latestValidApplicationDetails = null;

		for (ApplicationDetails applicationDetails : applicationDetailsList) {

			if (!applicationDetails.hasValidType()) {

				continue;
			}

			if (latestValidApplicationDetails == null || applicationDetails.getDate() != null && applicationDetails.getDate().isAfter(latestValidApplicationDetails.getDate())) {

				latestValidApplicationDetails = applicationDetails;
			}
		}

		if (latestValidApplicationDetails == null) {

			log.warn("Found no valid application details in response for flow instance " + flowInstance);
			return null;
		}

		return latestValidApplicationDetails.getDate();
	}

	private void autoAssignManagers(ImmutableFlowInstance flowInstance, String posterCitizenIdentifier, IntegrationSettings integrationSettings) throws SQLException, AutoAssignManagerException {

		if (!integrationSettings.getEndpoint().hasMinaIFOArendeConfiguration()) {

			throw new AutoAssignManagerException("Missing MinaIFOArende configuration, can not auto assign manager for flow instance " + flowInstance);
		}

		Set<String> citizenIdentifiers = getSigningPartiesCitizenIdentifiers(flowInstance);

		citizenIdentifiers.add(posterCitizenIdentifier);

		if (citizenIdentifiers.isEmpty()) {

			throw new AutoAssignManagerException("Missing citizenIdentifier, can not auto assign manager for flow instance " + flowInstance);
		}

		TreservaMinaIFOArendeClient client = getTreservaMinaIFOArendeClient(integrationSettings);

		int addedManagersCount = 0;

		List<String> errorMessages = null;

		for (String citizenIdentifer : citizenIdentifiers) {

			try {

				autoAssignManager(flowInstance, integrationSettings, client, citizenIdentifer);

				addedManagersCount++;

			} catch (AutoAssignManagerException e) {

				log.warn(e.getMessage(), e.getCause());

				errorMessages = CollectionUtils.addAndInstantiateIfNeeded(errorMessages, e.getMessage());
			}
		}

		if (addedManagersCount == 0) {

			throw new AutoAssignManagerException("Could not auto assign managers to flowInstance " + flowInstance + " due to previous errors: " + StringUtils.toCommaSeparatedString(errorMessages));
		}
	}

	private void autoAssignManager(ImmutableFlowInstance flowInstance, IntegrationSettings integrationSettings, TreservaMinaIFOArendeClient client, String citizenIdentifer) throws AutoAssignManagerException, SQLException {

		log.info("Sending get arende data by personnr request for flow instance " + flowInstance + " to endpoint " + integrationSettings.getEndpoint());

		CaseData caseData = null;

		try {

			caseData = client.getLatestValidCaseData(citizenIdentifer);

		} catch (InvalidKeyException | NoSuchAlgorithmException | IOException e) {

			throw new AutoAssignManagerException("Error getting arende data by personnr for flow instance " + flowInstance, e);
		}

		if (caseData == null) {

			throw new AutoAssignManagerException("Found no valid case data in response for flow instance " + flowInstance);
		}

		String email = caseData.getEmail();

		if (StringUtils.isEmpty(email)) {

			throw new AutoAssignManagerException("Case data has no email, can not auto assign manager for flow instance " + flowInstance);
		}

		User newManager = findNewManager(email);

		if (newManager == null) {

			throw new AutoAssignManagerException("Could not find user with email " + email + ", can not auto assign manager for flow instance " + flowInstance);
		}

		addManagerToFlowInstance((FlowInstance) flowInstance, newManager);
	}

	public Set<String> getSigningPartiesCitizenIdentifiers(ImmutableFlowInstance flowInstance) {

		Set<String> citizenIdentifiers = new HashSet<>();

		try {

			ImmutableFlowInstanceManager instanceManager = new ImmutableFlowInstanceManager(getFlowInstanceForManager(flowInstance.getFlowInstanceID()), queryHandler, null, new DefaultInstanceMetadata(flowInstanceAdminModule.getSiteProfile(flowInstance)), null);

			Set<SigningParty> signingParties = MultiSignUtils.getSigningParties(instanceManager);

			if (signingParties != null) {

				for (SigningParty signingParty : signingParties) {

					citizenIdentifiers.add(signingParty.getSocialSecurityNumber());
				}
			}

		} catch (Exception e) {

			log.error("Error getting instance manager for flowInstance " + flowInstance + ", can not get signing parties", e);
		}

		return citizenIdentifiers;
	}

	private User findNewManager(String email) {

		List<User> users = systemInterface.getUserHandler().getUsersByEmail(Collections.singletonList(email), true, false);

		return CollectionUtils.find(users, this::isManagerAvailableForAutoAssignment);
	}

	private boolean isManagerAvailableForAutoAssignment(User user) {

		if (!user.isEnabled()) {

			return false;
		}

		if (!CollectionUtils.isEmpty(autoAssignManagerGroupIDs) && CollectionUtils.find(user.getGroups(), group -> autoAssignManagerGroupIDs.contains(group.getGroupID())) == null) {

			return false;
		}

		return true;
	}

	private void addManagerToFlowInstance(FlowInstance flowInstance, User newManager) throws SQLException, AutoAssignManagerException {

		FlowFamily flowFamily = flowInstance.getFlow().getFlowFamily();
		List<User> previousManagers = flowInstance.getManagers();

		if (CollectionUtils.contains(previousManagers, newManager)) {

			log.info("The user " + newManager + " is already manager of flowInstance " + flowInstance);
			return;
		}

		if (!flowAdminModule.hasFlowFamilyManagerAccess(flowFamily, newManager)) {

			throw new AutoAssignManagerException("Tried to add manager " + newManager + " to flowInstance " + flowInstance + " but the user did not have the required access");
		}

		List<User> flowInstanceManagers = CollectionUtils.addAndInstantiateIfNeeded(previousManagers, newManager);

		flowInstance.setManagers(flowInstanceManagers);

		RelationQuery updateQuery = new RelationQuery(FlowInstance.MANAGERS_RELATION);

		flowInstanceDAO.update(flowInstance, updateQuery);

		String detailString = FlowInstanceUtils.getManagersString(flowInstance.getManagers(), flowInstance.getManagerGroups());

		SiteProfile profile = flowAdminModule.getSiteProfile(flowInstance);

		FlowInstanceEvent flowInstanceEvent = flowAdminModule.getFlowInstanceEventGenerator().addFlowInstanceEvent(flowInstance, EventType.MANAGERS_UPDATED, detailString, null);

		systemInterface.getEventHandler().sendEvent(FlowInstance.class, new CRUDEvent<>(CRUDAction.UPDATE, flowInstance), EventTarget.ALL);

		systemInterface.getEventHandler().sendEvent(FlowInstance.class, new ManagersChangedEvent(flowInstance, flowInstanceEvent, profile, previousManagers, flowInstance.getManagerGroups(), null), EventTarget.ALL);

		log.info("Added the user " + newManager + " as manager of flowInstance " + flowInstance);
	}

	@WebPublic(alias = "update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));

		}

		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			if (integrationSettings == null) {

				integrationSettings = new IntegrationSettings();
				integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
			}

			try {
				POPULATOR.populate(integrationSettings, req);

				List<ValidationError> validationErrors = new ArrayList<>();

				IntegrationEndpoint endpoint = getEndpoint(req);

				if (endpoint == null) {

					validationErrors.add(new ValidationError("NoEndpointSelected"));

				} else {

					integrationSettings.setEndpoint(endpoint);
				}

				String[] parameterIDs = req.getParameterValues("parameterID");

				if (parameterIDs != null) {

					List<IntegrationParameter> parameters = new ArrayList<>(parameterIDs.length);

					for (String parameterID : parameterIDs) {

						String name = ValidationUtils.validateParameter("name-" + parameterID, paramNameDisplayName, req, true, 1, 255, StringPopulator.getPopulator(), validationErrors);
						String value = ValidationUtils.validateParameter("value-" + parameterID, paramValueDisplayName, req, true, 1, 255, StringPopulator.getPopulator(), validationErrors);

						if (validationErrors.isEmpty()) {

							IntegrationParameter parameter = new IntegrationParameter();
							parameter.setName(name);
							parameter.setValue(value);
							parameters.add(parameter);
						}
					}

					integrationSettings.setParameters(parameters);

				} else {

					integrationSettings.setParameters(null);
				}

				if (!validationErrors.isEmpty()) {

					throw new ValidationException(validationErrors);
				}

				log.info("User " + user + " updating integration settings for flow " + flow);

				if (integrationSettings.isPopulated()) {

					settingsDAO.addOrUpdate(integrationSettings, null);

				} else {

					settingsDAO.delete(integrationSettings);
				}

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + getShowFlowModuleAnchorLink());

				return null;

			} catch (ValidationException e) {

				validationException = e;
			}
		}

		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);
		XMLUtils.append(doc, integrationSettingsElement, "Endpoints", endpointDAO.getAll());
		XMLUtils.appendNewElement(doc, integrationSettingsElement, "MaxRetryCount", maxRetryCount);

		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, getDefaultBreadcrumb());
	}

	@Override
	protected void handleMaxRetryCountExceeded(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler, IntegrationRequestException lastRequestException) throws SQLException {

		super.handleMaxRetryCountExceeded(flowInstance, event, queuedRequest, transactionHandler, lastRequestException);

		IntegrationSettings settings;

		if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

			if (settings.isSendMaxRetriesExceededEmail() && !CollectionUtils.isEmpty(settings.getMaxRetriesExceededEmailAddresses())) {

				log.info("Sending max retries exceeded error emails for flow instance" + flowInstance);

				sendEmails(settings.getMaxRetriesExceededEmailAddresses(), errorEmailSubject, maxRetriesExceededEmailMessage, flowInstance, getTagReplacer(flowInstance));

			} else {

				log.error(queuedRequest + " hit its max retry count and max retries exceeded email is disabled");
			}
		}
	}

	@Override
	protected QueuedIntegrationRequest getNewQueuedIntegrationRequest() {

		return new QueuedIntegrationRequest();
	}

	@Override
	protected String getShowFlowModuleAnchorLink() {

		return "#treservaintegrationsettings";
	}

	public static String generateCitizenIdentifierJSON(String citizenIdentifier) {

		JsonObject json = new JsonObject();

		json.putField(JSON_CITIZEN_IDENTIFIER, citizenIdentifier);

		return json.toJson();
	}

	public TreservaMinaIFOArendeClient getTreservaMinaIFOArendeClient(IntegrationSettings settings) {

		if (settings.getEndpoint() != null) {

			return new TreservaMinaIFOArendeClient(settings.getEndpoint(), connectionTimeout, requestTimeout, minaIFOArendeAPILoggingEnabled, settings.getValidCaseDataTypes());
		}

		return null;
	}

	public IntegrationSettings getIntegrationSettings(ImmutableFlowInstance flowInstance) throws SQLException {

		return getIntegrationSettings(flowInstance.getFlow());
	}

}
