package com.nordicpeak.flowengine.integration.treserva.clients;

import java.io.IOException;
import java.nio.charset.Charset;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import se.unlogic.standardutils.base64.Base64;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationEndpoint;

public class TreservaEkonomisktBistandClient {

	private static final TimeZone TIME_ZONE = TimeZone.getTimeZone("UTC");

	private static final Charset UTF8 = Charset.forName("UTF-8");

	private final String hmacID;

	private final String hmacKey;

	private final String apiBaseURL;

	private final String hmacBaseURL;

	private final Integer connectionTimeout;

	private final Integer requestTimeout;

	public TreservaEkonomisktBistandClient(String hmacID, String hmacKey, String apiBaseURL, String hmacBaseURL, Integer connectionTimeout, Integer requestTimeout) {

		super();
		this.hmacID = hmacID;
		this.hmacKey = hmacKey;
		this.apiBaseURL = apiBaseURL;

		if (hmacBaseURL != null) {

			this.hmacBaseURL = hmacBaseURL;

		} else {

			this.hmacBaseURL = apiBaseURL;
		}

		this.connectionTimeout = connectionTimeout;
		this.requestTimeout = requestTimeout;
	}

	public TreservaEkonomisktBistandClient(IntegrationEndpoint endpoint, Integer connectionTimeout, Integer requestTimeout) {

		this(endpoint.getEkonomisktBistandHmacID(), endpoint.getEkonomisktBistandHmacKey(), endpoint.getEkonomisktBistandAPIBaseURL(), endpoint.getEkonomisktBistandHmacBaseURL(), connectionTimeout, requestTimeout);
	}

	public String getVersion() throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetVersion", "", false);
	}

	public String getStadsDelList() throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetStadsDelList", "", false);
	}

	public String getAktuliseringsOrsakList() throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetAktuliseringsOrsakList", "", false);
	}

	public String saveApplication(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("SaveApplication", data, true);
	}

	private String apiGetCall(String apiMethod, String data, boolean post) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSXXX");
		dateFormat.setTimeZone(TIME_ZONE);
		String isoTime = dateFormat.format(new Date());

		SimpleRequest request = new SimpleRequest(apiBaseURL + apiMethod);
		request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
		request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

		String hmac = getHMAC(hmacKey, data + hmacBaseURL + apiMethod + isoTime + hmacID);

		request.addHeader("HMAC", hmac);
		request.addHeader("AccessId", hmacID);
		request.addHeader("RequestDateTime", isoTime);
		request.addHeader("X-AuthenticationMethod", "Hmac");

		if (post) {

			request.addHeader("Content-Type", "application/json;charset=UTF-8");

			return HTTPUtils.sendHTTPPostRequest(request, StringUtils.getInputStream(data, UTF8), UTF8);

		} else {

			HTTPResponse response = HTTPUtils.sendHTTPGetRequest(request, UTF8);

			return response.getValue();
		}
	}

	public static String getHMAC(String secretKey, String data) throws NoSuchAlgorithmException, InvalidKeyException {

		SecretKeySpec keySpec = new SecretKeySpec(secretKey.getBytes(), "HmacSHA256");

		Mac mac = Mac.getInstance("HmacSHA256");
		mac.init(keySpec);

		byte[] result = mac.doFinal(data.getBytes(UTF8));

		return Base64.encodeBytes(result);
	}
}
