package com.nordicpeak.flowengine.integration.treserva.clients;

import java.io.IOException;
import java.nio.charset.Charset;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import org.apache.log4j.Logger;

import se.unlogic.standardutils.base64.Base64;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.integration.treserva.TreservaIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.treserva.beans.CaseData;
import com.nordicpeak.flowengine.integration.treserva.beans.IntegrationEndpoint;

import flexjson.JSONDeserializer;

public class TreservaMinaIFOArendeClient {

	private Logger log = Logger.getLogger(this.getClass());

	private static final TimeZone TIME_ZONE = TimeZone.getTimeZone("UTC");

	private static final Charset UTF8 = Charset.forName("UTF-8");

	private final String hmacID;

	private final String hmacKey;

	private final String apiBaseURL;

	private final String hmacBaseURL;

	private final Integer connectionTimeout;

	private final Integer requestTimeout;

	private final boolean apiLoggingEnabled;

	private final List<String> validCaseDataTypes;

	public TreservaMinaIFOArendeClient(String hmacID, String hmacKey, String apiBaseURL, String hmacBaseURL, Integer connectionTimeout, Integer requestTimeout, boolean apiLoggingEnabled, List<String> validCaseDataTypes) {

		this.hmacID = hmacID;
		this.hmacKey = hmacKey;
		this.apiBaseURL = apiBaseURL;

		if (hmacBaseURL != null) {

			this.hmacBaseURL = hmacBaseURL;

		} else {

			this.hmacBaseURL = apiBaseURL;
		}

		this.connectionTimeout = connectionTimeout;
		this.requestTimeout = requestTimeout;
		this.apiLoggingEnabled = apiLoggingEnabled;
		this.validCaseDataTypes = validCaseDataTypes;
	}

	public TreservaMinaIFOArendeClient(IntegrationEndpoint endpoint, Integer connectionTimeout, Integer requestTimeout, boolean apiLoggingEnabled, List<String> validCaseDataTypes) {

		this(endpoint.getMinaIFOArendeHmacID(), endpoint.getMinaIFOArendeHmacKey(), endpoint.getMinaIFOArendeAPIBaseURL(), endpoint.getMinaIFOArendeHmacBaseURL(), connectionTimeout, requestTimeout, apiLoggingEnabled, validCaseDataTypes);
	}

	public String getVersion() throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetVersion", "", false);
	}

	public String getMinaIfoArende(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetMinaIfoArende", data, true);
	}

	public CaseData getLatestValidCaseData(String citizenIdentifier) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return getLatestValidCaseData(citizenIdentifier, null);
	}

	public CaseData getLatestValidCaseData(String citizenIdentifier, String posterCitizenIdentifier) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		String response = getArendeDataByPersonNr(TreservaIntegrationProviderModule.generateCitizenIdentifierJSON(citizenIdentifier));

		if (response == null) {

			return null;
		}

		List<CaseData> caseDatas = new JSONDeserializer<List<CaseData>>().use("values", CaseData.class).deserialize(response, ArrayList.class);

		if (CollectionUtils.isEmpty(caseDatas)) {

			return null;
		}

		CaseData latestValidCaseData = null;

		for (CaseData caseData : caseDatas) {

			if (validCaseDataTypes != null && CollectionUtils.find(validCaseDataTypes, type -> type.equalsIgnoreCase(caseData.getType())) == null) {

				continue;
			}

			if (latestValidCaseData == null || caseData.getParsedFromDate() != null && caseData.getParsedFromDate().isAfter(latestValidCaseData.getParsedFromDate())) {

				latestValidCaseData = caseData;
			}
		}

		if (latestValidCaseData == null || posterCitizenIdentifier != null && !latestValidCaseData.isCoApplicant(posterCitizenIdentifier)) {

			return null;
		}

		return latestValidCaseData;
	}

	public String getArendeDataByPersonNr(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetArendeDataByPersonNr", data, true);
	}

	public String getAktualiseringarForPerson(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetAktualiseringarForPerson", data, true);
	}

	public String getBetalningForPerson(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetBetalningForPerson", data, true);
	}

	public String GetBeslutDataByPerson(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetBeslutDataByPerson", data, true);
	}

	public String getBeslutDataByPersonkrets(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetBeslutDataByPersonkrets", data, true);
	}

	public String getArendeDocumentsForPerson(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetArendeDocumentsForPerson", data, true);
	}

	public String getDocumentaForPerson(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetDocumentaForPerson", data, true);
	}

	public String getNormberakningDetails(String data) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		return apiGetCall("GetNormberakningDetails", data, true);
	}

	private String apiGetCall(String apiMethod, String data, boolean post) throws InvalidKeyException, NoSuchAlgorithmException, IOException {

		if (apiLoggingEnabled) {

			log.warn(apiMethod + " Request: " + data);
		}

		DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSXXX");
		dateFormat.setTimeZone(TIME_ZONE);
		String isoTime = dateFormat.format(new Date());

		SimpleRequest request = new SimpleRequest(apiBaseURL + apiMethod);
		request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
		request.setReadTimeout(requestTimeout * MillisecondTimeUnits.SECOND);

		String hmac = getHMAC(hmacKey, data + hmacBaseURL + apiMethod + isoTime + hmacID);

		request.addHeader("HMAC", hmac);
		request.addHeader("AccessId", hmacID);
		request.addHeader("RequestDateTime", isoTime);
		request.addHeader("X-AuthenticationMethod", "Hmac");

		String responseText;

		if (post) {

			request.addHeader("Content-Type", "application/json;charset=UTF-8");

			responseText = HTTPUtils.sendHTTPPostRequest(request, StringUtils.getInputStream(data, UTF8), UTF8);

		} else {

			HTTPResponse response = HTTPUtils.sendHTTPGetRequest(request, UTF8);

			responseText = response.getValue();
		}

		if (apiLoggingEnabled) {

			log.warn(apiMethod + " Response: " + responseText);
		}

		return responseText;
	}

	public static String getHMAC(String secretKey, String data) throws NoSuchAlgorithmException, InvalidKeyException {

		SecretKeySpec keySpec = new SecretKeySpec(secretKey.getBytes(), "HmacSHA256");

		Mac mac = Mac.getInstance("HmacSHA256");
		mac.init(keySpec);

		byte[] result = mac.doFinal(data.getBytes(UTF8));

		return Base64.encodeBytes(result);
	}

}
