package com.nordicpeak.flowengine.userinvoices;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.charset.Charset;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.AccessInterface;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.extensionlinks.ExtensionLink;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.rest.URIParam;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.io.FileUtils;
import se.unlogic.standardutils.mime.MimeUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.CharsetValidator;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.http.enums.ContentDisposition;

import com.nordicpeak.flowengine.UserFlowInstanceMenuModule;
import com.nordicpeak.flowengine.interfaces.UserMenuProvider;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

public class UserInvoiceModule extends AnnotatedRESTModule implements UserMenuProvider {

	@XSLVariable(prefix = "java.")
	private String userMenuTabTitle = "My invoices";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Title", description = "The title displayed above the recommended flows", required = true)
	private String title = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Description", description = "The descripton dsiplayed above the recommended flows", required = true)
	private String description = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Help text", description = "The help text for the recommended flows", required = true)
	private String helpText = "Not set";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Invoice list API URL", description = "URL to the invoice list API, $citizenIdentifier will be replaced with the users citizen identifier.")
	private String invoiceListURL;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Invoice download API URL", description = "URL to the invoice download API, $invoiceID will be replaced with the invoice ID.")
	private String invoiceDownloadURL;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API username", description = "")
	private String apiUsername;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "API password", description = "")
	private String apiPassword;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Encoding", description = "Encoding to use when parsing the API response", required = true, formatValidator = CharsetValidator.class)
	private String encoding = "UTF-8";

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection", required = true)
	private Integer connectionTimeout = 15;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read timeout", description = "The maximum time in seconds to wait for a response", required = true)
	private Integer readTimeout = 15;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "User menu item slot", description = "User menu item slot")
	protected String userMenuExtensionLinkSlot = "60";
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Show decimals in amount", description = "Show decimals for invoice amount")
	private boolean showDecimals = false;
	
	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Show currency in amount", description = "Show currency in invoice amount")
	private boolean showCurrency = false;

	protected UserFlowInstanceMenuModule userFlowInstanceMenuModule;

	protected ExtensionLink userMenuLink;
	
	private DecimalFormat numberFormat;
	
	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		this.userMenuLink = new ExtensionLink(userMenuTabTitle, getFullAlias(), ">", userMenuExtensionLinkSlot);

		if (userFlowInstanceMenuModule != null) {

			userFlowInstanceMenuModule.sortProviders();
		}
		
		numberFormat = (DecimalFormat)NumberFormat.getCurrencyInstance(new Locale("sv", "SE"));
		
		if(!showDecimals) {
			numberFormat.setMaximumFractionDigits(0);
		}
		if(!showCurrency) {
			DecimalFormatSymbols symbols = numberFormat.getDecimalFormatSymbols();
			symbols.setCurrencySymbol("");
			numberFormat.setDecimalFormatSymbols(symbols);
		}
	}

	@InstanceManagerDependency
	public void setUserFlowInstanceMenuModule(UserFlowInstanceMenuModule userFlowInstanceMenuModule) {

		if (userFlowInstanceMenuModule == null && this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.removeUserMenuProvider(this);
		}

		this.userFlowInstanceMenuModule = userFlowInstanceMenuModule;

		if (this.userFlowInstanceMenuModule != null) {

			this.userFlowInstanceMenuModule.addUserMenuProvider(this);
		}
	}

	@Override
	public void unload() throws Exception {

		if (userFlowInstanceMenuModule != null) {

			setUserFlowInstanceMenuModule(null);
		}

		super.unload();
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		Document doc = createDocument(uriParser);

		Element userInvoiceListElement = XMLUtils.appendNewElement(doc, doc.getDocumentElement(), "InvoiceList");

		XMLUtils.append(doc, userInvoiceListElement, "Invoices", getInvoices(user));

		XMLUtils.appendNewElement(doc, userInvoiceListElement, "title", title);
		XMLUtils.appendNewElement(doc, userInvoiceListElement, "description", description);
		XMLUtils.appendNewElement(doc, userInvoiceListElement, "helpText", helpText);

		return new SimpleForegroundModuleResponse(doc, userMenuTabTitle);
	}

	@RESTMethod(alias = "invoice/{invoiceID}", method = "get")
	public void downloadInvoice(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, @URIParam(name = "invoiceID") String invoiceID) throws Throwable {

		if (invoiceDownloadURL != null) {

			List<Invoice> invoices = getInvoices(user);

			if (invoices != null) {

				for (Invoice invoice : invoices) {

					if (invoice.getInvoiceID() != null && invoice.getInvoiceID().equals(invoiceID)) {

						if (invoice.getLongDescription() != null) {

							log.warn("User " + user + " tried to download invoice ID " + invoiceID + " which has long description set");
							break;
						}

						log.info("Getting invoice ID " + invoiceID + " for user " + user + "...");

						SimpleRequest simpleRequest = new SimpleRequest(invoiceDownloadURL.replace("$invoiceID", invoiceID));

						simpleRequest.setFollowRedirects(false);
						simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
						simpleRequest.setReadTimeout(readTimeout * 1000);

						if (apiUsername != null && apiPassword != null) {

							simpleRequest.setUsername(apiUsername);
							simpleRequest.setPassword(apiPassword);
						}

						res.setHeader("Content-Disposition", ContentDisposition.INLINE + ";" + FileUtils.toContentDispositionFilename("Faktura " + invoiceID + ".pdf"));
						res.setContentType(MimeUtils.getMimeType("Faktura " + invoiceID + ".pdf"));

						List<Entry<String, String>> headerEntries = new ArrayList<>(2);

						headerEntries.add(new SimpleEntry<>("NoLoginRedirect", "NoLoginRedirect"));
						simpleRequest.setHeaders(headerEntries);

						try {
							HTTPUtils.sendHTTPGetRequest(simpleRequest, res.getOutputStream());

						} catch (IOException e) {

							log.warn("Error downloading and sending invoice ID " + invoiceID + " to user " + user);
							break;
						}

						return;
					}
				}
			}
		}

		throw new URINotFoundException(uriParser);
	}

	private List<Invoice> getInvoices(User user) {

		String citizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(user);

		if (invoiceListURL != null && citizenID != null) {

			log.info("Getting invoice list for user " + user + "...");

			SimpleRequest simpleRequest = new SimpleRequest(invoiceListURL.replace("$citizenIdentifier", citizenID));

			simpleRequest.setFollowRedirects(false);
			simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
			simpleRequest.setReadTimeout(readTimeout * 1000);

			if (apiUsername != null && apiPassword != null) {

				simpleRequest.setUsername(apiUsername);
				simpleRequest.setPassword(apiPassword);
			}
			
			List<Entry<String, String>> headerEntries = new ArrayList<>(2);

			headerEntries.add(new SimpleEntry<>("NoLoginRedirect", "NoLoginRedirect"));
			simpleRequest.setHeaders(headerEntries);

			try {
				HTTPResponse response = HTTPUtils.sendHTTPGetRequest(simpleRequest, Charset.forName(encoding));

				try {
					XMLParser parser = new XMLParser(XMLUtils.parseXML(response.getValue(), encoding, false, false));

					List<XMLParser> items = parser.getNodes("/Invoices/Invoice");

					if (!items.isEmpty()) {

						List<Invoice> invoices = new ArrayList<>(items.size());

						for (XMLParser item : items) {

							String invoiceID = item.getString("InvoiceID");

							String amount = getAmount(item.getString("Amount"));

							String shortDescription = item.getString("Description");
							String longDescription = item.getString("LongDescription");
							String status = item.getString("Status");

							if (StringUtils.isEmpty(longDescription)) {

								longDescription = null;
							}

							if (StringUtils.isEmpty(status)) {

								status = null;
							}

							boolean payed = item.getPrimitiveBoolean("Payed");

							Date invoiceDate = null;
							Date dueDate = null;

							String invoiceDateString = item.getString("InvoiceDate");
							String dueDateString = item.getString("DueDate");

							if (!StringUtils.isEmpty(invoiceDateString)) {
								invoiceDate = DateUtils.DATE_FORMATTER.parse(invoiceDateString);
							}

							if (!StringUtils.isEmpty(dueDateString)) {
								dueDate = DateUtils.DATE_FORMATTER.parse(dueDateString);
							}

							invoices.add(new Invoice(invoiceID, amount, invoiceDate, dueDate, status, payed, shortDescription, longDescription));
						}

						log.info("Got " + invoices.size() + " invoices for user " + user);

						return invoices;
					}

					log.info("Got no invoices for user " + user);

				} catch (Exception e) {

					log.warn("Error parsing invoice list API response for user " + user, e);
				}

			} catch (Throwable e) {

				log.warn("Error connecting to invoice list API for user " + user, e);
			}
		}

		return null;
	}

	private String getAmount(String amountGiven) {

		BigDecimal amount = NumberUtils.toBigDecimalLocaleIndependent(amountGiven);
		
		return amount == null ? null : numberFormat.format(amount);
	}

	public Document createDocument(URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");
		doc.appendChild(documentElement);

		XMLUtils.appendNewElement(doc, documentElement, "BaseURL", uriParser.getContextPath() + this.getFullAlias());

		return doc;
	}

	@Override
	public AccessInterface getAccessInterface() {

		return moduleDescriptor;
	}

	@Override
	public String getUserMenuAlias() {

		return getFullAlias();
	}

	@Override
	public ExtensionLink getUserMenuExtensionLink(User user) {

		return userMenuLink;
	}

	@Override
	public String getUserMenuPriority() {

		return userMenuExtensionLinkSlot;
	}
}
