package com.nordicpeak.flowengine.evaluators.setweightbyageevaluator;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.xml.XMLGenerator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.beans.QueryModification;
import com.nordicpeak.flowengine.evaluators.baseevaluator.BaseEvaluationProviderModule;
import com.nordicpeak.flowengine.evaluators.setweightevaluator.SetWeightEvaluationEvent;
import com.nordicpeak.flowengine.interfaces.CitizenIdentifierQuery;
import com.nordicpeak.flowengine.interfaces.CitizenIdentifierQueryInstance;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.interfaces.QueryInstance;

public class SetWeightByAgeEvaluationProviderModule extends BaseEvaluationProviderModule<SetWeightByAgeEvaluator> {

	protected SetWeightByAgeEvaluatorCRUD evaluatorCRUD;

	protected AnnotatedDAO<SetWeightByAgeEvaluator> evaluatorDAO;

	protected QueryParameterFactory<SetWeightByAgeEvaluator, Integer> evaluatorIDParamFactory;

	@InstanceManagerDependency(required = true)
	protected QueryHandler queryHandler;

	private boolean supportsQueryInstance(QueryInstance queryInstance) {

		return queryInstance instanceof CitizenIdentifierQueryInstance;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return CitizenIdentifierQuery.class.isAssignableFrom(queryClass);
	}

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, SetWeightByAgeEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(SetWeightByAgeEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new SetWeightByAgeEvaluatorCRUD(SetWeightByAgeEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, SetWeightByAgeEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		if (!supportsQueryInstance(queryInstance)) {

			log.warn("Query instance " + queryInstance + " is not supported for evaluator " + evaluatorDescriptor);

			return null;
		}

		if ((evaluator.getMinAge() == null && evaluator.getMaxAge() == null) || evaluator.getWeightType() == null || evaluator.getWeight() == null) {

			return null;
		}

		String citizenIdentifier = ((CitizenIdentifierQueryInstance) queryInstance).getCitizenIdentifier();
		Integer age;

		if (citizenIdentifier == null || (age = getAge(citizenIdentifier, evaluator.isCalculateBasedOnYearOnly())) == null) {

			return null;
		}

		BigDecimal weight = BigDecimal.ZERO;

		if ((evaluator.getMinAge() == null || age >= evaluator.getMinAge()) && (evaluator.getMaxAge() == null || age <= evaluator.getMaxAge())) {

			weight = BigDecimal.valueOf(evaluator.getWeight());
			
			if (evaluator.isSetAttributeIfAgeInRange()) {
				
				attributeHandler.setAttribute(evaluator.getAttributeName(), evaluator.getAttributeValue());
			}
		}

		List<QueryModification> queryModifications = callback.sendEvent(new SetWeightEvaluationEvent(evaluator.getEvaluatorID(), evaluator.getWeightType(), weight, attributeHandler), true);

		if (queryModifications != null) {

			return new EvaluationResponse(queryModifications);
		}

		return null;

	}

	private Integer getAge(String citizenIdentifier, boolean calculateBasedOnYearOnly) {

		if (calculateBasedOnYearOnly) {

			if (citizenIdentifier.length() < 4) {

				return null;
			}

			Integer birthYear = NumberUtils.toInt(citizenIdentifier.substring(0, 4));

			if (birthYear == null) {

				return null;
			}

			Integer currentYear = Calendar.getInstance().get(Calendar.YEAR);

			return currentYear - birthYear;

		} else {

			if (citizenIdentifier.length() < 8) {

				return null;
			}

			Integer year = NumberUtils.toInt(citizenIdentifier.substring(0, 4));
			Integer month = NumberUtils.toInt(citizenIdentifier.substring(4, 6));
			Integer day = NumberUtils.toInt(citizenIdentifier.substring(6, 8));

			if (year == null || month == null || day == null) {

				return null;
			}

			Calendar calendar = Calendar.getInstance();
			calendar.set(Calendar.YEAR, year);
			calendar.set(Calendar.MONTH, month - 1); //month is 0-based
			calendar.set(Calendar.DAY_OF_MONTH, day);

			return DateUtils.getYearsBetween(calendar, Calendar.getInstance());
		}
	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		SetWeightByAgeEvaluator evaluator = new SetWeightByAgeEvaluator();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());

		return evaluator;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		SetWeightByAgeEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<SetWeightByAgeEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	protected Class<SetWeightByAgeEvaluator> getEvaluatorClass() {

		return SetWeightByAgeEvaluator.class;
	}

	protected AnnotatedDAO<SetWeightByAgeEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	protected QueryParameterFactory<SetWeightByAgeEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return evaluatorIDParamFactory;
	}

	@Override
	public Evaluator getEvaluator(MutableEvaluatorDescriptor descriptor, InstanceMetadata instanceMetadata) throws Throwable {

		SetWeightByAgeEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<SetWeightByAgeEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())));

		if (evaluator != null) {

			evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());
		}

		return evaluator;
	}

	@Override
	public boolean deleteEvaluator(ImmutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		evaluatorDAO.delete(new HighLevelQuery<SetWeightByAgeEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())), transactionHandler);

		return true;
	}

	public QueryHandler getQueryHandler() {

		return this.queryHandler;
	}

	@Override
	public SetWeightByAgeEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		SetWeightByAgeEvaluator evaluator = getEvaluatorClass().newInstance();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		evaluator.populate(descriptor.getImportParser().getNode(XMLGenerator.getElementName(evaluator.getClass())));

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}
}
