package com.nordicpeak.flowengine.evaluators.weightcalculatedquerystateevaluator;

import java.math.BigDecimal;
import java.math.MathContext;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.webutils.http.URIParser;

import com.udojava.evalex.Expression;

import com.nordicpeak.flowengine.evaluators.calculatequerystateevaluator.CalculatedQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.utils.TextTagReplacer;

public class WeightCalculatedQueryStateEvaluationProviderModule extends CalculatedQueryStateEvaluationProviderModule<WeightCalculatedQueryStateEvaluator> {

	protected WeightCalculatedQueryStateEvaluatorCRUD evaluatorCRUD;

	protected AnnotatedDAO<WeightCalculatedQueryStateEvaluator> evaluatorDAO;

	protected QueryParameterFactory<WeightCalculatedQueryStateEvaluator, Integer> evaluatorIDParamFactory;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, WeightCalculatedQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(WeightCalculatedQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new WeightCalculatedQueryStateEvaluatorCRUD(WeightCalculatedQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public Evaluator getEvaluator(MutableEvaluatorDescriptor descriptor, InstanceMetadata instanceMetadata) throws Throwable {

		WeightCalculatedQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<WeightCalculatedQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())));
		
		if (evaluator != null) {
			
			if (instanceMetadata != null && instanceMetadata.getSiteProfile() != null) {
				
				TextTagReplacer.replaceTextTags(evaluator, instanceMetadata.getSiteProfile());
			}
			
			evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());
		}
		
		return evaluator;
	}
	
	@Override
	protected BigDecimal getCalculatedValue(QueryInstance queryInstance, WeightCalculatedQueryStateEvaluator evaluator, User poster) throws ValidationException {

		if (evaluator.getExpression() == null) {
			
			return null;
		}
		
		String populatedExpression = evaluator.getPopulatedExpression();
		
		try {
			Expression expression = new Expression(populatedExpression, MathContext.DECIMAL128);
			
			return expression.eval();
			
		} catch (Exception e) {
			
			log.info("Error evaluating expression " + populatedExpression + " due to " + e);
		}
		
		return null;
	}

	@Override
	protected Class<WeightCalculatedQueryStateEvaluator> getEvaluatorClass() {

		return WeightCalculatedQueryStateEvaluator.class;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}
	
	@Override
	protected boolean supportsQueryInstance(QueryInstance queryInstance) {

		return true;
	}

	@Override
	protected AnnotatedDAO<WeightCalculatedQueryStateEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	@Override
	protected QueryParameterFactory<WeightCalculatedQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return evaluatorIDParamFactory;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		WeightCalculatedQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<WeightCalculatedQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
		
	}

	@Override
	public WeightCalculatedQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		WeightCalculatedQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}

	@Override
	protected List<Integer> getTargetQueryIDs(WeightCalculatedQueryStateEvaluator evaluator, EvaluationCallback callback) {

		return Collections.singletonList(callback.getQueryInstance().getQueryInstanceDescriptor().getQueryDescriptor().getQueryID());
	}
}
