package com.nordicpeak.flowengine.evaluators.weightcalculatedquerystateevaluator;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.annotations.DAOManaged;
import se.unlogic.standardutils.dao.annotations.Table;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLElement;
import se.unlogic.standardutils.xml.XMLParser;
import se.unlogic.standardutils.xml.XMLValidationUtils;

import com.nordicpeak.flowengine.annotations.TextTagReplace;
import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.beans.QueryModification;
import com.nordicpeak.flowengine.evaluators.calculatequerystateevaluator.CalculatedQueryStateEvaluator;
import com.nordicpeak.flowengine.evaluators.setweightevaluator.SetWeightEvaluationEvent;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.EvaluationEventListener;
import com.nordicpeak.flowengine.interfaces.EvaluationHandler;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.QueryInstance;

@Table(name = "weight_calculated_query_state_evaluators")
@XMLElement
public class WeightCalculatedQueryStateEvaluator extends CalculatedQueryStateEvaluator implements EvaluationEventListener {
	
	
	private static final Pattern WEIGHT_TAG_PATTERN = Pattern.compile("(\\$weight)[{](.*?)}");
	
	private static final long serialVersionUID = -1383195490043894435L;
	
	@DAOManaged
	@TextTagReplace
	@WebPopulate(required = true, maxLength = 65535)
	@XMLElement(cdata = true)
	private String expression;
	
	private Set<String> supportedWeights;
	
	private HashMap<String, HashMap<Integer, BigDecimal>> weightMap;
	
	@Override
	public void init(MutableEvaluatorDescriptor evaluatorDescriptor, String configURL) {
		
		super.init(evaluatorDescriptor, configURL);
		
		if (expression != null) {
			supportedWeights = getTags(expression);
		}
	}
	
	@Override
	public boolean requiresInitialization() {
		
		return true;
	}
	
	public String getExpression() {
		
		return expression;
	}
	
	public void setExpression(String expression) {
		
		this.expression = expression;
	}
	
	@Override
	public void populate(XMLParser xmlParser) throws ValidationException {
		
		super.populate(xmlParser);
		
		List<ValidationError> errors = new ArrayList<ValidationError>(1);
		
		expression = XMLValidationUtils.validateParameter("expression", xmlParser, false, 1, 65535, StringPopulator.getPopulator(), errors);
		
		if (!errors.isEmpty()) {
			
			throw new ValidationException(errors);
		}
	}
	
	@Override
	public boolean supportsEvent(Object event) {
		
		if (event instanceof SetWeightEvaluationEvent) {
			
			return true;
		}
		
		return false;
	}
	
	@Override
	public List<QueryModification> processEvent(Object event, QueryInstance queryInstance, User user, User poster, EvaluationCallback callback, EvaluationHandler evaluationHandler) {
		
		SetWeightEvaluationEvent weightEvent = (SetWeightEvaluationEvent) event;
		
		if (supportedWeights != null && supportedWeights.contains(weightEvent.getWeightType())) {
			
			if (weightMap == null) {
				
				weightMap = new HashMap<String, HashMap<Integer, BigDecimal>>();
			}
			
			HashMap<Integer, BigDecimal> weightSourceMap = weightMap.get(weightEvent.getWeightType());
			
			if (weightSourceMap == null) {
				
				weightSourceMap = new HashMap<Integer, BigDecimal>();
				weightMap.put(weightEvent.getWeightType(), weightSourceMap);
			}
			
			weightSourceMap.put(weightEvent.getSourceID(), weightEvent.getWeight());
		}
		
		EvaluationResponse response = super.evaluate(queryInstance, user, poster, callback, evaluationHandler, false, weightEvent.getAttributeHandler());
		
		if (response != null) {
			
			return response.getModifications();
		}
		
		return null;
	}
	
	public String getPopulatedExpression() {
		
		String populatedExpression = expression;
		
		if (this.supportedWeights != null) {
			
			for (String weightType : supportedWeights) {
				
				BigDecimal weight = getWeight(weightType);
				
				populatedExpression = populatedExpression.replace("$weight{" + weightType + "}", weight.toPlainString());
			}
		}
		
		return populatedExpression;
	}
	
	private BigDecimal getWeight(String weightType) {
		
		if (weightMap != null) {
			
			HashMap<Integer, BigDecimal> weightSourceMap = weightMap.get(weightType);
			
			if (weightSourceMap != null) {
				
				BigDecimal totalWeight = BigDecimal.ZERO;
				
				for (BigDecimal weight : weightSourceMap.values()) {
					
					totalWeight = totalWeight.add(weight);
				}
				
				return totalWeight;
			}
		}
		
		return BigDecimal.ZERO;
	}
	
	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, EvaluationCallback callback, EvaluationHandler evaluationHandler, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {
		
		return null;
	}
	
	public static Set<String> getTags(String text) {
		
		Matcher matcher = WEIGHT_TAG_PATTERN.matcher(text);
		
		Set<String> tags = null;
		
		while (matcher.find()) {
			
			String tag = matcher.group(2);
			
			tags = CollectionUtils.addAndInstantiateIfNeeded(tags, tag);
		}
		
		return tags;
	}
	
}
